#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <getopt.h>
#include <iostream>
#include <iomanip>
#include <glibmm.h>
#include <stdexcept>

#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#define EX_OK           0
#define EX_USAGE        64
#define EX_DATAERR      65
#define EX_NOINPUT      66
#define EX_UNAVAILABLE  69
#endif  

#include "engine.h"
#include "aircraft.h"
#include "icaofpl.h"
#include "cfmuautoroute.hh"
#include "cfmuautorouteproxy.hh"

class Command : public CFMUAutorouteProxy::Command {
public:
	Command(void) : CFMUAutorouteProxy::Command() {}
	Command(const std::string& cmd) : CFMUAutorouteProxy::Command(cmd) {}

	using CFMUAutorouteProxy::Command::set_option;
	
	void set_option(const std::string& name, CFMUAutoroute::opttarget_t ot, bool replace = false);
	std::pair<CFMUAutoroute::opttarget_t,bool> get_option_opttarget(const std::string& name) const;
};

static CFMUAutoroute *autoroute(0);
static Glib::RefPtr<Glib::MainLoop> mainloop;
typedef void (*cmd_t)(const Command& cmdin, Command& cmdout);
typedef std::map<std::string,cmd_t> cmdlist_t;
cmdlist_t cmdlist;
Glib::RefPtr<Glib::IOChannel> inputchan;
Glib::RefPtr<Glib::IOChannel> outputchan;

void Command::set_option(const std::string& name, CFMUAutoroute::opttarget_t ot, bool replace)
{
	set_option(name, ::to_str(ot), replace);
}

std::pair<CFMUAutoroute::opttarget_t,bool> Command::get_option_opttarget(const std::string& name) const
{
        std::pair<const std::string&,bool> o(get_option(name));
        std::pair<CFMUAutoroute::opttarget_t,bool> ret(CFMUAutoroute::opttarget_time, o.second);
        if (!ret.second)
                return ret;
        static const CFMUAutoroute::opttarget_t ot[] = {
		CFMUAutoroute::opttarget_time,
		CFMUAutoroute::opttarget_fuel
        };
        unsigned int i;
        for (i = 0; i < sizeof(ot)/sizeof(ot[0]); ++i)
                if (::to_str(ot[i]) == o.first)
                        break;
        if (i >= sizeof(ot)/sizeof(ot[0])) {
                error("invalid " + name + "=" + o.first);
                ret.second = false;
        } else {
                ret.first = ot[i];
        }
        return ret;
}

static void machine_writecmd(const Command& cmd)
{
        outputchan->write(cmd.get_cmdstring() + "\n");
}

static bool machine_input(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(inputchan->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
                mainloop->quit();
                return true;
        }
        while (!line.empty()) {
                char ch(line[line.size() - 1]);
                if (ch != '\r' && ch != '\n')
                        break;
                line.erase(line.size() - 1, 1);
        }
        Command cmdin(line);
        Command cmdout;
        {
                std::string cmdname("autoroute");
                cmdname = cmdin.get_cmdname();
                if (cmdname.empty())
                        cmdname = "autoroute";
                cmdout.set_cmdname(cmdname);
                {
                        std::pair<const std::string&,bool> seq(cmdin.get_option("cmdseq"));
                        if (seq.second)
                                cmdout.set_option("cmdseq", seq.first);
                }
        }
        try {
                Command cmdout1(cmdout);
                cmdlist_t::const_iterator cmdi(cmdlist.find(cmdin.get_cmdname()));
                if (cmdi == cmdlist.end()) {
                        cmdout1.set_option("error", "command not found");
                } else {
                        (cmdi->second)(cmdin, cmdout1);
                }
                machine_writecmd(cmdout1);
        } catch (const std::runtime_error& e) {
                cmdout.set_option("error", e.what());
		machine_writecmd(cmdout);
        } catch (const Glib::Exception& e) {
                cmdout.set_option("error", e.what());
                machine_writecmd(cmdout);
        }
        outputchan->flush();
        return true;
}

static void machine_statuschange(CFMUAutoroute::statusmask_t status)
{
	if (status & (CFMUAutoroute::statusmask_starting |
		      CFMUAutoroute::statusmask_stoppingdone |
		      CFMUAutoroute::statusmask_stoppingerror)) {
		Command cmd("autoroute");
		if (status & CFMUAutoroute::statusmask_starting) {
			cmd.set_option("status", "starting");
		} else if (status & (CFMUAutoroute::statusmask_stoppingdone |
				     CFMUAutoroute::statusmask_stoppingerror)) {
			cmd.set_option("status", "stopping");
			cmd.set_option("routesuccess", (unsigned int)!!(status & CFMUAutoroute::statusmask_stoppingdone));
		}
		machine_writecmd(cmd);
		outputchan->flush();
	}
}

static void machine_autoroutelog(CFMUAutoroute::log_t item, const Glib::ustring& line)
{
	if (item != CFMUAutoroute::log_fplproposal) {
		Command cmd("log");
		cmd.set_option("item", to_str(item));
		cmd.set_option("text", line);
		machine_writecmd(cmd);
		outputchan->flush();
		return;
	}
	// new flight plan
	const FPlanRoute& route(autoroute->get_route());
	{
		Command cmd("fplbegin");
		cmd.set_option("note", route.get_note());
                cmd.set_option("curwpt", route.get_curwpt());
                cmd.set_option("offblock", route.get_time_offblock_unix());
                cmd.set_option("save_offblock", route.get_save_time_offblock_unix());
                cmd.set_option("onblock", route.get_time_onblock_unix());
                cmd.set_option("save_onblock", route.get_save_time_onblock_unix());
		machine_writecmd(cmd);
	}
	for (unsigned int i = 0; i < route.get_nrwpt(); ++i) {
		const FPlanWaypoint& wpt(route[i]);
		Command cmd("fplwpt");
                cmd.set_option("icao", wpt.get_icao());
                cmd.set_option("name", wpt.get_name());
                cmd.set_option("pathname", wpt.get_pathname());
		cmd.set_option("pathcode", wpt.get_pathcode_name());
                cmd.set_option("note", wpt.get_note());
                cmd.set_option("time", wpt.get_time_unix());
                cmd.set_option("save_time", wpt.get_save_time_unix());
                cmd.set_option("frequency", wpt.get_frequency());
                cmd.set_option("coord", wpt.get_coord());
                cmd.set_option("lat", wpt.get_coord().get_lat_deg_dbl());
                cmd.set_option("lon", wpt.get_coord().get_lon_deg_dbl());
                cmd.set_option("altitude", wpt.get_altitude());
                cmd.set_option("flags", wpt.get_flags());
		cmd.set_option("winddir", wpt.get_winddir_deg());
		cmd.set_option("windspeed", wpt.get_windspeed_kts());
		cmd.set_option("qff", wpt.get_qff_hpa());
		cmd.set_option("isaoffset", wpt.get_isaoffset_kelvin());
		machine_writecmd(cmd);
	}
	{
		Command cmd("fplend");
		cmd.set_option("gcdist", autoroute->get_gcdistance());
		cmd.set_option("routedist", autoroute->get_routedistance());
		cmd.set_option("mintime", autoroute->get_mintime());
		cmd.set_option("routetime", autoroute->get_routetime());
		cmd.set_option("minfuel", autoroute->get_minfuel());
		cmd.set_option("routefuel", autoroute->get_routefuel());
		cmd.set_option("fpl", autoroute->get_simplefpl());
		cmd.set_option("iteration", autoroute->get_iterationcount());
		cmd.set_option("localiteration", autoroute->get_local_iterationcount());
		cmd.set_option("remoteiteration", autoroute->get_remote_iterationcount());
		machine_writecmd(cmd);
	}
	outputchan->flush();
}

static void cmd_nop(const Command& cmdin, Command& cmdout)
{
}

static void cmd_quit(const Command& cmdin, Command& cmdout)
{
	mainloop->quit();
}

static void cmd_departure(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<const std::string&,bool> icao(cmdin.get_option("icao"));
		std::pair<const std::string&,bool> name(cmdin.get_option("name"));
		if (icao.second || name.second)	{
			if (!autoroute->set_departure(icao.first, name.first))
				throw std::runtime_error("departure aerodrome " + icao.first + " " + name.first + " not found");
		}
	}
	if (cmdin.is_option("ifr"))
		autoroute->set_departure_ifr(true);
	else if (cmdin.is_option("vfr"))
		autoroute->set_departure_ifr(false);
	if (cmdin.is_option("sid")) {
		std::pair<Point,bool> pt(cmdin.get_option_coord("sid"));
		autoroute->set_sid(pt.first);
	}
	{
		std::pair<double,bool> lim(cmdin.get_option_float("sidlimit"));
		if (lim.second)
			autoroute->set_sidlimit(lim.first);
	}
	{
		std::pair<unsigned long,bool> dtime(cmdin.get_option_uint("time"));
		if (dtime.second)
			autoroute->set_deptime(dtime.first);
	}
	if (autoroute->get_departure().is_valid()) {
		cmdout.set_option("icao", autoroute->get_departure().get_icao());
		cmdout.set_option("name", autoroute->get_departure().get_name());
		cmdout.set_option("coord", autoroute->get_departure().get_coord());
	}
	if (autoroute->get_sid().is_invalid())
		cmdout.set_option("sid", "");
	else
		cmdout.set_option("sid", autoroute->get_sid());
	cmdout.set_option("sidlimit", autoroute->get_sidlimit());
	cmdout.set_option(autoroute->get_departure_ifr() ? "ifr" : "vfr", "");
	cmdout.set_option("time", autoroute->get_deptime());
}

static void cmd_destination(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<const std::string&,bool> icao(cmdin.get_option("icao"));
		std::pair<const std::string&,bool> name(cmdin.get_option("name"));
		if (icao.second || name.second)	{
			if (!autoroute->set_destination(icao.first, name.first))
				throw std::runtime_error("destination aerodrome " + icao.first + " " + name.first + " not found");
		}
	}
	if (cmdin.is_option("ifr"))
		autoroute->set_destination_ifr(true);
	else if (cmdin.is_option("vfr"))
		autoroute->set_destination_ifr(false);
	if (cmdin.is_option("star")) {
		std::pair<Point,bool> pt(cmdin.get_option_coord("star"));
		autoroute->set_star(pt.first);
	}
	{
		std::pair<double,bool> lim(cmdin.get_option_float("starlimit"));
		if (lim.second)
			autoroute->set_starlimit(lim.first);
	}
	if (autoroute->get_destination().is_valid()) {
		cmdout.set_option("icao", autoroute->get_destination().get_icao());
		cmdout.set_option("name", autoroute->get_destination().get_name());
		cmdout.set_option("coord", autoroute->get_destination().get_coord());
	}
	if (autoroute->get_star().is_invalid())
		cmdout.set_option("star", "");
	else
		cmdout.set_option("star", autoroute->get_star());
	cmdout.set_option("starlimit", autoroute->get_starlimit());
	cmdout.set_option(autoroute->get_destination_ifr() ? "ifr" : "vfr", "");
}

static void cmd_enroute(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<double,bool> lim(cmdin.get_option_float("dctlimit"));
		if (lim.second)
			autoroute->set_dctlimit(lim.first);
	}
	{
		std::pair<double,bool> lim(cmdin.get_option_float("dctpenalty"));
		if (lim.second)
			autoroute->set_dctpenalty(lim.first);
	}
	{
		std::pair<double,bool> lim(cmdin.get_option_float("vfraspclimit"));
		if (lim.second)
			autoroute->set_vfraspclimit(lim.first);
	}
	{
		std::pair<unsigned long,bool> o(cmdin.get_option_uint("forceenrouteifr"));
		if (o.second)
			autoroute->set_force_enroute_ifr(!!o.first);
	}
	{
		std::pair<unsigned long,bool> o(cmdin.get_option_uint("honourawylevels"));
		if (o.second)
			autoroute->set_honour_awy_levels(!!o.first);
	}
	cmdout.set_option("dctlimit", autoroute->get_dctlimit());
	cmdout.set_option("dctpenalty", autoroute->get_dctpenalty());
	cmdout.set_option("vfraspclimit", autoroute->get_vfraspclimit());
	cmdout.set_option("forceenrouteifr", (unsigned int)!!autoroute->get_force_enroute_ifr());
	cmdout.set_option("honourawylevels", (unsigned int)!!autoroute->get_honour_awy_levels());
}

static void cmd_levels(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<long,bool> ob(cmdin.get_option_int("base"));
		std::pair<long,bool> ot(cmdin.get_option_int("top"));
		autoroute->set_levels(ob.second ? ob.first : autoroute->get_baselevel(),
				     ot.second ? ot.first : autoroute->get_toplevel());
	}
	cmdout.set_option("base", autoroute->get_baselevel());
	cmdout.set_option("top", autoroute->get_toplevel()); 
}

static void cmd_exclude(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<unsigned long,bool> o(cmdin.get_option_uint("clear"));
		if (o.second && o.first)
			autoroute->clear_excluderegions();
	}
	std::pair<unsigned long,bool> index(cmdin.get_option_uint("index"));
	{
		std::pair<long,bool> ob(cmdin.get_option_int("base"));
		std::pair<long,bool> ot(cmdin.get_option_int("top"));
		std::pair<const std::string&,bool> aspcid(cmdin.get_option("aspcid"));
		std::pair<const std::string&,bool> aspctype(cmdin.get_option("aspctype"));
		std::pair<Point,bool> ptsw(cmdin.get_option_coord("sw"));
		std::pair<Point,bool> ptne(cmdin.get_option_coord("ne"));
		if (aspcid.second) {
			autoroute->add_excluderegion(CFMUAutoroute::ExcludeRegion(aspcid.first, aspctype.second ? aspctype.first : "",
										  ob.second ? ob.first : 0, ot.second ? ot.first : 999));
			if (!index.second) {
				index.first = autoroute->get_excluderegions().size() - 1;
				index.second = true;
			}
		}
		if (ptsw.second && ptne.second) {
			autoroute->add_excluderegion(CFMUAutoroute::ExcludeRegion(Rect(ptsw.first, ptne.first),
										  ob.second ? ob.first : 0, ot.second ? ot.first : 999));
			if (!index.second) {
				index.first = autoroute->get_excluderegions().size() - 1;
				index.second = true;
			}
		}
	}
	cmdout.set_option("count", autoroute->get_excluderegions().size());
	if (index.second && index.first < autoroute->get_excluderegions().size()) {
		const CFMUAutoroute::ExcludeRegion& er(autoroute->get_excluderegions()[index.first]);
		cmdout.set_option("base", er.get_minlevel());
		cmdout.set_option("top", er.get_maxlevel());
		if (er.is_airspace()) {
			cmdout.set_option("aspcid", er.get_airspace_id());
			cmdout.set_option("aspctype", er.get_airspace_type());
		} else {
			cmdout.set_option("sw", er.get_bbox().get_southwest());
			cmdout.set_option("ne", er.get_bbox().get_northeast());
		}
	}

}

static void cmd_tfr(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<unsigned long,bool> o(cmdin.get_option_uint("enabled"));
		if (o.second)
			autoroute->set_tfr_enabled(!!o.first);
	}
	{
		std::pair<const std::string&,bool> o(cmdin.get_option("trace"));
		if (o.second)
			autoroute->set_tfr_trace(o.first);
	}
	{
		std::pair<const std::string&,bool> o(cmdin.get_option("disable"));
		if (o.second)
			autoroute->set_tfr_disable(o.first);
	}
	cmdout.set_option("enabled", (unsigned int)!!autoroute->get_tfr_enabled());
	cmdout.set_option("available", (unsigned int)!!autoroute->get_tfr_available());
	cmdout.set_option("trace", autoroute->get_tfr_trace());
	cmdout.set_option("disable", autoroute->get_tfr_disable());
}

static void cmd_atmosphere(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<double,bool> o(cmdin.get_option_float("qnh"));
		if (o.second)
			autoroute->set_qnh(o.first);
	}
	{
		std::pair<double,bool> o(cmdin.get_option_float("isa"));
		if (o.second)
			autoroute->set_isaoffs(o.first);
	}
	cmdout.set_option("qnh", autoroute->get_qnh());
	cmdout.set_option("isa", autoroute->get_isaoffs());
}

static void cmd_cruise(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<double,bool> o(cmdin.get_option_float("rpm"));
		if (o.second)
			autoroute->set_engine_rpm(o.first);
	}
	{
		std::pair<double,bool> o(cmdin.get_option_float("mp"));
		if (o.second)
			autoroute->set_engine_mp(o.first);
	}
	cmdout.set_option("rpm", autoroute->get_engine_rpm());
	cmdout.set_option("mp", autoroute->get_engine_mp());
}

static void cmd_optimization(const Command& cmdin, Command& cmdout)
{
	{
		std::pair<CFMUAutoroute::opttarget_t,bool> ot(cmdin.get_option_opttarget("target"));
		if (ot.second)
			autoroute->set_opttarget(ot.first);
	}
	cmdout.set_option("target", to_str(autoroute->get_opttarget()));
}

static void cmd_aircraft(const Command& cmdin, Command& cmdout)
{
	std::pair<const std::string&,bool> file(cmdin.get_option("file"));
	if (file.second)
		cmdout.set_option("status", (int)!!autoroute->get_aircraft().load_file(file.first));
	std::pair<const std::string&,bool> data(cmdin.get_option("data"));
	if (data.second)
		cmdout.set_option("status", (int)!!autoroute->get_aircraft().load_string(data.first));
}

static void cmd_start(const Command& cmdin, Command& cmdout)
{
	autoroute->start();
}

static void cmd_stop(const Command& cmdin, Command& cmdout)
{
	autoroute->stop();
}

static void cmd_continue(const Command& cmdin, Command& cmdout)
{
	autoroute->cont();
}

static void cmd_preload(const Command& cmdin, Command& cmdout)
{
	autoroute->preload();
}

static void cmd_clear(const Command& cmdin, Command& cmdout)
{
	autoroute->clear();
}

static void statuschange(CFMUAutoroute::statusmask_t status)
{
	if (false && status & CFMUAutoroute::statusmask_newfpl) {
		std::cout << ">> " << autoroute->get_simplefpl() << std::endl;
	}
	if (false && status & CFMUAutoroute::statusmask_newvalidateresponse) {
		const CFMUAutoroute::validationresponse_t& vr(autoroute->get_validationresponse());
		for (CFMUAutoroute::validationresponse_t::const_iterator i(vr.begin()), e(vr.end()); i != e; ++i)
			std::cout << "<< " << *i << std::endl;
	}
	if (status & CFMUAutoroute::statusmask_starting) {
		std::cerr << "Starting..." << std::endl;
	}
	if (status & CFMUAutoroute::statusmask_stoppingdone) {
		std::cout << "DONE: " << autoroute->get_simplefpl()
			  << " GC " << Conversions::dist_str(autoroute->get_gcdistance())
			  << " route " << Conversions::dist_str(autoroute->get_routedistance())
			  << " fuel " << std::fixed << std::setprecision(1) << autoroute->get_routefuel()
			  << std::endl;
		autoroute->clear();
		mainloop->quit();
		exit(EX_OK);
	}
	if (status & CFMUAutoroute::statusmask_stoppingerror) {
		std::cout << "ERROR" << std::endl;
		autoroute->clear();
		mainloop->quit();
		exit(EX_UNAVAILABLE);
	}
}

static void autoroutelog(CFMUAutoroute::log_t item, const Glib::ustring& line)
{
	switch (item) {
	case CFMUAutoroute::log_fplproposal:
	{
		std::ostringstream oss;
		unsigned int mintime(autoroute->get_mintime());
		unsigned int rtetime(autoroute->get_routetime());
		oss << "# GC Distance " << std::fixed << std::setprecision(1) << autoroute->get_gcdistance()
		    << " Fuel " << std::fixed << std::setprecision(1) << autoroute->get_minfuel()
		    << " Time " << (mintime / 3600U) << ':' << std::setw(2) << std::setfill('0') << ((mintime / 60U) % 60U)
		    << ':' << std::setw(2) << std::setfill('0') << (mintime % 60U)
		    << " Route Distance " << std::fixed << std::setprecision(1) << autoroute->get_routedistance()
		    << " Fuel " << std::fixed << std::setprecision(1) << autoroute->get_routefuel()
		    << " Time " << (rtetime / 3600U) << ':' << std::setw(2) << std::setfill('0') << ((rtetime / 60U) % 60U)
		    << ':' << std::setw(2) << std::setfill('0') << (rtetime % 60U)
		    << " Iteration " << autoroute->get_local_iterationcount()
		    << '/' << autoroute->get_remote_iterationcount() << std::endl;
		std::cout << ">> " << line << std::endl << oss.str();
		break;
	}

	case CFMUAutoroute::log_fplvalidation:
		if (line.empty())
			break;
		std::cout << "<< " << line << std::endl;
		break;

	case CFMUAutoroute::log_graphrule:
		std::cout << "* " << line << std::endl;
		break;

	case CFMUAutoroute::log_graphruledesc:
	case CFMUAutoroute::log_graphruleoprgoal:
		std::cout << "*  " << line << std::endl;
		break;

	case CFMUAutoroute::log_graphchange:
		std::cout << "*   " << line << std::endl;
		break;

	case CFMUAutoroute::log_normal:
	case CFMUAutoroute::log_debug0:
	case CFMUAutoroute::log_debug1:
	default:
		std::cerr << line << std::endl;
		break;
	};
}

int main(int argc, char *argv[])
{
        try {
                Glib::init();
                Glib::set_application_name("CFMU Autoroute");
                Glib::thread_init();
		mainloop = Glib::MainLoop::create(Glib::MainContext::get_default());
		autoroute = new CFMUAutoroute();
		Glib::ustring dir_main(""), dir_aux("");
		Glib::ustring sidname, starname;
		Engine::auxdb_mode_t auxdbmode(Engine::auxdb_prefs);
		bool dis_aux(false), machine(false);
		bool tfrenable(true);

		srand(time(0));

		{
			static struct option long_options[] = {
				{ "maindir", required_argument, 0, 'm' },
				{ "auxdir", required_argument, 0, 'a' },
				{ "disableaux", no_argument, 0, 'A' },
				{ "aircraft", required_argument, 0, 0x100 },
				{ "dctlimit", required_argument, 0, 0x101 },
				{ "depvfr", no_argument, 0, 0x102 },
				{ "depifr", no_argument, 0, 0x103 },
				{ "destvfr", no_argument, 0, 0x104 },
				{ "destifr", no_argument, 0, 0x105 },
				{ "sidlimit", required_argument, 0, 0x106 },
				{ "starlimit", required_argument, 0, 0x107 },
				{ "sid", required_argument, 0, 0x108 },
				{ "star", required_argument, 0, 0x109 },
				{ "airspacelimit", required_argument, 0, 0x10a },
				{ "enrifr", no_argument, 0, 0x10b },
				{ "enrouteifr", no_argument, 0, 0x10b },
				{ "forceenrouteifr", no_argument, 0, 0x10b },
				{ "dctpenalty", no_argument, 0, 0x10c },
				{ "machineinterface", no_argument, 0, 0x10d },
				{ "time", no_argument, 0, 0x10e },
				{ "fuel", no_argument, 0, 0x10f },
				{ "disable-airway-levels", no_argument, 0, 0x110 },
				{ "enable-airway-levels", no_argument, 0, 0x111 },
				{ "disable-tfr", no_argument, 0, 0x112 },
				{ "enable-tfr", no_argument, 0, 0x113 },
				{ "rpm", required_argument, 0, 'R' },
				{ "mp", required_argument, 0, 'M' },
				{ "qnh", required_argument, 0, 'q' },
				{ "isaoffs", required_argument, 0, 't' },
				{0, 0, 0, 0}
			};
			int c, err(0);
			while ((c = getopt_long(argc, argv, "m:a:AR:M:q:t:", long_options, 0)) != EOF) {
				switch (c) {
				case 'm':
					if (optarg)
						dir_main = optarg;
					break;

				case 'a':
					if (optarg)
						dir_aux = optarg;
					break;

				case 'A':
					dis_aux = true;
					break;

				case 0x100:
					if (autoroute->get_aircraft().load_file(optarg))
						break;
					std::cerr << "Unable to load aircraft file \"" << optarg << "\"" << std::endl;
					return EX_NOINPUT;

				case 0x101:
					autoroute->set_dctlimit(strtod(optarg, 0));
					break;

				case 0x102:
				case 0x103:
					autoroute->set_departure_ifr(c & 1);
					break;

				case 0x104:
				case 0x105:
					autoroute->set_destination_ifr(c & 1);
					break;

				case 0x106:
					autoroute->set_sidlimit(strtod(optarg, 0));
					break;

				case 0x107:
					autoroute->set_starlimit(strtod(optarg, 0));
					break;

				case 0x108:
					sidname = optarg;
					break;

				case 0x109:
					starname = optarg;
					break;

				case 0x10a:
					autoroute->set_vfraspclimit(strtod(optarg, 0));
					break;

				case 0x10b:
					autoroute->set_force_enroute_ifr(true);
					break;

				case 0x10c:
					autoroute->set_dctpenalty(strtod(optarg, 0));
					break;

				case 0x10d:
					machine = true;
					break;

				case 0x10e:
					autoroute->set_opttarget(CFMUAutoroute::opttarget_time);
					break;

				case 0x10f:
					autoroute->set_opttarget(CFMUAutoroute::opttarget_fuel);
					break;

				case 0x110:
				case 0x111:
					autoroute->set_honour_awy_levels(c & 1);
					break;

				case 0x112:
				case 0x113:
					tfrenable = c & 1;
					break;

				case 'R':
					autoroute->set_engine_rpm(strtod(optarg, 0));
					break;

				case 'M':
					autoroute->set_engine_mp(strtod(optarg, 0));
					break;

				case 'q':
					autoroute->set_qnh(strtod(optarg, 0));
					break;

				case 't':
					autoroute->set_isaoffs(strtod(optarg, 0));
					break;

				default:
					err++;
					break;
				}
			}
			if (err || (optind + 2 >= argc && !machine)) {
				std::cerr << "usage: cfmuautoroute [-m <maindir>] [-a <auxdir>] [-A] [--aircraft=file.xml]"
					" [--dctlimit=lim] [--dctpenalty=p] [--depvfr] [--depifr] [--destvfr] [--destifr]"
					" [--sidlimit=lim] [--starlimit=lim] [--sid=wpt] [--star=wpt] [--airspacelimit=lim]"
					" [--forceenrouteifr] [--qnh=qnh] [--isaoffs=temp] [--time] [--fuel]"
					" <dep> <dest> <baselvl> <toplvl>" << std::endl;
				return EX_USAGE;
			}
		}
		if (dis_aux)
			auxdbmode = Engine::auxdb_none;
		else if (!dir_aux.empty())
			auxdbmode = Engine::auxdb_override;
		autoroute->set_db_directories(dir_main, auxdbmode, dir_aux);
		autoroute->set_tfr_enabled(tfrenable);
		if (optind + 0 < argc) {
			if (!autoroute->set_departure(argv[optind], argv[optind])) {
				std::cerr << "Departure Airport " << argv[optind] << " not found" << std::endl;
				return EX_DATAERR;
			}
		}
		if (optind + 1 < argc) {
			if (!autoroute->set_destination(argv[optind + 1], argv[optind + 1])) {
				std::cerr << "Destination Airport " << argv[optind + 1] << " not found" << std::endl;
				return EX_DATAERR;
			}
		}
		if (!sidname.empty()) {
			if (!autoroute->set_sid(sidname)) {
				std::cerr << "SID " << sidname << " not found" << std::endl;
				return EX_DATAERR;
			}
		}
		if (!starname.empty()) {
			if (!autoroute->set_star(starname)) {
				std::cerr << "STAR " << starname << " not found" << std::endl;
				return EX_DATAERR;
			}
		}
		if (optind + 2 < argc)
			autoroute->set_levels(strtol(argv[optind + 2], 0, 0), autoroute->get_toplevel());
		if (optind + 3 < argc)
			autoroute->set_levels(autoroute->get_baselevel(), strtol(argv[optind + 3], 0, 0));
		if (machine) {
			cmdlist["nop"] = cmd_nop;
			cmdlist["quit"] = cmd_quit;
			cmdlist["departure"] = cmd_departure;
			cmdlist["destination"] = cmd_destination;
			cmdlist["enroute"] = cmd_enroute;
			cmdlist["levels"] = cmd_levels;
			cmdlist["exclude"] = cmd_exclude;
			cmdlist["tfr"] = cmd_tfr;
			cmdlist["atmosphere"] = cmd_atmosphere;
			cmdlist["cruise"] = cmd_cruise;
			cmdlist["optimization"] = cmd_optimization;
			cmdlist["aircraft"] = cmd_aircraft;
			cmdlist["start"] = cmd_start;
			cmdlist["stop"] = cmd_stop;
			cmdlist["continue"] = cmd_continue;
			cmdlist["preload"] = cmd_preload;
			cmdlist["clear"] = cmd_clear;
			inputchan = Glib::IOChannel::create_from_fd(0);
			outputchan = Glib::IOChannel::create_from_fd(1);
			inputchan->set_encoding();
			outputchan->set_encoding();
			Glib::signal_io().connect(sigc::ptr_fun(&machine_input), inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
			autoroute->signal_statuschange().connect(sigc::ptr_fun(&machine_statuschange));
			autoroute->signal_log().connect(sigc::ptr_fun(&machine_autoroutelog));
			{
				Command cmd("autoroute");
				cmd.set_option("version", VERSION);
				cmd.set_option("provider", "cfmu");
				machine_writecmd(cmd);
			}
			outputchan->flush();
			mainloop->run();
			return EX_OK;
		}
		autoroute->signal_statuschange().connect(sigc::ptr_fun(&statuschange));
		autoroute->signal_log().connect(sigc::ptr_fun(&autoroutelog));
		std::cout << "Departure: " << autoroute->get_departure().get_icao_name() << " ("
			  << (autoroute->get_departure_ifr() ? 'I' : 'V') << "FR)" << std::endl;
		std::cout << "Destination: " << autoroute->get_destination().get_icao_name() << " ("
			  << (autoroute->get_destination_ifr() ? 'I' : 'V') << "FR)" << std::endl;
		if (autoroute->get_sid().is_invalid())
			std::cout << "SID Limit (nmi): " << Conversions::dist_str(autoroute->get_sidlimit()) << std::endl;
		else
			std::cout << "SID: " << autoroute->get_sid().get_lat_str2() << ' ' << autoroute->get_sid().get_lon_str2() << std::endl;
		if (autoroute->get_star().is_invalid())
			std::cout << "STAR Limit (nmi): " << Conversions::dist_str(autoroute->get_starlimit()) << std::endl;
		else
			std::cout << "STAR: " << autoroute->get_star().get_lat_str2() << ' ' << autoroute->get_star().get_lon_str2() << std::endl;
		std::cout << "DCT Limit (nmi): " << Conversions::dist_str(autoroute->get_dctlimit())
			  << " Penalty: " << autoroute->get_dctpenalty() << std::endl;
		if (autoroute->get_departure_ifr() || autoroute->get_destination_ifr() || autoroute->get_force_enroute_ifr())
			std::cout << "Optimization target: " << ((autoroute->get_opttarget() == CFMUAutoroute::opttarget_fuel) ? "Fuel" : "Time") << std::endl;
		else
			std::cout << "VFR Airspace Area Limit (nmi^2): " << Conversions::dist_str(autoroute->get_vfraspclimit()) << std::endl;
		std::cout << "Levels: F" << std::setw(3) << std::setfill('0') << autoroute->get_baselevel()
			  << "...F" << std::setw(3) << std::setfill('0') << autoroute->get_toplevel() << std::endl
			  << "QNH: " << autoroute->get_qnh() << " ISA" << std::showpos << autoroute->get_isaoffs() << std::noshowpos
			  << " Engine: " << autoroute->get_engine_rpm() << '/' << autoroute->get_engine_mp() << std::endl;
		autoroute->start();
		mainloop->run();
	} catch (const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return EX_DATAERR;
        } catch (const Glib::Exception& ex) {
                std::cerr << "Glib exception: " << ex.what() << std::endl;
                return EX_DATAERR;
        } catch (const std::exception& ex) {
                std::cerr << "exception: " << ex.what() << std::endl;
                return EX_DATAERR;
        }
        return autoroute->is_errorfree() ? EX_OK : EX_UNAVAILABLE;
}
