#ifndef CFMUAUTOROUTE_H
#define CFMUAUTOROUTE_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sysdeps.h"
#include "fplan.h"
#include "dbobj.h"
#include "engine.h"
#include "aircraft.h"
#include "baro.h"
#include "tfr.hh"

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <glibmm.h>
#include <sigc++/sigc++.h>

// TODO:
// Path Probing
// Implement Area deletion (for EDGGADF)
// Implement per country DCT limit
// Implement SID/STAR for VFR and IFR

class CFMUAutoroute : public sigc::trackable {
public:
	CFMUAutoroute();
	~CFMUAutoroute();

	void set_db_directories(const std::string& dir_main = "", Engine::auxdb_mode_t auxdbmode = Engine::auxdb_prefs, const std::string& dir_aux = "");
	const std::string& get_db_maindir(void) const { return m_dbdir_main; }
	const std::string& get_db_auxdir(void) const { return m_dbdir_aux; }
	Preferences& get_prefs(void) { return m_prefs; }

	Aircraft& get_aircraft(void) { return m_aircraft; }
	const Aircraft& get_aircraft(void) const { return m_aircraft; }

	typedef enum {
		opttarget_time,
		opttarget_fuel
	} opttarget_t;
	void set_opttarget(opttarget_t t = opttarget_time);
	opttarget_t get_opttarget(void) const { return m_opttarget; }

	const AirportsDb::Airport& get_departure(void) const { return m_airport[0]; }
	void set_departure(const AirportsDb::Airport& el);
	bool set_departure(const Glib::ustring& icao, const Glib::ustring& name);
	bool get_departure_ifr(void) const { return m_airportifr[0]; }
	void set_departure_ifr(bool ifr);
	const Point& get_sid(void) const { return m_airportconn[0]; }
	void set_sid(const Point& pt = Point::invalid);
	bool set_sid(const Glib::ustring& name);
	double get_sidlimit(void) const { return m_airportconnlimit[0]; }
	void set_sidlimit(double l = 20.0);

	const AirportsDb::Airport& get_destination(void) const { return m_airport[1]; }
	void set_destination(const AirportsDb::Airport& el);
	bool set_destination(const Glib::ustring& icao, const Glib::ustring& name);
	bool get_destination_ifr(void) const { return m_airportifr[1]; }
	void set_destination_ifr(bool ifr);
	const Point& get_star(void) const { return m_airportconn[1]; }
	void set_star(const Point& pt = Point::invalid);
	bool set_star(const Glib::ustring& name);
	double get_starlimit(void) const { return m_airportconnlimit[1]; }
	void set_starlimit(double l = 20.0);

	double get_dctlimit(void) const { return m_dctlimit; }
	void set_dctlimit(double l);
	double get_dctpenalty(void) const { return m_dctpenalty; }
	void set_dctpenalty(double l);
	double get_vfraspclimit(void) const { return m_vfraspclimit; }
	void set_vfraspclimit(double l);

	int get_baselevel(void) const { return m_levels[0]; }
	int get_toplevel(void) const { return m_levels[1]; }
	void set_levels(int base, int top);

	bool get_honour_awy_levels(void) const { return m_honourawylevels; }
	void set_honour_awy_levels(bool awylvl);

	class ExcludeRegion {
	public:
		ExcludeRegion(const std::string& aspcid, const std::string& aspctype, int minlevel = 0, int maxlevel = 999)
			: m_aspcid(aspcid), m_aspctype(aspctype), m_minlevel(minlevel), m_maxlevel(maxlevel) {}
		ExcludeRegion(const Rect& bbox, int minlevel = 0, int maxlevel = 999)
			: m_bbox(bbox), m_minlevel(minlevel), m_maxlevel(maxlevel) {}
		bool is_airspace(void) const { return !(m_aspcid.empty() && m_aspctype.empty()); }
		const std::string& get_airspace_id(void) const { return m_aspcid; }
		const std::string& get_airspace_type(void) const { return m_aspctype; }
		const Rect& get_bbox(void) const { return m_bbox; }
		int get_minlevel(void) const { return m_minlevel; }
		int get_maxlevel(void) const { return m_maxlevel; }

	protected:
		std::string m_aspcid;
		std::string m_aspctype;
		Rect m_bbox;
		int m_minlevel;
		int m_maxlevel;
	};
	typedef std::vector<ExcludeRegion> excluderegions_t;

	const excluderegions_t& get_excluderegions(void) const { return m_excluderegions; }
	void set_excluderegions(const excluderegions_t& r);
	void add_excluderegion(const ExcludeRegion& r);
	void clear_excluderegions(void);

	bool get_force_enroute_ifr(void) const { return m_forceenrifr; }
	void set_force_enroute_ifr(bool ifr);

	bool get_tfr_available(void) const;
	bool get_tfr_enabled(void) const { return m_tfrenable; }
	void set_tfr_enabled(bool ena = true);

	void set_tfr_trace(const std::string& rules);
	std::string get_tfr_trace(void) const;
	void set_tfr_disable(const std::string& rules);
	std::string get_tfr_disable(void) const;

	double get_qnh(void) const { return m_qnh; }
	void set_qnh(double q = IcaoAtmosphere::std_sealevel_pressure);
	double get_isaoffs(void) const { return m_isaoffs; }
	void set_isaoffs(double t = 0);
	double get_engine_rpm(void) const { return m_enginerpm; }
	void set_engine_rpm(double l);
	double get_engine_mp(void) const { return m_enginemp; }
	void set_engine_mp(double l);

	time_t get_deptime(void) const { return m_route.get_time_offblock_unix(); }
	void set_deptime(time_t t);

	const FPlanRoute& get_route(void) const { return m_route; }
	Glib::ustring get_simplefpl(void) const;
	typedef std::vector<Glib::ustring> validationresponse_t;
	const validationresponse_t& get_validationresponse(void) const { return m_validationresponse; }

	double get_gcdistance(void) const;
	double get_routedistance(void) const;
	unsigned int get_mintime(void) const;
	unsigned int get_routetime(void) const;
	double get_minfuel(void) const;
	double get_routefuel(void) const;
	unsigned int get_local_iterationcount(void) const { return m_iterationcount[0]; }
	unsigned int get_remote_iterationcount(void) const { return m_iterationcount[1]; }
	unsigned int get_iterationcount(void) const { return get_local_iterationcount() + get_remote_iterationcount(); }

	bool is_running(void) const;
	bool is_errorfree(void) const;

	void start(void);
	void cont(void);
	void stop(void);
	void preload(void);
	void clear(void);

	typedef enum {
		statusmask_starting = 1U << 0,
		statusmask_stoppingdone = 1U << 1,
		statusmask_stoppingerror = 1U << 2,
		statusmask_newfpl = 1U << 3,
		statusmask_newvalidateresponse = 1U << 4
	} statusmask_t;
	sigc::signal<void,statusmask_t> signal_statuschange(void) { return m_signal_statuschange; }

	typedef enum {
		log_fplproposal,
		log_fplvalidation,
		log_graphrule,
		log_graphruledesc,
		log_graphruleoprgoal,
		log_graphchange,
		log_normal,
		log_debug0,
		log_debug1
	} log_t;
	sigc::signal<void,log_t,const Glib::ustring&> signal_log(void) { return m_signal_log; }

protected:
	typedef bool (CFMUAutoroute::*parseresponse_t)(Glib::MatchInfo&);
	static const struct parsefunctions {
		const char *regex;
		parseresponse_t func;
	} parsefunctions[];
	static const char *ignoreregex[];

	class Performance {
	public:
		Performance(int alt = 0, double secpernmi = 0, double fuelpersec = 0, double timepenalty = 0, double fuelpenalty = 0)
			: m_secpernmi(secpernmi), m_fuelpersec(fuelpersec), m_timepenalty(timepenalty), m_fuelpenalty(fuelpenalty), m_altitude(alt) {}

		int get_altitude(void) const { return m_altitude; }

		// Cruise Performance
		double get_secpernmi(void) const { return m_secpernmi; }
		double get_fuelpersec(void) const { return m_fuelpersec; }

		// Climb to next level Performance
		double get_timepenalty(void) const { return m_timepenalty; }
		double get_fuelpenalty(void) const { return m_fuelpenalty; }

		// For FPL
		double get_tas(void) const { return 3600.0 / m_secpernmi; }
		std::string get_fpltas(void) const;

	protected:
		double m_secpernmi;
		double m_fuelpersec;
		double m_timepenalty;
		double m_fuelpenalty;
		int m_altitude;
	};

        Preferences m_prefs;
	AirportsDb m_airportdb;
	NavaidsDb m_navaiddb;
	WaypointsDb m_waypointdb;
	AirwaysDb m_airwaydb;
	AirspacesDb m_airspacedb;
	MapelementsDb m_mapelementdb;
	TopoDb30 m_topodb;
        FPlan m_fplandb;
	TrafficFlowRestrictions m_tfr;
	std::string m_dbdir_main;
	std::string m_dbdir_aux;
	Aircraft m_aircraft;
	FPlanRoute m_route;
	sigc::signal<void,statusmask_t> m_signal_statuschange;
	sigc::signal<void,log_t,const Glib::ustring&> m_signal_log;
	typedef std::vector<Performance> performance_t;
	performance_t m_performance;
	unsigned int m_routetime;
	double m_routefuel;
	double m_deptimepenalty;
	double m_depfuelpenalty;
	AirportsDb::Airport m_airport[2];
	Point m_airportconn[2];
	double m_airportconnlimit[2];
	bool m_airportifr[2];
	double m_dctlimit;
	double m_dctpenalty;
	double m_vfraspclimit;
	double m_qnh;
	double m_isaoffs;
	double m_enginerpm;
	double m_enginemp;
	int m_levels[2];
	time_t m_starttime;
	excluderegions_t m_excluderegions;
	Glib::ustring m_callsign;
	validationresponse_t m_validationresponse;
	sigc::connection m_connchildtimeout;
	sigc::connection m_connchildwatch;
	sigc::connection m_connchildstdout;
	Glib::RefPtr<Glib::IOChannel> m_childchanstdin;
	Glib::RefPtr<Glib::IOChannel> m_childchanstdout;
	Glib::Pid m_childpid;
	unsigned int m_childtimeoutcount;
	unsigned int m_iterationcount[2];
	int m_pathprobe;
	opttarget_t m_opttarget;
	bool m_childrun;
	bool m_forceenrifr;
	bool m_tfrenable;
	bool m_tfrloaded;
	bool m_honourawylevels;
	bool m_running;
	bool m_done;

	void opendb(void);
	void closedb(void);
	std::string get_tfr_filename(void) const { return Glib::build_filename(get_db_auxdir(), "eRAD.xml"); }
	bool find_airport(AirportsDb::Airport& el, const Glib::ustring& icao, const Glib::ustring& name);
	bool find_point(Point& pt, const Glib::ustring& name, bool ifr, const Point& ptnear);
	bool is_optfuel(void) const { return m_opttarget == opttarget_fuel; }
	void updatefpl(void);
	Rect get_bbox(void) const;
	void logmessage(const TrafficFlowRestrictions::Message& msg);
	void computeperformance(void);
	unsigned int findperfindex(int alt) const;
	void start(bool cont);
	bool start_vfr(void);
	bool start_ifr(bool cont = false, bool optfuel = false);
	bool start_ifr_specialcase(void);
	bool start_ifr_specialcase_pogo(void);
	void child_watch(GPid pid, int child_status);
	void child_run(void);
	void child_close(void);
	bool child_stdout_handler(Glib::IOCondition iocond);
	bool child_timeout(void);
	void parse_response(void);
	bool parse_response_route49(Glib::MatchInfo& minfo);
	bool parse_response_route52(Glib::MatchInfo& minfo);
	bool parse_response_route130(Glib::MatchInfo& minfo);
	bool parse_response_route134(Glib::MatchInfo& minfo);
	bool parse_response_route135(Glib::MatchInfo& minfo);
	bool parse_response_route139(Glib::MatchInfo& minfo);
	bool parse_response_route140(Glib::MatchInfo& minfo);
	bool parse_response_route165(Glib::MatchInfo& minfo);
	bool parse_response_route168(Glib::MatchInfo& minfo);
	bool parse_response_route172(Glib::MatchInfo& minfo);
	bool parse_response_prof50(Glib::MatchInfo& minfo);
	bool parse_response_prof194(Glib::MatchInfo& minfo);
	bool parse_response_prof195(Glib::MatchInfo& minfo);
	bool parse_response_prof197(Glib::MatchInfo& minfo);
	bool parse_response_prof198(Glib::MatchInfo& minfo);
	bool parse_response_prof199(Glib::MatchInfo& minfo);
	bool parse_response_prof201(Glib::MatchInfo& minfo);
	bool parse_response_prof201b(Glib::MatchInfo& minfo);
	bool parse_response_prof204(Glib::MatchInfo& minfo);
	bool parse_response_prof204b(Glib::MatchInfo& minfo);
	bool parse_response_prof204c(Glib::MatchInfo& minfo);
	bool parse_response_prof204d(Glib::MatchInfo& minfo);
	bool parse_response_prof205(Glib::MatchInfo& minfo);
	bool parse_response_prof205b(Glib::MatchInfo& minfo);
	bool parse_response_prof205c(Glib::MatchInfo& minfo);
	bool parse_response_prof206(Glib::MatchInfo& minfo);

	class VFRRouter {
	public:
		typedef Engine::AirwayGraphResult::Graph Graph;
		typedef Engine::AirwayGraphResult::Vertex Vertex;
		typedef Engine::AirwayGraphResult::Edge Edge;

		VFRRouter(CFMUAutoroute& ar);
		bool load_graph(const Rect& bbox);
		bool set_departure(const AirportsDb::Airport& arpt);
		bool set_destination(const AirportsDb::Airport& arpt);
		void add_dct(double dctlim, double sidlim, double starlim);
		bool add_sid(const Point& pt = Point::invalid);
		bool add_sid(const AirportsDb::Airport& arpt);
		bool add_star(const Point& pt = Point::invalid);
		bool add_star(const AirportsDb::Airport& arpt);
		void set_metric(void);
		void exclude_airspace(int baselevel, int toplevel, double maxarea = std::numeric_limits<double>::max());
		void exclude_regions(void);
		bool autoroute(FPlanRoute& route);
		const Graph& get_graph(void) const { return m_graph; }
		Graph& get_graph(void) { return m_graph; }

	protected:
		CFMUAutoroute& m_ar;
		Graph m_graph;
		Rect m_bbox;
		Graph::vertex_descriptor m_dep;
		Graph::vertex_descriptor m_dest;
		typedef std::vector<AirportsDb::Airport::VFRRoute> sidstar_t;
		sidstar_t m_sid;
		sidstar_t m_star;

		Graph::vertex_descriptor find_airport(const AirportsDb::Airport& arpt);

		void remove_sid(void);
		void remove_star(void);
	};

	class DctCache {
	public:
		typedef unsigned long pointid_t;
		typedef TrafficFlowRestrictions::DctParameters::AltRange::set_t set_t;

		class PointID {
		public:
			PointID(const Glib::ustring& ident = "", const Point& pt = Point(), pointid_t id = 0) : m_ident(ident), m_point(pt), m_id(id) {}
			const Glib::ustring& get_ident(void) const { return m_ident; }
			const Point& get_point(void) const { return m_point; }
			pointid_t get_id(void) const { return m_id; }
			bool operator<(const PointID& x) const;

		protected:
			Glib::ustring m_ident;
			Point m_point;
			pointid_t m_id;
		};

		class DctAlt {
		public:
			DctAlt(pointid_t pt0 = 0, pointid_t pt1 = 0, const set_t& alt = set_t(), bool dirty = false) : m_alt(alt), m_point0(pt0), m_point1(pt1), m_dirty(dirty) {}
			const set_t& get_alt(void) const { return m_alt; }
		        set_t& get_alt(void) { return m_alt; }
			pointid_t get_point0(void) const { return m_point0; }
			pointid_t get_point1(void) const { return m_point1; }
			bool operator<(const DctAlt& x) const;
			bool is_dirty(void) const { return m_dirty; }
			void set_dirty(bool x = true) { m_dirty = true; }

		protected:
			set_t m_alt;
			pointid_t m_point0;
			pointid_t m_point1;
			bool m_dirty;
		};

		DctCache(const Glib::ustring& path = "", const Glib::ustring& revision = "1900-01-01T00:00:00");

		Glib::ustring get_property(const Glib::ustring& key);
		void set_property(const Glib::ustring& key, const Glib::ustring& value);

		void load(const Rect& bbox);
		void flush(void);

		pointid_t find_point(const Glib::ustring& ident, const Point& pt);
		const DctAlt *find_dct(pointid_t pt0, pointid_t pt1) const;
		void add_dct(pointid_t pt0, pointid_t pt1, const set_t& alt);

		unsigned int get_nrpoints(void) const { return m_points.size(); }
		unsigned int get_nrdcts(void) const { return m_dcts.size(); }

	protected:
		sqlite3x::sqlite3_connection m_db;
		typedef std::set<PointID> points_t;
		points_t m_points;
		typedef std::set<DctAlt> dcts_t;
		dcts_t m_dcts;
		pointid_t m_pointid;
		pointid_t m_pointidalloc;
	};

	typedef uint32_t lgraphairwayindex_t;
	static const lgraphairwayindex_t airwaymatchall = 0x7FFFFFFFU;
	static const lgraphairwayindex_t airwaymatchnone = airwaymatchall - 1U;
	static const lgraphairwayindex_t airwaymatchawy = airwaymatchnone - 1U;
	static const lgraphairwayindex_t airwaymatchdctawy = airwaymatchawy - 1U;
	static const lgraphairwayindex_t airwaymatchdctawysidstar = airwaymatchdctawy - 1U;
	static const lgraphairwayindex_t airwaymatchdctawysidstarlevel = airwaymatchdctawysidstar - 1U;
	static const lgraphairwayindex_t airwaymatchsidstar = airwaymatchdctawysidstarlevel - 1U;
	static const lgraphairwayindex_t airwaymatchlevel = airwaymatchsidstar - 1U;
	static const lgraphairwayindex_t airwaylevelup = airwaymatchlevel - 1U;
	static const lgraphairwayindex_t airwayleveldown = airwaylevelup - 1U;
	static const lgraphairwayindex_t airwaysid = airwayleveldown - 1U;
	static const lgraphairwayindex_t airwaystar = airwaysid - 1U;
	static const lgraphairwayindex_t airwaydct = airwaystar - 1U;
	typedef std::set<Glib::ustring> lgraphairwayidents_t;

	class LVertex {
	public:
		LVertex(unsigned int superindex = 0)
			: m_superindex(superindex) {}

		unsigned int get_superindex(void) const { return m_superindex; }

	protected:
		unsigned int m_superindex;
	};

	struct Combine;

	class LEdge {
	public:
		static const double maxmetric;
		static const lgraphairwayindex_t identmask = 0x7FFFFFFFU;
		static const lgraphairwayindex_t solutionmask = 0x80000000U;
		LEdge(lgraphairwayindex_t id = airwaydct, double metric = maxmetric)
			: m_ident(id & identmask), m_metric(metric) {}

		void set_ident(lgraphairwayindex_t id) { m_ident ^= (m_ident ^ id) & identmask; }
		lgraphairwayindex_t get_ident(void) const { return m_ident & identmask; }
		bool is_dct(void) const { return get_ident() == airwaydct; }
		static bool is_match(lgraphairwayindex_t edgeid, lgraphairwayindex_t id);
		bool is_match(lgraphairwayindex_t id) const { return is_match(get_ident(), id); }
		void set_solution(bool s) { m_ident ^= (m_ident ^ -!!s) & solutionmask; }
		bool is_solution(void) const { return !!(m_ident & solutionmask); }
		void set_metric(double metric = maxmetric) { m_metric = metric; }
		double get_metric(void) const { return m_metric; }

	protected:
	        lgraphairwayindex_t m_ident;

	public:
		double m_metric;
	};

	typedef boost::adjacency_list<boost::multisetS, boost::vecS, boost::bidirectionalS, LVertex, LEdge> LGraph;

	class LSuperVertex {
	public:
		LSuperVertex(Engine::DbObject::ptr_t obj, LGraph::vertex_descriptor v0, LGraph::vertex_descriptor v1)
			: m_obj(obj), m_coord(Point::invalid), m_v0(v0), m_v1(v1) {
			if (m_obj)
				m_coord = m_obj->get_coord();
		}

		bool is_airport(void) const;
		bool is_navaid(void) const;
		bool is_intersection(void) const;
		bool is_mapelement(void) const;

		bool get(AirportsDb::Airport& el) const;
		bool get(NavaidsDb::Navaid& el) const;
		bool get(WaypointsDb::Waypoint& el) const;
		bool get(MapelementsDb::Mapelement& el) const;

		bool set(FPlanWaypoint& el) const;
		unsigned int insert(FPlanRoute& route, uint32_t nr = ~0U) const;

		Engine::DbObject::ptr_t get_object(void) const { return m_obj; }
		const Glib::ustring& get_ident(void) const;
		bool is_ident_valid(void) const;
		const Point& get_coord(void) const { return m_coord; }
		LGraph::vertex_descriptor get_v0(void) const { return m_v0; }
		LGraph::vertex_descriptor get_v1(void) const { return m_v1; }

	protected:
		Engine::DbObject::ptr_t m_obj;
		Point m_coord;
		LGraph::vertex_descriptor m_v0;
		LGraph::vertex_descriptor m_v1;
	};

	typedef std::vector<LSuperVertex> LSuperVertices;

	class LVertexEdge {
	public:
		// v should be the target of e
		LVertexEdge(LGraph::vertex_descriptor v = 0, lgraphairwayindex_t e = airwaymatchnone) : m_vertex(v), m_edge(e) {}
		LVertexEdge(LGraph::edge_descriptor e, const LGraph& g);
		LGraph::vertex_descriptor get_vertex(void) const { return m_vertex; }
		lgraphairwayindex_t get_edge(void) const { return m_edge; }
		bool is_match(lgraphairwayindex_t id) const { return LEdge::is_match(get_edge(), id); }
		bool operator<(const LVertexEdge& x) const;
		bool operator>(const LVertexEdge& x) const { return x < *this; }

	protected:
		LGraph::vertex_descriptor m_vertex;
		lgraphairwayindex_t m_edge;
	};

	class LRoute : public std::vector <LVertexEdge> {
	public:
		LRoute(void) : m_metric(std::numeric_limits<double>::quiet_NaN()) {}

		bool operator<(const LRoute& x) const;
		void set_metric(double metric) { m_metric = metric; }
		double get_metric(void) const { return m_metric; }
		bool is_inf(void) const { return std::isinf(get_metric()); }
		bool is_nan(void) const { return std::isnan(get_metric()); }
		bool is_repeated_nodes(const LGraph& g, const LSuperVertices& supervertices) const;
		bool is_existing(const LGraph& g) const;
		std::pair<LGraph::edge_descriptor,bool> get_edge(size_type i, const LGraph& g) const;
		static std::pair<LGraph::edge_descriptor,bool> get_edge(const LVertexEdge& u, const LVertexEdge& v, const LGraph& g);
		static std::pair<LGraph::edge_descriptor,bool> get_edge(LGraph::vertex_descriptor u, const LVertexEdge& v, const LGraph& g);

	protected:
		double m_metric;
	};

	class LGraphSolutionTreeNode;
	class LGraphSolutionTreeNode : public std::map<LVertexEdge,LGraphSolutionTreeNode> {
	};

	class LEdgeAdd {
	public:
		LEdgeAdd(LGraph::vertex_descriptor v0 = 0, LGraph::vertex_descriptor v1 = 0, double metric = std::numeric_limits<double>::max());
		LGraph::vertex_descriptor get_v0(void) const { return m_v0; }
		LGraph::vertex_descriptor get_v1(void) const { return m_v1; }
		double get_metric(void) const { return m_metric; }
		bool operator<(const LEdgeAdd& x) const;

	protected:
		LGraph::vertex_descriptor m_v0;
		LGraph::vertex_descriptor m_v1;
		double m_metric;
	};

	class LGraphAwyEdgeFilter {
	public:
		LGraphAwyEdgeFilter(void) : m_lgraph(0), m_airway(airwaymatchnone) {}
		LGraphAwyEdgeFilter(LGraph& g, lgraphairwayindex_t awy) : m_lgraph(&g), m_airway(awy) {}
		template <typename E> bool operator()(const E& e) const { return (*m_lgraph)[e].is_match(m_airway); }

	protected:
		LGraph *m_lgraph;
		lgraphairwayindex_t m_airway;
	};

	class LGraphSolutionEdgeFilter {
	public:
		LGraphSolutionEdgeFilter(void) : m_lgraph(0) {}
		LGraphSolutionEdgeFilter(LGraph& g) : m_lgraph(&g) {}
		template <typename E> bool operator()(const E& e) const { return (*m_lgraph)[e].is_solution(); }

	protected:
		LGraph *m_lgraph;
	};

	class DijkstraCore;

	class LGMandatoryPoint {
	public:
		LGMandatoryPoint(LGraph::vertex_descriptor v0 = 0, LGraph::vertex_descriptor v1 = 0,
				 lgraphairwayindex_t awy = airwaymatchnone) : m_v0(v0), m_v1(v1), m_airway(awy) {}
	        LGraph::vertex_descriptor get_v0(void) const { return m_v0; }
	        LGraph::vertex_descriptor get_v1(void) const { return m_v1; }
		lgraphairwayindex_t get_airway(void) const { return m_airway; }

	protected:
		LGraph::vertex_descriptor m_v0;
		LGraph::vertex_descriptor m_v1;
		lgraphairwayindex_t m_airway;
	};

	class LGMandatorySequence : public std::vector<LGMandatoryPoint> {
	};

	class LGMandatoryAlternative : public std::vector<LGMandatorySequence> {
	};

	class LGMandatory : public std::vector<LGMandatoryAlternative> {
	};

	LGraph m_lgraph;
	typedef std::vector<Glib::ustring> lgraphairwaytable_t;
	lgraphairwaytable_t m_lgraphairwaytable;
	typedef std::map<Glib::ustring,lgraphairwayindex_t> lgraphairwayrevtable_t;
	lgraphairwayrevtable_t m_lgraphairwayrevtable;
	LSuperVertices m_lsupervertices;
	typedef std::set<LGraph::vertex_descriptor> solutionvertices_t;
	solutionvertices_t m_solutionvertices;
	// Yen's k shortest loopless paths algorithm
	LGraphSolutionTreeNode m_solutiontree;
	typedef std::set<LRoute> lgraphsolutionpool_t;
	lgraphsolutionpool_t m_solutionpool;
	unsigned int m_superdep;
	unsigned int m_superdest;
	static const bool lgraphrestartedgescan = false;
	static const bool lgraphawyvertdct = false;

	void clearlgraph(void);
	lgraphairwayindex_t lgraphairway2index(const Glib::ustring& awy, bool create = false);
	const Glib::ustring& lgraphindex2airway(lgraphairwayindex_t idx, bool specialnames = false);
	unsigned int lgraphaddvertex(Engine::DbObject::ptr_t obj, bool supernode = true, bool levelchg = true);
	bool lgraphload(const Rect& bbox);
	void lgraphkillinvalidsuper(void);
	static bool lgraphcheckintersect(const Point& pt0, const Point& pt1, const Rect& bbox);
	static bool lgraphcheckintersect(const Point& pt0, const Point& pt1, const Rect& bbox, const MultiPolygonHole& poly);
	void lgraphexcluderegions(void);
	void lgraphadddct(void);
	bool lgraphaddsidstar(void);
	void lgraphupdatemetric(LRoute& r);
	void lgraphclearcurrent(void);
	bool lgraphsetcurrent(const LRoute& r);
	void lgraphmodified(void);
	bool lgraphisinsolutiontree(const LRoute& r) const;
	bool lgraphinsertsolutiontree(const LRoute& r);
	std::pair<LGraph::edge_descriptor,bool> lgraphfindbestedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, bool solutiononly) const;
	LRoute lgraphdijkstraroute(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, const std::vector<LGraph::vertex_descriptor>& predecessors) const;
	std::string lgraphprint(const LRoute& r);
	LRoute lgraphdijkstra(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v, const LRoute& baseroute = LRoute(), bool solutionedgeonly = false, LGMandatory mandatory = LGMandatory());
	LRoute lgraphnextsolution(void);
	void lgraphroute(void);
	bool lgraphlocaltfrcheck(TrafficFlowRestrictions::Result& res);
	LGMandatory lgraphmandatoryroute(const TrafficFlowRestrictions::Result& res);
	void lgraphsubmitroute(void);
	bool lgraphtfrruleresult(const TrafficFlowRestrictions::RuleResult& rule, const TrafficFlowRestrictions::Fpl& fpl);
	void lgraphtfrmessage(const TrafficFlowRestrictions::Message& msg, const TrafficFlowRestrictions::Fpl& fpl);
	bool lgraphkilllevelchg(unsigned int superindex);
	bool lgraphkilllevelchg(const LSuperVertex& super);
	Glib::ustring lgraphvertexname(LGraph::vertex_descriptor v);
	void lgraphlogkilledge(LGraph::edge_descriptor e);
	void lgraphlogaddedge(LGraph::edge_descriptor e);
	void lgraphlogrenameedge(LGraph::edge_descriptor e, lgraphairwayindex_t awyname);
	void lgraphlogscaleedge(LGraph::edge_descriptor e, double scale);
	bool lgraphhasdctedge(LGraph::edge_descriptor e);
	bool lgraphhasdctedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v);
	bool lgraphcanadddctedge(LGraph::vertex_descriptor u, LGraph::vertex_descriptor v);
	bool lgraphdisconnectvertex(const Glib::ustring& ident, int lvlfrom = 0, int lvlto = 600);
	bool lgraphdisconnectsolutionvertex(const Glib::ustring& ident);
	bool lgraphdisconnectsolutionedge(const Glib::ustring& id0, const Glib::ustring& id1);
	bool lgraphscalesolutionvertex(const Glib::ustring& ident, double scale);
	bool lgraphscalesolutionedge(const Glib::ustring& id0, const Glib::ustring& id1, double scale);
	bool lgraphmodifyedge(const Glib::ustring& id0, const Glib::ustring& id1, lgraphairwayindex_t awyremove,
			      lgraphairwayindex_t awyadd = airwaymatchnone, int lvlfrom = 0, int lvlto = 600, bool bidir = false);
	bool lgraphmodifyawysegment(const Glib::ustring& id0, const Glib::ustring& id1, lgraphairwayindex_t awymatch,
				    lgraphairwayindex_t awyrename = airwaymatchnone, int lvlfrom = 0, int lvlto = 600, bool bidir = false, bool solutiononly = false);
	bool lgraphchangeoutgoing(const Glib::ustring& id, lgraphairwayindex_t awyname);
	bool lgraphchangeincoming(const Glib::ustring& id, lgraphairwayindex_t awyname);
	bool lgraphedgetodct(lgraphairwayindex_t awy);
	bool lgraphkilloutgoing(const Glib::ustring& id);
	bool lgraphkillincoming(const Glib::ustring& id);
	bool lgraphsolutiongroundclearance(void);
	static bool is_metric_invalid(double m);
};

inline CFMUAutoroute::statusmask_t& operator|=(CFMUAutoroute::statusmask_t& a, CFMUAutoroute::statusmask_t b) { return a = (CFMUAutoroute::statusmask_t)(a | b); }
inline CFMUAutoroute::statusmask_t& operator&=(CFMUAutoroute::statusmask_t& a, CFMUAutoroute::statusmask_t b) { return a = (CFMUAutoroute::statusmask_t)(a & b); }
inline CFMUAutoroute::statusmask_t& operator^=(CFMUAutoroute::statusmask_t& a, CFMUAutoroute::statusmask_t b) { return a = (CFMUAutoroute::statusmask_t)(a ^ b); }
inline CFMUAutoroute::statusmask_t operator|(CFMUAutoroute::statusmask_t a, CFMUAutoroute::statusmask_t b) { a |= b; return a; }
inline CFMUAutoroute::statusmask_t operator&(CFMUAutoroute::statusmask_t a, CFMUAutoroute::statusmask_t b) { a &= b; return a; }
inline CFMUAutoroute::statusmask_t operator^(CFMUAutoroute::statusmask_t a, CFMUAutoroute::statusmask_t b) { a ^= b; return a; }

extern const Glib::ustring& to_str(CFMUAutoroute::opttarget_t t);
extern const Glib::ustring& to_str(CFMUAutoroute::log_t l);

#endif /* CFMUAUTOROUTE_H */
