#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <iomanip>
#include <gdkmm.h>
#include <gtkmm.h>

#include "cfmuvalidateserver.hh"
#include "fplan.h"
#include "icaofpl.h"

CFMUValidateServer::CFMUValidateServer(int& argc, char**& argv)
	: m_engine(0), m_portal_webview(0), m_validate_webview(0),
	  m_portal_webview_close_sighandler(0), m_portal_webview_create_sighandler(0),
	  m_portal_webview_ready_sighandler(0), m_portal_webview_loadstatus_sighandler(0),
	  m_portal_webview_navpolicy_sighandler(0), m_portal_webview_newwndpolicy_sighandler(0),
	  m_portal_webview_rsrcload_sighandler(0), m_portal_webview_consolemsg_sighandler(0),
	  m_validate_webview_close_sighandler(0), m_validate_webview_create_sighandler(0),
	  m_validate_webview_ready_sighandler(0), m_validate_webview_loadstatus_sighandler(0),
	  m_validate_webview_navpolicy_sighandler(0), m_validate_webview_newwndpolicy_sighandler(0),
	  m_validate_webview_rsrcload_sighandler(0), m_validate_webview_consolemsg_sighandler(0),
	  m_verbose(0), m_wvstate(wvstate_failed), m_quit(false), m_local(false), m_tfrloaded(false)
{
        bool version(false);
	Glib::ustring fpl;
	int verbose(0);
	// option group
	{
		bool dis_aux(false);
		Glib::OptionGroup options("cfmuvalidateserver", "CFMU Flight Plan Validation Server");
		// version
		Glib::OptionEntry entry_version;
		entry_version.set_long_name(_("version"));
		entry_version.set_description(_("Show version and exit."));
		options.add_entry(entry_version, version);
		// fpl
                Glib::OptionEntry entry_fpl;
                entry_fpl.set_long_name(_("fpl"));
                entry_fpl.set_description(_("Query flight plan given as argument and return."));
                options.add_entry(entry_fpl, fpl);
 		// verbose      
		Glib::OptionEntry entry_verbose;
		entry_verbose.set_long_name(_("verbose"));
		entry_verbose.set_description(_("Set Verbosity."));
		options.add_entry(entry_verbose, verbose);
		// maindir
                Glib::OptionEntry entry_maindir;
                entry_maindir.set_short_name('m');
                entry_maindir.set_long_name("maindir");
                entry_maindir.set_description("Directory containing the main database");
                entry_maindir.set_arg_description("DIR");
                options.add_entry_filename(entry_maindir, m_dir_main);
		// auxdir
                Glib::OptionEntry entry_auxdir;
                entry_auxdir.set_short_name('a');
                entry_auxdir.set_long_name("auxdir");
                entry_auxdir.set_description("Directory containing the auxilliary database");
                entry_auxdir.set_arg_description("DIR");
                options.add_entry_filename(entry_auxdir, m_dir_aux);
		// disableaux
                Glib::OptionEntry entry_disableaux;
                entry_disableaux.set_short_name('A');
                entry_disableaux.set_long_name("disableaux");
                entry_disableaux.set_description("Disable the auxilliary database");
                entry_disableaux.set_arg_description("BOOL");
                options.add_entry(entry_disableaux, dis_aux);
		// local
                Glib::OptionEntry entry_local;
                entry_local.set_short_name('l');
                entry_local.set_long_name("local");
                entry_local.set_description("Local Checks (default Remote Checks)");
                entry_local.set_arg_description("BOOL");
                options.add_entry(entry_local, m_local);

		// option context
		Glib::OptionContext context(_("-- CFMU Flight Plan Validation Server"));
		//context.add_group(options);
		context.set_main_group(options);
		context.set_help_enabled(true);
		context.set_ignore_unknown_options(true);
		Glib::set_prgname(Glib::path_get_basename(*argv));
		if (!context.parse(argc, argv)) {
			std::cerr << "Command Line Option Error: Usage: " << Glib::get_prgname() << std::endl
				  << context.get_help(false);
			exit(1);
		}
		m_auxdbmode = Engine::auxdb_prefs;
		if (dis_aux)
                        m_auxdbmode = Engine::auxdb_none;
                else if (!m_dir_aux.empty())
                        m_auxdbmode = Engine::auxdb_override;
	}
        if (version) {
                std::cout << Glib::get_prgname() << (" Version " VERSION) << std::endl;
		exit(0);
	}
	m_verbose = std::max(verbose, 0);
        m_mainloop = Glib::MainLoop::create(Glib::MainContext::get_default());	
        m_inputchan = Glib::IOChannel::create_from_fd(0);
        m_outputchan = Glib::IOChannel::create_from_fd(1);
        m_inputchan->set_encoding();
        m_outputchan->set_encoding();
 	if (!fpl.empty()) {
		m_flightplans.push_back(fpl);
		m_quit = true;
	} else {
		m_inputreaderconn = Glib::signal_io().connect(sigc::mem_fun(this, &CFMUValidateServer::input_line_handler),
							      m_inputchan, Glib::IO_IN | Glib::IO_ERR | Glib::IO_HUP);
	}
	prepare_webview();
	if (!m_flightplans.empty())
		handle_input();
}

CFMUValidateServer::~CFMUValidateServer()
{
	remove_portal_sighandlers();
	remove_validate_sighandlers();
	if (m_portal_webview)
		g_object_unref(m_portal_webview);
	if (m_validate_webview)
		g_object_unref(m_validate_webview);
}

void CFMUValidateServer::run(void)
{
        m_mainloop->run();
}

bool CFMUValidateServer::input_line_handler(Glib::IOCondition iocond)
{
        Glib::ustring line;
        Glib::IOStatus iostat(m_inputchan->read_line(line));
        if (iostat == Glib::IO_STATUS_ERROR ||
            iostat == Glib::IO_STATUS_EOF) {
		m_quit = true;
		switch (m_wvstate) {
		case wvstate_loadmainpage:
		case wvstate_loadquerypage:
		case wvstate_ready:
			if (m_flightplans.empty())
				m_mainloop->quit();
			break;

		default:
			break;
		}
		m_inputreaderconn.disconnect();
		return true;
        }
        while (!line.empty()) {
                char ch(line[line.size() - 1]);
                if (ch != '\r' && ch != '\n')
                        break;
                line.erase(line.size() - 1, 1);
        }
	if (line.empty())
		return true;
	if (m_verbose > 1U)
		std::cerr << "Queue Flight Plan: " << line << std::endl;
	m_flightplans.push_back(line);
	handle_input();
	return true;
}

void CFMUValidateServer::handle_input(void)
{
	for (;;) {
		if (!m_local && m_wvstate != wvstate_ready)
			return;
		if (m_flightplans.empty()) {
			if (m_quit)
				m_mainloop->quit();
			return;
		}
		// check for pseudo instructions
		const std::string& fplan(m_flightplans.front());
		if (!fplan.compare(0, 6, "trace?")) {
			m_flightplans.pop_front();
			print_tracerules();
			continue;
		}
		if (!fplan.compare(0, 6, "trace:") || !fplan.compare(0, 6, "trace+")) {
			Glib::ustring txt1(fplan);
			Glib::ustring::size_type n(txt1.find('\n'));
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			n = txt1.find(' ');
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			m_tfr.tracerules_add(txt1.substr(6));
			m_flightplans.pop_front();
			print_tracerules();
			continue;
		}
		if (!fplan.compare(0, 6, "trace-")) {
			Glib::ustring txt1(fplan);
			Glib::ustring::size_type n(txt1.find('\n'));
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			n = txt1.find(' ');
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			m_tfr.tracerules_erase(txt1.substr(6));
			m_flightplans.pop_front();
			print_tracerules();
			continue;
		}
		if (!fplan.compare(0, 6, "trace*")) {
			m_tfr.tracerules_clear();
			m_flightplans.pop_front();
			print_tracerules();
			continue;
		}
		if (!fplan.compare(0, 9, "disabled?")) {
			m_flightplans.pop_front();
			print_disabledrules();
			continue;
		}
		if (!fplan.compare(0, 9, "disabled:") || !fplan.compare(0, 9, "disabled+")) {
			Glib::ustring txt1(fplan);
			Glib::ustring::size_type n(txt1.find('\n'));
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			n = txt1.find(' ');
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			m_tfr.disabledrules_add(txt1.substr(9));
			m_flightplans.pop_front();
			print_disabledrules();
			continue;
		}
		if (!fplan.compare(0, 9, "disabled-")) {
			Glib::ustring txt1(fplan);
			Glib::ustring::size_type n(txt1.find('\n'));
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			n = txt1.find(' ');
			if (n != Glib::ustring::npos)
				txt1.erase(n);
			m_tfr.disabledrules_erase(txt1.substr(9));
			m_flightplans.pop_front();
			print_disabledrules();
			continue;
		}
		if (!fplan.compare(0, 9, "disabled*")) {
			m_tfr.disabledrules_clear();
			m_flightplans.pop_front();
			print_disabledrules();
			continue;
		}
		if (!fplan.compare(0, 14, "validate:local")) {
			m_local = true;
			m_flightplans.pop_front();
			continue;
		}
		if (!fplan.compare(0, 15, "validate:remote")) {
			m_local = false;
			m_flightplans.pop_front();
			continue;
		}
		// process remote flight plan
		if (!m_local) {
			if (!m_validate_webview) {
				m_flightplans.pop_front();
				continue;
			}
			WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
			WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
			if (!document) {
				m_flightplans.pop_front();
				continue;
			}
			WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD"));
			if (!element || !WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element)) {
				m_flightplans.pop_front();
				continue;
			}
			webkit_dom_html_text_area_element_set_value(WEBKIT_DOM_HTML_TEXT_AREA_ELEMENT(element), fplan.c_str());
			m_flightplans.pop_front();
			element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL");
			if (!element)
				continue;
			zap_errors();
			webkit_web_view_execute_script(webview, "document.getElementById(\"FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL\").click();");
			set_wvstate(wvstate_query);
			continue;
		}
		// process local flight plan
		if (!m_tfrloaded) {
			// first open the databases
			delete m_engine;
			m_engine = new Engine(m_dir_main, m_auxdbmode, m_dir_aux, false);
			std::string dir_aux1 = m_engine->get_aux_dir(m_auxdbmode, m_dir_aux);
			m_airportdb.open(m_dir_main);
			if (!dir_aux1.empty() && m_airportdb.is_dbfile_exists(dir_aux1))
				m_airportdb.attach(dir_aux1);
			if (m_airportdb.is_aux_attached())
				std::cerr << "Auxillary airports database attached" << std::endl;
			m_navaiddb.open(m_dir_main);
			if (!dir_aux1.empty() && m_navaiddb.is_dbfile_exists(dir_aux1))
				m_navaiddb.attach(dir_aux1);
			if (m_navaiddb.is_aux_attached())
				std::cerr << "Auxillary navaids database attached" << std::endl;
			m_waypointdb.open(m_dir_main);
			if (!dir_aux1.empty() && m_waypointdb.is_dbfile_exists(dir_aux1))
				m_waypointdb.attach(dir_aux1);
			if (m_waypointdb.is_aux_attached())
				std::cerr << "Auxillary waypoints database attached" << std::endl;
			m_airwaydb.open(m_dir_main);
			if (!dir_aux1.empty() && m_airwaydb.is_dbfile_exists(dir_aux1))
				m_airwaydb.attach(dir_aux1);
			if (m_airwaydb.is_aux_attached())
				std::cerr << "Auxillary airways database attached" << std::endl;
			m_airspacedb.open(m_dir_main);
			if (!dir_aux1.empty() && m_airspacedb.is_dbfile_exists(dir_aux1))
				m_airspacedb.attach(dir_aux1);
			if (m_airspacedb.is_aux_attached())
				std::cerr << "Auxillary airspaces database attached" << std::endl;
			Glib::ustring tfrfilename(Glib::build_filename(dir_aux1, "eRAD.xml"));
			m_outputchan->write("Loading TFR rules from " + tfrfilename + ", please wait...\n");
			std::vector<TrafficFlowRestrictions::Message> msg;
			bool ld(m_tfr.load_rules(msg, tfrfilename, m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb));
			for (unsigned int i = 0; i < msg.size(); ++i)
				print(msg[i]);
			if (!ld) {
				m_flightplans.pop_front();
				m_outputchan->write("Error loading TFR file " + tfrfilename + "\n");
				m_outputchan->flush();
				continue;
			}
			m_outputchan->write("Origin:    " + m_tfr.get_origin() + "\n");
			m_outputchan->write("Created:   " + m_tfr.get_created() + "\n");
			m_outputchan->write("Effective: " + m_tfr.get_effective() + "\n");
			m_tfr.reset_rules();
			if (true) {
				Glib::TimeVal tv;
				tv.assign_current_time();
				m_tfr.simplify_rules_typeofflight('G');
				m_tfr.simplify_rules_mil(false);
				m_tfr.simplify_rules_gat(true);
				{
					Glib::TimeVal tv1;
					tv1.assign_current_time();
					tv = tv1 - tv;
				}
				std::ostringstream oss;
				oss << m_tfr.count_rules() << " rules after fixing type of flight G CIV GAT, "
				    << std::fixed << std::setprecision(3) << tv.as_double() << 's';
				m_outputchan->write(oss.str() + "\n");
			}
			m_tfrloaded = true;
		}
		// parse flight plan
		FPlanRoute fpl(*(FPlan *)0);
		char type_of_flight('G');
		std::string acft_type;
		std::string equipment;
		{
			IcaoFlightPlan icaofpl(*m_engine);
			std::string err(icaofpl.parse(fplan, false));
			m_flightplans.pop_front();
			if (!err.empty()) {
				m_outputchan->write("EFPL: " + err + "\n");
				m_outputchan->flush();
				continue;
			}
			icaofpl.set_route(fpl);
			type_of_flight = icaofpl.get_flighttype();
			acft_type = icaofpl.get_aircrafttype();
			equipment = icaofpl.get_equipment();
			if (true) {
				std::ostringstream oss;
				oss << "I: FPL Parsed, " << fpl.get_nrwpt() << " waypoints";
				m_outputchan->write(oss.str() + "\n");
			}
		}
		// print Parsed Flight Plan
		if (false) {
			for (unsigned int i = 0; i < fpl.get_nrwpt(); ++i) {
				const FPlanWaypoint& wpt(fpl[i]);
				std::ostringstream oss;
				oss << "PFPL[" << i << "]: " << wpt.get_icao() << '/' << wpt.get_name() << ' ';
				switch (wpt.get_pathcode()) {
				case FPlanWaypoint::pathcode_sid:
				case FPlanWaypoint::pathcode_star:
				case FPlanWaypoint::pathcode_airway:
					oss << wpt.get_pathname();
					break;

				case FPlanWaypoint::pathcode_directto:
					oss << "DCT";
					break;

				default:
					oss << '-';
					break;
				}
				oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
				    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
				    << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
			        m_outputchan->write(oss.str() + "\n");
			}
		}
		TrafficFlowRestrictions::Result res(m_tfr.check_fplan(fpl, type_of_flight, acft_type, equipment,
								      m_airportdb, m_navaiddb, m_waypointdb, m_airwaydb, m_airspacedb));
		// print Flight Plan
		{
			{
				std::ostringstream oss;
				oss << "FR: " << res.get_aircrafttype() << ' ' << res.get_aircrafttypeclass() << ' '
				    << res.get_equipment() << ' ' << res.get_typeofflight();
				m_outputchan->write(oss.str() + "\n");
			}
			const TrafficFlowRestrictions::Fpl& fpl(res.get_fplan());
			for (unsigned int i = 0; i < fpl.size(); ++i) {
				const TrafficFlowRestrictions::FplWpt& wpt(fpl[i]);
				std::ostringstream oss;
				oss << "F[" << i << '/' << wpt.get_wptnr() << "]: " << wpt.get_ident() << ' ' << to_str(wpt.get_type());
				if (wpt.get_type() == TrafficFlowRestrictions::Vertex::type_navaid)
					oss << '/' << NavaidsDb::Navaid::get_navaid_typename(wpt.get_navaid_type());
				oss << ' ';
				switch (wpt.get_pathcode()) {
				case FPlanWaypoint::pathcode_sid:
				case FPlanWaypoint::pathcode_star:
				case FPlanWaypoint::pathcode_airway:
					oss << wpt.get_pathname();
					break;

				case FPlanWaypoint::pathcode_directto:
					oss << "DCT";
					break;

				default:
					oss << '-';
					break;
				}
				oss << ' ' << wpt.get_coord().get_lat_str2() << ' ' << wpt.get_coord().get_lon_str2()
				    << ' ' << (wpt.is_standard() ? 'F' : 'A') << std::setw(3) << std::setfill('0') << ((wpt.get_altitude() + 50) / 100)
				    << ' ' << wpt.get_icao() << '/' << wpt.get_name() << ' ' << (wpt.is_ifr() ? 'I' : 'V') << "FR";
				m_outputchan->write(oss.str() + "\n");
			}
		}
		print(res);
		m_outputchan->write("\n");
		m_outputchan->flush();
	}
}

void CFMUValidateServer::print(const TrafficFlowRestrictions::Message& m)
{
	std::ostringstream oss;
	oss << m.get_type_char() << ':';
	if (!m.get_rule().empty())
		oss << " R:" << m.get_rule();
	if (!m.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(m.get_vertexset().begin()), e(m.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!m.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::Message::set_t::const_iterator i(m.get_edgeset().begin()), e(m.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	oss << ' ' << m.get_text();
	m_outputchan->write(oss.str() + "\n");
}

void CFMUValidateServer::print(const TrafficFlowRestrictions::Result& res)
{
	// print Messages
	for (unsigned int i = 0; i < res.messages_size(); ++i)
		print(res.get_message(i));
	m_outputchan->write(res ? "OK\n" : "FAIL\n");
	for (unsigned int i = 0; i < res.size(); ++i)
		print(res[i]);
}

void CFMUValidateServer::print(const TrafficFlowRestrictions::RuleResult& rule)
{
	std::ostringstream oss;
	oss << "Rule: " << rule.get_rule() << ' ';
	switch (rule.get_codetype()) {
	case TrafficFlowRestrictions::RuleResult::codetype_mandatory:
		oss << "Mandatory";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_forbidden:
		oss << "Forbidden";
		break;

	case TrafficFlowRestrictions::RuleResult::codetype_closed:
		oss << "ClosedForCruising";
		break;

	default:
		oss << '?';
		break;
	}
	if (rule.is_dct())
		oss << " (DCT)";
	if (!rule.get_vertexset().empty()) {
		oss << " V:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_vertexset().begin()), e(rule.get_vertexset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!rule.get_edgeset().empty()) {
		oss << " E:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_edgeset().begin()), e(rule.get_edgeset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	if (!rule.get_reflocset().empty()) {
		oss << " L:";
		bool subseq(false);
		for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(rule.get_reflocset().begin()), e(rule.get_reflocset().end()); i != e; ++i) {
			if (subseq)
				oss << ',';
			subseq = true;
			oss << *i;
		}
	}
	m_outputchan->write(oss.str() + "\n");
	m_outputchan->write("Desc: " + rule.get_desc() + "\n");
	m_outputchan->write("OprGoal: " + rule.get_oprgoal() + "\n");
	m_outputchan->write("Cond: " + rule.get_cond() + "\n");
	for (unsigned int i = 0; i < rule.size(); ++i) {
		const TrafficFlowRestrictions::RuleResult::Alternative& alt(rule[i]);
		{
			std::ostringstream oss;
			oss << "  Alternative " << i;
			if (!alt.get_vertexset().empty()) {
				oss << " V:";
				bool subseq(false);
				for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_vertexset().begin()), e(alt.get_vertexset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			if (!alt.get_edgeset().empty()) {
				oss << " E:";
				bool subseq(false);
				for (TrafficFlowRestrictions::RuleResult::set_t::const_iterator i(alt.get_edgeset().begin()), e(alt.get_edgeset().end()); i != e; ++i) {
					if (subseq)
						oss << ',';
					subseq = true;
					oss << *i;
				}
			}
			m_outputchan->write(oss.str() + "\n");
		}
		for (unsigned int j = 0; j < alt.size(); ++j) {
			const TrafficFlowRestrictions::RuleResult::Alternative::Sequence& el(alt[j]);
			std::ostringstream oss;
			oss << "    " << el.get_type_string();
			switch (el.get_type()) {
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway:
				oss << ' ' << el.get_ident();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_sid:
				oss << ' ' << el.get_ident() << ' ' << el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_star:
				oss << ' ' << el.get_ident() << ' ' << el.get_airport();
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airspace:
				oss << ' ' << el.get_ident() << '/'
				    << AirspacesDb::Airspace::get_class_string(el.get_bdryclass(), AirspacesDb::Airspace::typecode_ead);
				break;

			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct:
			case TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point:
			default:
				break;
			}
			oss << " F" << std::setw(3) << std::setfill('0') << ((el.get_lower_alt() + 50) / 100) << "...F"
			    << std::setw(3) << std::setfill('0') << ((std::min((int32_t)el.get_upper_alt(), (int32_t)99900) + 50) / 100);
			m_outputchan->write(oss.str() + "\n");
			if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_point) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point0());
				std::ostringstream oss;
				oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
				    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
				m_outputchan->write(oss.str() + "\n");
			}
			if (el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_airway ||
			    el.get_type() == TrafficFlowRestrictions::RuleResult::Alternative::Sequence::type_dct) {
				const TrafficFlowRestrictions::RuleResult::Alternative::Sequence::RulePoint& pt(el.get_point1());
				std::ostringstream oss;
				oss << "      " << pt.get_ident() << ' ' << pt.get_coord().get_lat_str2() << ' '
				    << pt.get_coord().get_lon_str2() << ' ' << to_str(pt.get_type());
				m_outputchan->write(oss.str() + "\n");
			}
		}
	}
}

void CFMUValidateServer::print_tracerules(void)
{
	std::string r;
	bool subseq(false);
	for (TrafficFlowRestrictions::tracerules_const_iterator_t ti(m_tfr.tracerules_begin()), te(m_tfr.tracerules_end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	m_outputchan->write("trace:" + r + "\n\n");
	m_outputchan->flush();
}

void CFMUValidateServer::print_disabledrules(void)
{
	std::string r;
	bool subseq(false);
	for (TrafficFlowRestrictions::disabledrules_const_iterator_t ti(m_tfr.disabledrules_begin()), te(m_tfr.disabledrules_end()); ti != te; ++ti) {
		if (subseq)
			r += ",";
		subseq = true;
		r += *ti;
	}
	m_outputchan->write("disabled:" + r + "\n\n");
	m_outputchan->flush();
}

void CFMUValidateServer::add_portal_sighandlers(void)
{
	remove_portal_sighandlers();
	m_portal_webview_close_sighandler = g_signal_connect(m_portal_webview, "close-web-view", G_CALLBACK(&CFMUValidateServer::close_webview_1), this);
	m_portal_webview_create_sighandler = g_signal_connect(m_portal_webview, "create-web-view", G_CALLBACK(&CFMUValidateServer::create_webview_1), this);
	m_portal_webview_ready_sighandler = g_signal_connect(m_portal_webview, "web-view-ready", G_CALLBACK(&CFMUValidateServer::webview_ready_1), this);
	m_portal_webview_loadstatus_sighandler = g_signal_connect(m_portal_webview, "notify::load-status", G_CALLBACK(&CFMUValidateServer::loadstatus_webview_1), this);
	m_portal_webview_navpolicy_sighandler = g_signal_connect(m_portal_webview, "navigation-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_portal_webview_newwndpolicy_sighandler = g_signal_connect(m_portal_webview, "new-window-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_portal_webview_rsrcload_sighandler = g_signal_connect(m_portal_webview, "resource-load-finished", G_CALLBACK(&CFMUValidateServer::resourceload_finished_1), this);
	m_portal_webview_consolemsg_sighandler = g_signal_connect(m_portal_webview, "console-message", G_CALLBACK(&CFMUValidateServer::console_message_1), this);
}

void CFMUValidateServer::remove_portal_sighandlers(void)
{
        if (m_portal_webview_close_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_close_sighandler);
		m_portal_webview_close_sighandler = 0;
        }
        if (m_portal_webview_create_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_create_sighandler);
		m_portal_webview_create_sighandler = 0;
        }
	if (m_portal_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_portal_webview, m_portal_webview_ready_sighandler);
		m_portal_webview_ready_sighandler = 0;
	}
        if (m_portal_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_loadstatus_sighandler);
		m_portal_webview_loadstatus_sighandler = 0;
        }
        if (m_portal_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_navpolicy_sighandler);
		m_portal_webview_navpolicy_sighandler = 0;
        }
        if (m_portal_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_newwndpolicy_sighandler);
		m_portal_webview_newwndpolicy_sighandler = 0;
        }
        if (m_portal_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_rsrcload_sighandler);
		m_portal_webview_rsrcload_sighandler = 0;
        }
        if (m_portal_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_portal_webview, m_portal_webview_consolemsg_sighandler);
		m_portal_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidateServer::add_validate_sighandlers(void)
{
	remove_validate_sighandlers();
	m_validate_webview_close_sighandler = g_signal_connect(m_validate_webview, "close-web-view", G_CALLBACK(&CFMUValidateServer::close_webview_1), this);
	m_validate_webview_create_sighandler = g_signal_connect(m_validate_webview, "create-web-view", G_CALLBACK(&CFMUValidateServer::create_webview_1), this);
	m_validate_webview_ready_sighandler = g_signal_connect(m_validate_webview, "web-view-ready", G_CALLBACK(&CFMUValidateServer::webview_ready_1), this);
	m_validate_webview_loadstatus_sighandler = g_signal_connect(m_validate_webview, "notify::load-status", G_CALLBACK(&CFMUValidateServer::loadstatus_webview_1), this);
	m_validate_webview_navpolicy_sighandler = g_signal_connect(m_validate_webview, "navigation-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_validate_webview_newwndpolicy_sighandler = g_signal_connect(m_validate_webview, "new-window-policy-decision-requested", G_CALLBACK(&CFMUValidateServer::policy_decision_required_1), this);
	m_validate_webview_rsrcload_sighandler = g_signal_connect(m_validate_webview, "resource-load-finished", G_CALLBACK(&CFMUValidateServer::resourceload_finished_1), this);
	m_validate_webview_consolemsg_sighandler = g_signal_connect(m_validate_webview, "console-message", G_CALLBACK(&CFMUValidateServer::console_message_1), this);
}

void CFMUValidateServer::remove_validate_sighandlers(void)
{
        if (m_validate_webview_close_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_close_sighandler);
		m_validate_webview_close_sighandler = 0;
        }
        if (m_validate_webview_create_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_create_sighandler);
		m_validate_webview_create_sighandler = 0;
        }
	if (m_validate_webview_ready_sighandler) {
		g_signal_handler_disconnect(m_validate_webview, m_validate_webview_ready_sighandler);
		m_validate_webview_ready_sighandler = 0;
	}
        if (m_validate_webview_loadstatus_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_loadstatus_sighandler);
		m_validate_webview_loadstatus_sighandler = 0;
        }
        if (m_validate_webview_navpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_navpolicy_sighandler);
		m_validate_webview_navpolicy_sighandler = 0;
        }
        if (m_validate_webview_newwndpolicy_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_newwndpolicy_sighandler);
		m_validate_webview_newwndpolicy_sighandler = 0;
        }
        if (m_validate_webview_rsrcload_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_rsrcload_sighandler);
		m_validate_webview_rsrcload_sighandler = 0;
        }
        if (m_validate_webview_consolemsg_sighandler) {
                g_signal_handler_disconnect(m_validate_webview, m_validate_webview_consolemsg_sighandler);
		m_validate_webview_consolemsg_sighandler = 0;
        }
}

void CFMUValidateServer::prepare_webview(void)
{
	remove_portal_sighandlers();
	if (m_portal_webview)
		g_object_unref(m_portal_webview);
	// add cookie jar
	std::string cookiefile(Glib::build_filename(FPlan::get_userdbpath(), "cookies.txt"));
	SoupSession *session(webkit_get_default_session());
	SoupCookieJar *cookiejar(soup_cookie_jar_text_new(cookiefile.c_str(), false));
	soup_session_add_feature(session, SOUP_SESSION_FEATURE(cookiejar));
	SoupCookie *bwarn(soup_cookie_new("eurocontrol_nm_portal_hideBroswerWarning", "hide", "www.public.cfmu.eurocontrol.int", "/", -1));
	soup_cookie_jar_add_cookie(cookiejar, bwarn);
	bwarn = soup_cookie_new("eurocontrol_nm_portal_hideCookieWarning", "hide", "www.public.cfmu.eurocontrol.int", "/", -1);
	soup_cookie_jar_add_cookie(cookiejar, bwarn);
	g_object_unref(cookiejar);
	// open webview
	m_portal_webview = GTK_WIDGET(g_object_ref_sink(webkit_web_view_new()));
	gtk_widget_show(m_portal_webview);
	add_portal_sighandlers();
	WebKitWebView *webview = WEBKIT_WEB_VIEW(m_portal_webview);
	webkit_web_view_load_uri(webview, "https://www.public.cfmu.eurocontrol.int/PUBPORTAL/gateway/spec/index.html");
	set_wvstate(wvstate_loadmainpage);
}

void CFMUValidateServer::set_wvstate(wvstate_t newstate)
{
	m_delayedstatechangeconn.disconnect();
	bool chg(m_wvstate != newstate);
	m_wvstate = newstate;
	switch (m_wvstate) {
	case wvstate_ready:
		handle_input();
		break;

	case wvstate_failed:
		if (chg) {
			m_outputchan->write("FAIL: WEBPAGE\n");
			m_outputchan->flush();
			break;
		}
		m_mainloop->quit();
		break;

	default:
		break;
	}
}

void CFMUValidateServer::parse_results(void)
{
	if (!m_validate_webview)
		return;
	typedef std::pair<guint,bool> lvlvisible_t;
	typedef std::pair<WebKitDOMNode *,lvlvisible_t> nodelvlvisible_t;
	std::stack<nodelvlvisible_t> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(std::make_pair(WEBKIT_DOM_NODE(element), std::make_pair(0U, true)));
	}
	typedef std::vector<Glib::ustring> results_t;
	results_t results;
	while (!subtree.empty()) {
		WebKitDOMNode *node;
		guint level;
		bool visible;
		{
			nodelvlvisible_t nl(subtree.top());
			node = nl.first;
			level = nl.second.first;
			visible = nl.second.second;
		}
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					visible = !!strcmp(disp, "none");
					g_free(disp);
				}
			}
		}
		if (visible) {
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(std::make_pair(ptr, std::make_pair(level + 1U, visible)));
			}
		}
		if (!WEBKIT_DOM_IS_HTML_ELEMENT(node))
			continue;
		char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
		if (m_verbose > 2U) {
			std::cerr << std::string(2U * level, ' ') << "element:";
			gchar *lname(webkit_dom_node_get_local_name(WEBKIT_DOM_NODE(node)));
			if (lname) {
				if (*lname)
					std::cerr << " name \"" << lname << "\"";
				g_free(lname);
			}
			lname = webkit_dom_html_element_get_id(WEBKIT_DOM_HTML_ELEMENT(node));
			if (lname) {
				if (*lname)
					std::cerr << " id \"" << lname << "\"";
				g_free(lname);
			}
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(node)));
			if (typ && *typ)
				std::cerr << " type \"" << typ << "\"";
			if (class_name && *class_name)
				std::cerr << " class \"" << class_name << "\"";
			WebKitDOMCSSStyleDeclaration *style(webkit_dom_element_get_style(WEBKIT_DOM_ELEMENT(node)));
			if (style) {
				gchar *disp(webkit_dom_css_style_declaration_get_property_value(style, "display"));
				if (disp) {
					if (*disp)
						std::cerr << " display \"" << disp << "\"" << std::endl;
					g_free(disp);
				}
			}
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt)
					std::cerr << " text \"" << txt << "\"";
				g_free(txt);
			}
			std::cerr << std::endl;
		}
		if (!class_name)
			continue;
		if (!visible) {
			g_free(class_name);
			continue;
		}
		if (!strcmp(class_name, "portal_dataValue") && WEBKIT_DOM_IS_HTML_DIV_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (false && !strcmp(class_name, "portal_decotableCell ") && WEBKIT_DOM_IS_HTML_TABLE_CELL_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		if (true && (!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
		    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
			gchar *txt(webkit_dom_html_element_get_inner_text(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (txt) {
				if (*txt) {
					results.push_back(Glib::ustring(txt));
					if (m_verbose > 2U)
						std::cerr << "*** " << txt << std::endl;
				}
				g_free(txt);
			}
		}
		g_free(class_name);
	}
	// locally look up failing rules and get a description
	for (results_t::const_iterator ri(results.begin()), re(results.end()); ri != re; ++ri) {
		m_outputchan->write(*ri + "\n");
		if (!m_tfrloaded)
			continue;
		Glib::ustring::size_type n(ri->rfind('['));
		if (n == Glib::ustring::npos)
			continue;
		Glib::ustring::size_type ne(ri->find(']', n + 1));
		if (ne == Glib::ustring::npos || ne - n < 2 || ne - n > 11)
			continue;
		std::vector<TrafficFlowRestrictions::RuleResult> rr(m_tfr.find_rules(ri->substr(n + 1, ne - n - 1), true));
		if (rr.empty())
			continue;
		print(rr[0]);
	}
	if (!results.empty()) {
		m_outputchan->write("\n");
		m_outputchan->flush();
		set_wvstate(wvstate_ready);
	}
}

void CFMUValidateServer::zap_errors(void)
{
	if (!m_validate_webview)
		return;
	std::stack<WebKitDOMNode *> subtree;
	{
		WebKitWebView *webview(WEBKIT_WEB_VIEW(m_validate_webview));
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA"));
		if (!element || !WEBKIT_DOM_IS_NODE(element))
			return;
		subtree.push(WEBKIT_DOM_NODE(element));
	}
	std::ostringstream results;
	while (!subtree.empty()) {
		WebKitDOMNode *node(subtree.top());
		subtree.pop();
		if (WEBKIT_DOM_IS_HTML_ELEMENT(node)) {
			char *class_name(webkit_dom_html_element_get_class_name(WEBKIT_DOM_HTML_ELEMENT(node)));
			if (class_name) {
		     		if ((!strcmp(class_name, "portal_rowOdd") || !strcmp(class_name, "portal_rowEven")) &&
				    WEBKIT_DOM_IS_HTML_TABLE_ROW_ELEMENT(node)) {
					g_free(class_name);
					while (webkit_dom_node_has_child_nodes(node)) {
						GError *error(0);
						webkit_dom_node_remove_child(node, webkit_dom_node_get_first_child(node), &error);
						if (error) {
							std::cerr << "Cannot remove child: " << error->message << std::endl;
							g_error_free(error);
						}
					}
					continue;
				}
				g_free(class_name);
			}
		}
		{
			WebKitDOMNodeList *list(webkit_dom_node_get_child_nodes(node));
			gulong i(webkit_dom_node_list_get_length(list));
			while (i) {
				--i;
				WebKitDOMNode *ptr(webkit_dom_node_list_item(list, i));
				if (!ptr)
					continue;
				subtree.push(ptr);
			}
		}
	}
}

gboolean CFMUValidateServer::close_webview_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->close_webview(webview);
}

gboolean CFMUValidateServer::close_webview(WebKitWebView* webview)
{
        set_wvstate(wvstate_failed);
	return true;
}

WebKitWebView *CFMUValidateServer::create_webview_1(WebKitWebView *webview, WebKitWebFrame *frame, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->create_webview(webview, frame);
}

WebKitWebView *CFMUValidateServer::create_webview(WebKitWebView *webview, WebKitWebFrame *frame)
{
	if (m_verbose > 1U)
		std::cerr << "new webview!" << std::endl;
	if (webview == WEBKIT_WEB_VIEW(m_validate_webview))
		return 0;
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview));
	if (m_wvstate != wvstate_loadmainpage)
		return 0;
	remove_validate_sighandlers();
	if (m_validate_webview)
		g_object_unref(m_validate_webview);
	m_validate_webview = GTK_WIDGET(g_object_ref_sink(webkit_web_view_new()));
	gtk_widget_show(m_validate_webview);
	webkit_web_view_set_settings(WEBKIT_WEB_VIEW(m_validate_webview), webkit_web_view_get_settings(WEBKIT_WEB_VIEW(m_portal_webview)));
	add_validate_sighandlers();
	set_wvstate(wvstate_loadquerypage);
	WebKitWebView *wv(WEBKIT_WEB_VIEW(m_validate_webview));
	return wv;
}

gboolean CFMUValidateServer::webview_ready_1(WebKitWebView* webview, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->webview_ready(webview);
}

gboolean CFMUValidateServer::webview_ready(WebKitWebView* webview)
{
	if (m_verbose > 1U)
		std::cerr << "WebView: ready" << std::endl;
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	gtk_widget_show_all(GTK_WIDGET(webview));
	return false;
}

void CFMUValidateServer::loadstatus_webview_1(WebKitWebView *webview, GParamSpec *pspec, gpointer userdata)
{
	static_cast<CFMUValidateServer *>(userdata)->loadstatus_webview(webview, pspec);
}

void CFMUValidateServer::loadstatus_webview(WebKitWebView *webview, GParamSpec *pspec)
{
        WebKitLoadStatus status(webkit_web_view_get_load_status(webview));
        GObject *object(G_OBJECT(webview));
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
        g_object_freeze_notify(object);
	switch (status) {
        case WEBKIT_LOAD_FINISHED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load finished" << std::endl;
		gtk_widget_show_all(GTK_WIDGET(webview));
		break;

        case WEBKIT_LOAD_PROVISIONAL:
		if (m_verbose > 1U)
			std::cerr << "WebView: load provisional" << std::endl;
                break;

        case WEBKIT_LOAD_COMMITTED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load committed" << std::endl;
                break;

        case WEBKIT_LOAD_FIRST_VISUALLY_NON_EMPTY_LAYOUT:
		if (m_verbose > 1U)
			std::cerr << "WebView: load first visually nonempty layout" << std::endl;
                break;

        case WEBKIT_LOAD_FAILED:
		if (m_verbose > 1U)
			std::cerr << "WebView: load failed" << std::endl;
		set_wvstate(wvstate_failed);
                break;

        default:
                break;
        }
        g_object_thaw_notify(object);
}

gboolean CFMUValidateServer::policy_decision_required_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							      WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->policy_decision_required(webview, webframe, request, action, decision);
}

gboolean CFMUValidateServer::policy_decision_required(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitNetworkRequest *request,
							    WebKitWebNavigationAction *action, WebKitWebPolicyDecision *decision)
{
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	WebKitWebNavigationReason reason(webkit_web_navigation_action_get_reason(action));
        gint button(webkit_web_navigation_action_get_button(action));
        gint state(webkit_web_navigation_action_get_modifier_state(action));
        const char *uri(webkit_network_request_get_uri(request));
	if (m_verbose > 1U)
		std::cerr << "policy_decision_required: reason " << reason << " button " << button << " state " << state
			  << " uri " << uri << std::endl;
	return false;
}

void CFMUValidateServer::resourceload_finished_1(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource, gpointer userdata)
{
	static_cast<CFMUValidateServer *>(userdata)->resourceload_finished(webview, webframe, webresource);
}

void CFMUValidateServer::resourceload_finished(WebKitWebView *webview, WebKitWebFrame *webframe, WebKitWebResource *webresource)
{
        g_assert(webview == WEBKIT_WEB_VIEW(m_portal_webview) || webview == WEBKIT_WEB_VIEW(m_validate_webview));
	if (m_verbose > 1U)
		std::cerr << "resource load finished! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_loadmainpage:
	{
		if (webview != WEBKIT_WEB_VIEW(m_portal_webview))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL"));
		if (!element)
			break;
		if (m_verbose > 2U)
			std::cerr << "Button found! Injecting script" << std::endl;
		webkit_web_view_execute_script(webview, "document.getElementById(\"IFPUV_LAUNCH_AREA.FREE_TEXT_EDIT_LINK_LABEL\").click();");
		set_wvstate(wvstate_loadquerypage);
		break;
	}

	case wvstate_loadquerypage:
	{
		if (webview != WEBKIT_WEB_VIEW(m_validate_webview))
			break;
		WebKitDOMDocument *document(webkit_web_view_get_dom_document(webview));
		WebKitDOMElement *element(webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.VALIDATE_ACTION_LABEL"));
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Validate element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.FLIGHT_DATA_AREA.GENERAL_DATA_ENTRY.INTRODUCE_FLIGHT_PLAN_FIELD");
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "FPL entry element type " << typ << std::endl;
		}
		if (!WEBKIT_DOM_IS_HTML_TEXT_AREA_ELEMENT(element))
			break;
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA.ERRORS_DATA_TABLE");
		if (!element)
			break;
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Result Errors element type " << typ << std::endl;
		}
		element = webkit_dom_document_get_element_by_id(document, "FREE_TEXT_EDITOR.VALIDATION_RESULTS_AREA");
		if (!element)
			break;
		webkit_dom_event_target_add_event_listener(WEBKIT_DOM_EVENT_TARGET(element), "DOMSubtreeModified", G_CALLBACK(&result_change_1), true, this);
		if (m_verbose > 1U) {
			const gchar *typ(g_type_name(G_TYPE_FROM_INSTANCE(element)));
			if (typ)
				std::cerr << "Results element type " << typ << std::endl;
		}
		m_delayedstatechangeconn.disconnect();
		m_delayedstatechangeconn = Glib::signal_timeout().connect(sigc::bind_return(sigc::bind(sigc::mem_fun(*this, &CFMUValidateServer::set_wvstate), wvstate_ready), false), 3000);
		break;
	}

	case wvstate_query:
	{
		if (webview != WEBKIT_WEB_VIEW(m_validate_webview))
			break;
		set_wvstate(wvstate_waitresult);
		parse_results();
		break;
	}

	default:
		break;
	}
}

gboolean CFMUValidateServer::result_change_1(WebKitDOMElement *target, WebKitDOMEvent *e, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->result_change(target, e);
}

gboolean CFMUValidateServer::result_change(WebKitDOMElement *target, WebKitDOMEvent *e)
{
	std::cerr << "result change! state " << (int)m_wvstate << std::endl;
	switch (m_wvstate) {
	case wvstate_waitresult:
		parse_results();
		break;

	default:
		break;
	}
	return true;
}

gboolean CFMUValidateServer::console_message_1(WebKitWebView *webview, gchar *message, gint line, gchar *source_id, gpointer userdata)
{
	return static_cast<CFMUValidateServer *>(userdata)->console_message(webview, message, line, source_id);
}

gboolean CFMUValidateServer::console_message(WebKitWebView *webview, gchar *message, gint line, gchar *source_id)
{
	if (m_verbose > 0U) {
		std::cerr << message << " (" << (source_id ? "*null*" : source_id) << ':' << line << ')' << std::endl;
	}
	return true;
}

int main(int argc, char *argv[])
{
        try {
		g_type_init();
                Glib::init();
		Gio::init();
                Glib::set_application_name("CFMU Validate Server");
                Glib::thread_init();
		{
			bool found(false);
			std::string disp(Glib::getenv("DISPLAY", found));
			if (found)
				gtk_init(&argc, &argv);
		}
                CFMUValidateServer app(argc, argv);
                app.run();
                return 0;
        } catch(const Glib::FileError& ex) {
                std::cerr << "FileError: " << ex.what() << std::endl;
                return 1;
        } catch(const Gtk::BuilderError& ex) {
                std::cerr << "BuilderError: " << ex.what() << std::endl;
                return 1;
        }
        return 1;
}
