//
// C++ Interface: aircraft
//
// Description: Aircraft Model
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2012, 2013
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef AIRCRAFT_H
#define AIRCRAFT_H

#include "sysdeps.h"

#include <limits>
#include <set>
#include <glibmm.h>
#include <libxml++/libxml++.h>

class FPlanRoute;
class Engine;

class Aircraft {
  protected:
	template <typename T> class Poly1D : public std::vector<T> {
	  public:
		Poly1D(void) : std::vector<T>() {}
		Poly1D(typename Poly1D<T>::size_type n) : std::vector<T>(n) {}
		Poly1D(typename Poly1D<T>::size_type n, const T& t) : std::vector<T>(n, t) {}
		Poly1D(const std::vector<T>& t) : std::vector<T>(t) {}
		template <class InputIterator> Poly1D(InputIterator a, InputIterator b) : std::vector<T>(a, b) {}

		T eval(T x) const;
		T evalderiv(T x) const;
		T inveval(T x, T tol, unsigned int maxiter, T t) const;
		T inveval(T x, T tol, unsigned int maxiter = 16) const;
		T inveval(T x) const;

		void reduce(void);

		Poly1D<T>& operator+=(T x);
		Poly1D<T>& operator-=(T x);
		Poly1D<T>& operator*=(T x);
		Poly1D<T>& operator/=(T x);
		Poly1D<T>& operator+=(const Poly1D<T>& x);
		Poly1D<T>& operator-=(const Poly1D<T>& x);
		Poly1D<T> operator*(const Poly1D<T>& x) const;

		inline Poly1D<T> operator+(T x) const { Poly1D<T> y(*this); y += x; return y; }
		inline Poly1D<T> operator-(T x) const { Poly1D<T> y(*this); y -= x; return y; }
		inline Poly1D<T> operator*(T x) const { Poly1D<T> y(*this); y *= x; return y; }
		inline Poly1D<T> operator/(T x) const { Poly1D<T> y(*this); y /= x; return y; }
		inline Poly1D<T> operator+(const Poly1D<T>& x) const { Poly1D<T> y(*this); y += x; return y; }
		inline Poly1D<T> operator-(const Poly1D<T>& x) const { Poly1D<T> y(*this); y -= x; return y; }
		inline Poly1D<T>& operator*=(const Poly1D<T>& x) { Poly1D<T> y((*this) * x); this->swap(y); return *this; }

		Poly1D<T> differentiate(void) const;
		Poly1D<T> integrate(void) const;

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

		std::ostream& save_octave(std::ostream& os, const Glib::ustring& name) const;
	};

  public:
	class WeightBalance {
	  public:
		class Element {
		  public:
			class Unit {
			  public:
				Unit(const Glib::ustring& name = "", double factor = 1, double offset = 0);

				const Glib::ustring& get_name(void) const { return m_name; }
				double get_factor(void) const { return m_factor; }
				double get_offset(void) const { return m_offset; }
				bool is_fixed(void) const { return get_factor() == 0; }

				void load_xml(const xmlpp::Element *el);
				void save_xml(xmlpp::Element *el) const;

			  protected:
				Glib::ustring m_name;
				double m_factor;
				double m_offset;
			};

			typedef enum {
				flag_none = 0,
				flag_fixed = (1 << 0),
				flag_binary = (1 << 1),
				flag_avgas = (1 << 2),
				flag_oil = (1 << 3),
				flag_gear = (1 << 4),
				flag_all = flag_fixed | flag_binary | flag_avgas | flag_oil | flag_gear
			} flags_t;

			Element(const Glib::ustring& name = "", double massmin = 0, double massmax = 0, double arm = 0, flags_t flags = flag_none);
			const Glib::ustring& get_name(void) const { return m_name; }
			double get_massmin(void) const { return m_massmin; }
			double get_massmax(void) const { return m_massmax; }
			double get_arm(void) const { return m_arm; }
			flags_t get_flags(void) const { return m_flags; }

			unsigned int get_nrunits(void) const { return m_units.size(); }
			const Unit& get_unit(unsigned int idx) const { return m_units[idx]; }

			void clear_units(void);
			void add_unit(const Glib::ustring& name = "", double factor = 1, double offset = 0);
			void add_auto_units(double kg_to_unit);

			void load_xml(const xmlpp::Element *el);
			void save_xml(xmlpp::Element *el) const;

		  protected:
			class FlagName {
			  public:
				const char *m_name;
				flags_t m_flag;
			};
			static const FlagName flagnames[];

			typedef std::vector<Unit> units_t;
			units_t m_units;
			Glib::ustring m_name;
			double m_massmin;
			double m_massmax;
			double m_arm;
			flags_t m_flags;
		};

		class ElementValue {
		  public:
			ElementValue(double value = 0, unsigned int unit = 0) : m_value(value), m_unit(unit) {}
			double get_value(void) const { return m_value; }
			unsigned int get_unit(void) const { return m_unit; }
			void set_value(double v) { m_value = v; }
			void set_unit(unsigned int u) { m_unit = u; }

		  protected:
			double m_value;
			unsigned int m_unit;
		};

		typedef std::vector<ElementValue> elementvalues_t;

		class Envelope {
		  public:
			class Point {
			  public:
				Point(double arm = 0, double mass = 0);
				double get_arm(void) const { return m_arm; }
				double get_mass(void) const { return m_mass; }

				static double area2(const Point& p0, const Point& p1, const Point& p2);
				double area2(const Point& p0, const Point& p1) const { return area2(p0, p1, *this); }

				void load_xml(const xmlpp::Element *el);
				void save_xml(xmlpp::Element *el) const;

			  protected:
				double m_arm;
				double m_mass;
			};

			Envelope(const Glib::ustring& name = "");

			const Glib::ustring& get_name(void) const { return m_name; }

			unsigned int size(void) const { return m_env.size(); }
			const Point& operator[](unsigned int idx) const { return m_env[idx]; }
			int windingnumber(const Point& pt) const;
			bool is_inside(const Point& pt) const { return !!windingnumber(pt); }
			void get_bounds(double& minarm, double& maxarm, double& minmass, double& maxmass) const;

			void load_xml(const xmlpp::Element *el);
			void save_xml(xmlpp::Element *el) const;

			void add_point(const Point& pt);
			void add_point(double arm = 0, double mass = 0);

		  public:
			typedef std::vector<Point> env_t;
			env_t m_env;
			Glib::ustring m_name;
		};

		WeightBalance(void);

		const Glib::ustring& get_armunitname(void) const { return m_armunitname; }
		const Glib::ustring& get_massunitname(void) const { return m_massunitname; }

		unsigned int get_nrelements(void) const { return m_elements.size(); }
		const Element& get_element(unsigned int idx) const { return m_elements[idx]; }

		unsigned int get_nrenvelopes(void) const { return m_envelopes.size(); }
		const Envelope& get_envelope(unsigned int idx) const { return m_envelopes[idx]; }

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

	  protected:
		typedef std::vector<Element> elements_t;
		elements_t m_elements;
		typedef std::vector<Envelope> envelopes_t;
		envelopes_t m_envelopes;
		Glib::ustring m_armunitname;
		Glib::ustring m_massunitname;
	};

	class Distances {
	  public:
		class Distance {
		  public:
			class Point {
			  public:
				Point(double da = 0, double pa = 0, double temp = 0, double gnddist = 0, double obstdist = 0);
				
				void load_xml(const xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
					      double tempoffset = 273.15, double distfactor = 1);
				void save_xml(xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
					      double tempoffset = 273.15, double distfactor = 1) const;

				double get_densityalt(void) const;
				double get_gnddist(void) const { return m_gnddist; }
				double get_obstdist(void) const { return m_obstdist; }

			  protected:
				double m_da;
				double m_pa;
				double m_temp;
				double m_gnddist;
				double m_obstdist;
			};

			Distance(const Glib::ustring& name = "", double vrotate = 0, double mass = 0);

			bool recalculatepoly(bool force = false);
			void add_point_da(double gnddist, double obstdist, double da);
			void add_point_pa(double gnddist, double obstdist, double pa, double temp = std::numeric_limits<double>::quiet_NaN());

			void load_xml(const xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
				      double tempoffset = 273.15, double distfactor = 1);
			void save_xml(xmlpp::Element *el, double altfactor = 1, double tempfactor = 1,
				      double tempoffset = 273.15, double distfactor = 1) const;

			const Glib::ustring& get_name(void) const { return m_name; }
			double get_vrotate(void) const { return m_vrotate; }
			double get_mass(void) const { return m_mass; }
			const Poly1D<double>& get_gndpoly(void) const { return m_gndpoly; }
			const Poly1D<double>& get_obstpoly(void) const { return m_obstpoly; }
			double calculate_gnddist(double da) const;
			double calculate_obstdist(double da) const;
			void calculate(double& gnddist, double& obstdist, double da, double headwind = 0) const;
			void calculate(double& gnddist, double& obstdist, double da, double headwind, double mass) const;

		  protected:
			Glib::ustring m_name;
			double m_vrotate;
			double m_mass;
			typedef Poly1D<double> poly_t;
			poly_t m_gndpoly;
			poly_t m_obstpoly;
			typedef std::vector<Point> points_t;
			points_t m_points;
		};

		Distances(double altfactor = 1, double tempfactor = 1, double tempoffset = 273.15, double distfactor = 1);

		unsigned int get_nrtakeoffdist(void) const { return m_takeoffdist.size(); }
		const Distance& get_takeoffdist(unsigned int i) const { return m_takeoffdist[i]; }
		Distance& get_takeoffdist(unsigned int i) { return m_takeoffdist[i]; }

		unsigned int get_nrldgdist(void) const { return m_ldgdist.size(); }
		const Distance& get_ldgdist(unsigned int i) const { return m_ldgdist[i]; }
		Distance& get_ldgdist(unsigned int i) { return m_ldgdist[i]; }

		bool recalculatepoly(bool force = false);

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

	  protected:
		typedef std::vector<Distance> distances_t;
		distances_t m_takeoffdist;
		distances_t m_ldgdist;
		double m_altfactor;
		double m_tempfactor;
		double m_tempoffset;
		double m_distfactor;
	};

	class Climb {
	  public:
		class Point {
		  public:
			Point(double da = 0, double rate = 0, double fuelflow = 0);
			
			void load_xml(const xmlpp::Element *el, double altfactor = 1, double ratefactor = 1, double fuelfactor = 1);
			void save_xml(xmlpp::Element *el, double altfactor = 1, double ratefactor = 1, double fuelfactor = 1) const;

			double get_densityalt(void) const { return m_da; }
			double get_rate(void) const { return m_rate; }
			double get_fuelflow(void) const { return m_fuelflow; }

		  protected:
			double m_da;
			double m_rate;
			double m_fuelflow;
		};

		Climb(double altfactor = 1, double ratefactor = 1, double fuelfactor = 1);

		void calculate(double& rate, double& fuelflow, double da);
		double get_ceiling(void) const { return m_ceiling; }
		double get_climbtime(void) const { return m_climbtime; }

		double time_to_altitude(double t) const;
		double time_to_distance(double t) const;
		double time_to_fuel(double t) const;
		double time_to_climbrate(double t) const;
		double time_to_tas(double t) const;
		double time_to_fuelflow(double t) const;
		double altitude_to_time(double a) const;
		double distance_to_time(double d) const;

		bool recalculatepoly(double vy, bool force = false);

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

	  protected:
		typedef Poly1D<double> poly_t;
		poly_t m_ratepoly;
		poly_t m_fuelflowpoly;
		poly_t m_climbtimepoly;
		poly_t m_climbaltpoly;
		poly_t m_climbdistpoly;
		poly_t m_climbfuelpoly;
		typedef std::vector<Point> points_t;
		points_t m_points;
		double m_altfactor;
		double m_ratefactor;
		double m_fuelfactor;
		double m_ceiling;
		double m_climbtime;
	};

	class Cruise {
	  public:
		class Point {
		  public:
			Point(double da = 0, double rpm = 0, double mp = 0, double bhp = 0, double tas = 0, double fuelflow = 0);
			
			void load_xml(const xmlpp::Element *el, double altfactor = 1, double tasfactor = 1, double fuelfactor = 1);
			void save_xml(xmlpp::Element *el, double altfactor = 1, double tasfactor = 1, double fuelfactor = 1) const;

			double get_densityalt(void) const { return m_da; }
			double get_rpm(void) const { return m_rpm; }
			double get_mp(void) const { return m_mp; }
			double get_bhp(void) const { return m_bhp; }
			double get_tas(void) const { return m_tas; }
			double get_fuelflow(void) const { return m_fuelflow; }

		  protected:
			double m_da;
			double m_rpm;
			double m_mp;
			double m_bhp;
			double m_tas;
			double m_fuelflow;
		};

		Cruise(double altfactor = 1, double tasfactor = 1, double fuelfactor = 1);

		bool is_constantspeed(void) const { return m_hasmp; }
		void calculate(double& tas, double& fuelflow, double& bhp, double& rpm, double& mp, double& da) const;

		void load_xml(const xmlpp::Element *el);
		void save_xml(xmlpp::Element *el) const;

	  protected:
		class Point1 {
		  public:
			Point1(double bhp = 0, double tas = 0, double fuelflow = 0) : m_bhp(bhp), m_tas(tas), m_fuelflow(fuelflow) {}
			double get_bhp(void) const { return m_bhp; }
			double get_tas(void) const { return m_tas; }
			double get_fuelflow(void) const { return m_fuelflow; }

		  protected:
			double m_bhp;
			double m_tas;
			double m_fuelflow;
		};

		typedef std::map<double,Point1> mpmap_t;
		typedef std::map<double,mpmap_t> rpmmap_t;
		typedef std::map<double,rpmmap_t> altmap_t;
		altmap_t m_altmap;
		typedef std::vector<Point> points_t;
		points_t m_points;
		double m_altfactor;
		double m_tasfactor;
		double m_fuelfactor;
		bool m_hasmp;

		void build_altmap(void);
		void calculate(double& tas, double& fuelflow, double& bhp, double& mp, double& rpm, altmap_t::const_iterator it) const;
		void calculate(double& tas, double& fuelflow, double& bhp, double& mp, rpmmap_t::const_iterator it) const;
	};

	class OtherInfo {
	  public:
		OtherInfo(const Glib::ustring& category = "", const Glib::ustring& text = "") : m_category(category), m_text(text) {}
		const Glib::ustring& get_category(void) const { return m_category; }
		const Glib::ustring& get_text(void) const { return m_text; }
		bool operator<(const OtherInfo& x) const { return m_category < x.m_category; }
		bool is_valid(void) const { return !m_category.empty(); }

	  protected:
		Glib::ustring m_category;
		Glib::ustring m_text;
	};
  protected:
	typedef std::set<OtherInfo> otherinfo_t;

  public:
	Aircraft(void);

	WeightBalance& get_wb(void) { return m_wb; }
	const WeightBalance& get_wb(void) const { return m_wb; }

	Distances& get_dist(void) { return m_dist; }
	const Distances& get_dist(void) const { return m_dist; }

	Climb& get_climb(void) { return m_climb; }
	const Climb& get_climb(void) const { return m_climb; }

	Cruise& get_cruise(void) { return m_cruise; }
	const Cruise& get_cruise(void) const { return m_cruise; }

	typedef otherinfo_t::const_iterator otherinfo_const_iterator_t;
	otherinfo_const_iterator_t otherinfo_begin(void) const { return m_otherinfo.begin(); }
	otherinfo_const_iterator_t otherinfo_end(void) const { return m_otherinfo.end(); }
	void otherinfo_add(const OtherInfo& oi);
	void otherinfo_add(const Glib::ustring& category, const Glib::ustring& text = "") { otherinfo_add(OtherInfo(category, text)); }
	const OtherInfo& otherinfo_find(const Glib::ustring& category) const;

	const Glib::ustring& get_callsign(void) const { return m_callsign; }
	const Glib::ustring& get_aircrafttype(void) const { return m_aircrafttype; }
	static const char *get_aircrafttypeclass(const Glib::ustring& acfttype);
	const char *get_aircrafttypeclass(void) const { return get_aircrafttypeclass(get_aircrafttype()); }
	const Glib::ustring& get_icaotype(void) const { return m_icaotype; }
	const Glib::ustring& get_equipment(void) const { return m_equipment; }
	const Glib::ustring& get_transponder(void) const { return m_transponder; }
	const Glib::ustring& get_pbn(void) const { return m_pbn; }
	const Glib::ustring& get_colormarking(void) const { return m_colormarking; }
	const Glib::ustring& get_emergencyradio(void) const { return m_emergencyradio; }
	const Glib::ustring& get_survival(void) const { return m_survival; }
	const Glib::ustring& get_lifejackets(void) const { return m_lifejackets; }
	const Glib::ustring& get_dinghies(void) const { return m_dinghies; }
	char get_wakecategory(void) const;
	double get_mtom(void) const { return m_mtom; }
	double get_vr0(void) const { return m_vr0; }
	double get_vr1(void) const { return m_vr1; }
	double get_vx0(void) const { return m_vx0; }
	double get_vx1(void) const { return m_vx1; }
	double get_vy(void) const { return m_vy; }
	double get_vs0(void) const { return m_vs0; }
	double get_vs1(void) const { return m_vs1; }
	double get_vno(void) const { return m_vno; }
	double get_vne(void) const { return m_vne; }
	double get_vref0(void) const { return m_vref0; }
	double get_vref1(void) const { return m_vref1; }
	double get_va(void) const { return m_va; }
	double get_vfe(void) const { return m_vfe; }
	double get_vgearext(void) const { return m_vgearext; }
	double get_vgearret(void) const { return m_vgearret; }
	double get_vbestglide(void) const { return m_vbestglide; }
	double get_glideslope(void) const { return m_glideslope; }
	double get_fuelmass(void) const { return m_fuelmass; }
	double get_maxbhp(void) const { return m_maxbhp; }

	void load_xml(const xmlpp::Element *el);
	void save_xml(xmlpp::Element *el) const;

	bool load_file(const Glib::ustring& filename);
	void save_file(const Glib::ustring& filename) const;

	bool load_string(const Glib::ustring& data);
	Glib::ustring save_string(void) const;

	void navfplan(const Glib::ustring& templatefile, Engine& engine, const FPlanRoute& fplan,
		      double cruise_rpm = std::numeric_limits<double>::quiet_NaN(),
		      double cruise_mp = std::numeric_limits<double>::quiet_NaN(),
		      const WeightBalance::elementvalues_t& wbev = WeightBalance::elementvalues_t()) const;

  protected:
	typedef char aircraft_type_class_t[10];
	static const aircraft_type_class_t aircraft_type_class_db[];

	class AircraftTypeClassDbCompare {
	  public:
		bool operator()(const char *a, const char *b) const { return strncmp(a, b, 4) < 0; }
	};

	WeightBalance m_wb;
	Distances m_dist;
	Climb m_climb;
	Cruise m_cruise;
	otherinfo_t m_otherinfo;
	// some basic data
	Glib::ustring m_callsign;
	Glib::ustring m_aircrafttype;
	Glib::ustring m_icaotype;
	Glib::ustring m_equipment;
	Glib::ustring m_transponder;
	Glib::ustring m_pbn;
	Glib::ustring m_colormarking;
	Glib::ustring m_emergencyradio;
	Glib::ustring m_survival;
	Glib::ustring m_lifejackets;
	Glib::ustring m_dinghies;
	double m_mtom;
	double m_vr0;
	double m_vr1;
	double m_vx0;
	double m_vx1;
	double m_vy;
	double m_vs0;
	double m_vs1;
	double m_vno;
	double m_vne;
	double m_vref0;
	double m_vref1;
	double m_va;
	double m_vfe;
	double m_vgearext;
	double m_vgearret;
	double m_vbestglide;
	double m_glideslope;
	double m_fuelmass;
	double m_maxbhp;

	class NavFPlan;

	void check_aircraft_type_class(void);
};

inline Aircraft::WeightBalance::Element::flags_t operator|(Aircraft::WeightBalance::Element::flags_t x, Aircraft::WeightBalance::Element::flags_t y) { return (Aircraft::WeightBalance::Element::flags_t)((unsigned int)x | (unsigned int)y); }
inline Aircraft::WeightBalance::Element::flags_t operator&(Aircraft::WeightBalance::Element::flags_t x, Aircraft::WeightBalance::Element::flags_t y) { return (Aircraft::WeightBalance::Element::flags_t)((unsigned int)x & (unsigned int)y); }
inline Aircraft::WeightBalance::Element::flags_t operator^(Aircraft::WeightBalance::Element::flags_t x, Aircraft::WeightBalance::Element::flags_t y) { return (Aircraft::WeightBalance::Element::flags_t)((unsigned int)x ^ (unsigned int)y); }
inline Aircraft::WeightBalance::Element::flags_t operator~(Aircraft::WeightBalance::Element::flags_t x){ return (Aircraft::WeightBalance::Element::flags_t)~(unsigned int)x; }
inline Aircraft::WeightBalance::Element::flags_t& operator|=(Aircraft::WeightBalance::Element::flags_t& x, Aircraft::WeightBalance::Element::flags_t y) { x = x | y; return x; }
inline Aircraft::WeightBalance::Element::flags_t& operator&=(Aircraft::WeightBalance::Element::flags_t& x, Aircraft::WeightBalance::Element::flags_t y) { x = x & y; return x; }
inline Aircraft::WeightBalance::Element::flags_t& operator^=(Aircraft::WeightBalance::Element::flags_t& x, Aircraft::WeightBalance::Element::flags_t y) { x = x ^ y; return x; }




#endif /* AIRCRAFT_H */
