/*****************************************************************************/

/*
 *      baro.cc  --  ICAO/U.S. Standard Atmosphere 1976 computations.
 *
 *      Copyright (C) 2003, 2007, 2012  Thomas Sailer (t.sailer@alumni.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#include "baro.h"
#include <cmath>
#include <stdexcept>

const float IcaoAtmosphere::const_g;                     /* gravity const. */
const float IcaoAtmosphere::const_m;                     /* Molecular Weight of air */
const float IcaoAtmosphere::const_r;                     /* gas const. for air */
const float IcaoAtmosphere::const_l0;                    /* Lapse rate first layer */
const float IcaoAtmosphere::const_gmr = (const_g*const_m/const_r);   /* Hydrostatic constant */
const float IcaoAtmosphere::degc_to_kelvin;
const float IcaoAtmosphere::std_sealevel_temperature;
const float IcaoAtmosphere::std_sealevel_pressure;
const float IcaoAtmosphere::std_sealevel_density = const_m / const_r * std_sealevel_pressure / std_sealevel_temperature;

const unsigned int IcaoAtmosphere::nr_layers;
const IcaoAtmosphere::AtmospericLayer IcaoAtmosphere::atmosperic_layers[nr_layers+1] = {
        {     0, -const_l0 },
        { 11000,    0 },
        { 20000,  1.0/1000 },
        { 32000,  2.8/1000 },
        { 47000,    0 },
        { 51000, -2.8/1000 },
        { 71000, -2.0/1000 },
        { 84852,    0 }
};

IcaoAtmosphere::IcaoAtmosphere(float qnh, float temp)
{
        set(qnh, temp);
}

void IcaoAtmosphere::compute_boundaries(void)
{
        for (unsigned int i = 0; i < nr_layers; i++) {
                if (atmosperic_layers[i].lapserate) {
                        m_bdry[i+1].t = m_bdry[i].t + atmosperic_layers[i].lapserate * (atmosperic_layers[i+1].limit - atmosperic_layers[i].limit);
                        //m_bdry[i+1].p = m_bdry[i].p * exp(log(m_bdry[i].t / m_bdry[i+1].t) * (const_gmr / 1000) / atmosperic_layers[i].lapserate);
                        m_bdry[i+1].p = m_bdry[i].p * pow(m_bdry[i].t / m_bdry[i+1].t, (const_gmr / 1000) / atmosperic_layers[i].lapserate);
                } else {
                        m_bdry[i+1].t = m_bdry[i].t;
                        m_bdry[i+1].p = m_bdry[i].p * exp((-const_gmr / 1000) * (atmosperic_layers[i+1].limit - atmosperic_layers[i].limit) / m_bdry[i].t);
                }
        }
}

void IcaoAtmosphere::set(float qnh, float temp)
{
        m_bdry[0].t = temp;
        m_bdry[0].p = qnh;
        compute_boundaries();
}

void IcaoAtmosphere::altitude_to_pressure(float *press, float *temp, float alt, const atmospheric_boundary b[nr_layers+1])
{
        if (false && alt < 0)
                throw std::runtime_error("IcaoAtmosphere::altitude_to_pressure: altitude negative");
        unsigned int i;
        for (i = 0; i < nr_layers; i++)
                if (alt <= atmosperic_layers[i+1].limit)
                        break;
        if (i >= nr_layers)
                throw std::runtime_error("IcaoAtmosphere::altitude_to_pressure: altitude above model limit");
        //std::cerr << "Alt " << alt << " layer " << i << std::endl;
        float t = b[i].t + atmosperic_layers[i].lapserate * (alt - atmosperic_layers[i].limit);
        if (temp)
                *temp = t;
        if (!press)
                return;
        if (atmosperic_layers[i].lapserate) {
                *press = b[i].p * pow(b[i].t / t, (const_gmr / 1000) / atmosperic_layers[i].lapserate);
                return;
        }
        *press = b[i].p * exp((-const_gmr / 1000) * (alt - atmosperic_layers[i].limit) / b[i].t);
}

void IcaoAtmosphere::pressure_to_altitude(float *alt, float *temp, float press, const atmospheric_boundary b[nr_layers+1])
{

        if (/*press > b[0].p ||*/ press <= 0)
                throw std::runtime_error("IcaoAtmosphere::pressure_to_altitude: pressure below zero");
        unsigned int i;
        for (i = 0; i < nr_layers; i++)
                if (press > b[i+1].p)
                        break;
        if (i >= nr_layers)
                throw std::runtime_error("IcaoAtmosphere::pressure_to_altitude: pressure lower than model limit");
        //std::cerr << "Pressure " << press << " layer " << i << std::endl;
        float a;
        if (atmosperic_layers[i].lapserate) {
		a = (b[i].t / pow(press / b[i].p, atmosperic_layers[i].lapserate * (1000 / const_gmr)) - b[i].t);
		a /= atmosperic_layers[i].lapserate;
		a += atmosperic_layers[i].limit;
	} else {
                a = log(press / b[i].p) * b[i].t * (-1000 / const_gmr) + atmosperic_layers[i].limit;
        }
        if (alt)
                *alt = a;
        if (temp)
                *temp = b[i].t + atmosperic_layers[i].lapserate * (a - atmosperic_layers[i].limit);
}

float IcaoAtmosphere::pressure_to_density(float pressure, float temp)
{
        return (const_m/const_r) * pressure / std::max(temp, 0.01f);
}

float IcaoAtmosphere::density_to_pressure(float density, float temp)
{
        return density * temp * (const_r / const_m);
}
