//
// C++ Implementation: dbobjtopo
//
// Description: Database Objects: Topography
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2009, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "sysdeps.h"

#include <limits>
#include <zlib.h>

#include "dbobj.h"

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::pixel_index_t TopoDbN<resolution,tilesize>::tile_size;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::nodata = std::numeric_limits<elev_t>::min();

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::ocean = std::numeric_limits<elev_t>::min() + 1;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::TopoCoordinate::coord_t TopoDbN<resolution,tilesize>::TopoCoordinate::from_point;

template<int resolution, int tilesize>
const uint64_t TopoDbN<resolution,tilesize>::TopoCoordinate::to_point;

template<int resolution, int tilesize>
const unsigned int TopoDbN<resolution,tilesize>::tile_cache_size;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::tile_index_t TopoDbN<resolution,tilesize>::lon_tiles;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::tile_index_t TopoDbN<resolution,tilesize>::lat_tiles;

template<int resolution, int tilesize>
const typename TopoDbN<resolution,tilesize>::tile_index_t TopoDbN<resolution,tilesize>::nr_tiles;

template<int resolution, int tilesize>
const uint16_t TopoDbN<resolution,tilesize>::TopoTileCoordinate::lon_tiles;

template<int resolution, int tilesize>
const uint16_t TopoDbN<resolution,tilesize>::TopoTileCoordinate::lat_tiles;

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoTileCoordinate::TopoTileCoordinate(const TopoCoordinate & tc)
{
        m_lon_tile = tc.get_lon() / tilesize;
        m_lat_tile = tc.get_lat() / tilesize;
        m_lon_offs = tc.get_lon() - m_lon_tile * tilesize;
        m_lat_offs = tc.get_lat() - m_lat_tile * tilesize;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_east(void)
{
        m_lon_tile++;
        if (m_lon_tile >= lon_tiles)
                m_lon_tile = 0;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_north(void)
{
        if (m_lat_tile < lat_tiles)
                m_lat_tile++;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_west(void)
{
        if (!m_lon_tile)
                m_lon_tile = lon_tiles;
        m_lon_tile--;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoTileCoordinate::advance_tile_south(void)
{
        if (m_lat_tile)
                m_lat_tile--;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoCoordinate::TopoCoordinate(const Point & pt)
{
        int64_t lat1 = pt.get_lat();
        lat1 += 0x40000000;
        uint64_t lon = (uint32_t)pt.get_lon();
        uint64_t lat = std::min(std::max(lat1, (int64_t)0), (int64_t)0x80000000);
        lon *= from_point;
        lat *= from_point;
        m_lon = lon >> 32;
        m_lat = lat >> 32;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoCoordinate::operator Point(void) const
{
        uint64_t lon(m_lon), lat(m_lat);
        lon *= to_point;
        lat *= to_point;
        lon += (((uint64_t)1) << 31);
        lat += (((uint64_t)1) << 31);
        Point pt(lon >> 32, (lat >> 32) - 0x40000000);
        return pt;
}

template<int resolution, int tilesize>
std::ostream & TopoDbN<resolution,tilesize>::TopoCoordinate::print(std::ostream & os)
{
        return os << '(' << get_lon() << ',' << get_lat() << ')';
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_east(void)
{
        m_lon++;
        if (m_lon == from_point)
                m_lon = 0;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_west(void)
{
        if (!m_lon)
                m_lon = from_point;
        m_lon--;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_north(void)
{
        if (m_lat < from_point/2)
                m_lat++;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::TopoCoordinate::advance_south(void)
{
        if (m_lat)
                m_lat--;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::TopoCoordinate::coord_t TopoDbN<resolution,tilesize>::TopoCoordinate::lon_dist(const TopoCoordinate & tc) const
{
        coord_t r = tc.get_lon() + from_point - get_lon();
        if (r >= from_point)
                r -= from_point;
        return r;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::TopoCoordinate::coord_t TopoDbN<resolution,tilesize>::TopoCoordinate::lat_dist(const TopoCoordinate & tc) const
{
        return tc.get_lat() - get_lat();
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::TopoCoordinate::coords_t TopoDbN<resolution,tilesize>::TopoCoordinate::lon_dist_signed(const TopoCoordinate& tc) const
{
	coords_t r(lon_dist(tc));
	if (r >= (from_point / 2U))
		r -= from_point;
	return r;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::TopoCoordinate::coords_t TopoDbN<resolution,tilesize>::TopoCoordinate::lat_dist_signed(const TopoCoordinate& tc) const
{
        return tc.get_lat() - get_lat();
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::Tile(tile_index_t index)
        : m_index(index), m_minelev(nodata), m_maxelev(nodata), m_dirty(false), m_loaded(false), m_minmaxinvalid(false)
{
        m_elev = new elev_t[tilesize * tilesize];
        for (pixel_index_t idx = 0; idx < tilesize * tilesize; idx++)
                m_elev[idx] = nodata;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::Tile(const Tile & tile)
        : m_index(tile.m_index), m_minelev(tile.m_minelev), m_maxelev(tile.m_maxelev), m_dirty(tile.m_dirty), m_loaded(tile.m_loaded), m_minmaxinvalid(tile.m_minmaxinvalid)
{
        m_elev = new elev_t[tilesize * tilesize];
        for (pixel_index_t idx = 0; idx < tilesize * tilesize; idx++)
                m_elev[idx] = tile.m_elev[idx];
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::Tile(sqlite3x::sqlite3_connection & db, tile_index_t index)
        : m_index(index), m_minelev(nodata), m_maxelev(nodata), m_dirty(false), m_loaded(false), m_minmaxinvalid(false)
{
        m_elev = new elev_t[tilesize * tilesize];
        for (pixel_index_t idx = 0; idx < tilesize * tilesize; idx++)
                m_elev[idx] = nodata;
        std::string qs("SELECT MINELEV,MAXELEV,ELEV FROM topo WHERE ID=?1;");
        if (false)
		std::cerr << "sql: " << qs << std::endl;
        sqlite3x::sqlite3_command cmd(db, qs);
        int nr = 1;
        cmd.bind(nr++, (int)m_index);
        sqlite3x::sqlite3_cursor cursor = cmd.executecursor();
        if (!cursor.step())
                return;
        m_loaded = true;
        elev_t minelev = m_minelev = cursor.getint(0);
        elev_t maxelev = m_maxelev = cursor.getint(1);
        if (m_minelev == ocean)
                m_minelev = 0;
        if (m_maxelev == ocean)
                m_maxelev = 0;
	int blobsize(0);
        if (cursor.isnull(2)) {
                elev_t e(nodata);
                if (minelev == maxelev)
                        e = minelev;
                for (pixel_index_t i = 0; i < tilesize * tilesize; i++)
                        m_elev[i] = e;
        } else {
                void const *blob(cursor.getblob(2, blobsize));
                z_stream d_stream;
                memset(&d_stream, 0, sizeof(d_stream));
                d_stream.zalloc = (alloc_func)0;
                d_stream.zfree = (free_func)0;
                d_stream.opaque = (voidpf)0;
                d_stream.next_in  = (Bytef *)blob;
                d_stream.avail_in = blobsize;
                int err = inflateInit(&d_stream);
                if (err != Z_OK) {
                        inflateEnd(&d_stream);
                        std::cerr << "zlib: inflateInit error " << err << std::endl;
                        return;
                }
                uint8_t *data = new uint8_t[2 * tilesize * tilesize];
                d_stream.next_out = data;
                d_stream.avail_out = 2 * tilesize * tilesize;
                err = inflate(&d_stream, Z_FINISH);
                inflateEnd(&d_stream);
                if (err != Z_STREAM_END) {
                        delete[] data;
                        std::cerr << "zlib: inflate error " << err << std::endl;
                        return;
                }
                for (pixel_index_t i = 0; i < tilesize * tilesize; i++)
                        m_elev[i] = data[2*i] + (data[2*i+1] << 8);
                delete[] data;
        }
        if (false)
		std::cerr << "TopoDb::Tile::Tile: index " << index << " blobsize " << blobsize << std::endl;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::Tile::~Tile(void)
{
        delete[] m_elev;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Tile::save(sqlite3x::sqlite3_connection & db)
{
        if (!m_dirty)
                return;
        if (m_minmaxinvalid)
                update_minmax();
        std::cerr << "Saving tile: ID " << m_index << " min: " << m_minelev << " max: " << m_maxelev << std::endl;
        sqlite3x::sqlite3_transaction tran(db);
        if (!m_loaded) {
                sqlite3x::sqlite3_command cmd(db, "INSERT INTO topo (ID) VALUES(?);");
                cmd.bind(1, (int)m_index);
                cmd.executenonquery();
        }
        {
                sqlite3x::sqlite3_command cmd(db, "UPDATE topo SET MINELEV=?2,MAXELEV=?3,ELEV=?4 WHERE ID=?1;");
                cmd.bind(1, (int)m_index);
                if (m_minelev == 0 && m_maxelev == 0 && m_elev[0] == ocean) {
                        cmd.bind(2, (long long int)ocean);
                        cmd.bind(3, (long long int)ocean);
                } else {
                        cmd.bind(2, (long long int)m_minelev);
                        cmd.bind(3, (long long int)m_maxelev);
                }
                cmd.bind(4);
                if (m_minelev != m_maxelev) {
                        unsigned int els(tilesize * tilesize);
                        uint8_t edata[2*els];
                        for (unsigned int i = 0; i < els; i++) {
                                edata[2*i] = m_elev[i];
                                edata[2*i+1] = m_elev[i] >> 8;
                        }
                        uint8_t cdata[2*els+64];
                        z_stream c_stream;
                        memset(&c_stream, 0, sizeof(c_stream));
                        c_stream.zalloc = (alloc_func)0;
                        c_stream.zfree = (free_func)0;
                        c_stream.opaque = (voidpf)0;
                        int err = deflateInit(&c_stream, Z_BEST_COMPRESSION);
                        if (err) {
                                std::cerr << "deflateInit error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        c_stream.next_out = cdata;
                        c_stream.avail_out = 2*els+64;
                        c_stream.next_in = (Bytef *)edata;
                        c_stream.avail_in = 2*els;
                        err = deflate(&c_stream, Z_FINISH);
                        if (err != Z_STREAM_END) {
                                std::cerr << "deflate error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                        if (c_stream.avail_in != 0 || c_stream.total_in != 2*els) {
                                std::cerr << "deflate did not consume all input" << std::endl;
                                goto outerr2;
                        }
                        cmd.bind(4, cdata, c_stream.total_out);
                        std::cerr << "Tile compression: input: " << (2*els) << " output: " << c_stream.total_out << std::endl;
                        err = deflateEnd(&c_stream);
                        if (err) {
                                std::cerr << "deflateEnd error " << err << ", " << c_stream.msg << std::endl;
                                goto outerr2;
                        }
                outerr2:;
                }
                cmd.executenonquery();
        }
        tran.commit();
        m_loaded = true;
        m_dirty = false;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::Tile::get_elev(pixel_index_t index) const
{
        if (index < 0 || index >= tilesize * tilesize)
                return nodata;
        return m_elev[index];
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Tile::set_elev(pixel_index_t index, elev_t elev)
{
        if (index < 0 || index >= tilesize * tilesize)
                return;
        if (elev != m_elev[index])
                m_dirty = true;
        m_elev[index] = elev;
        m_minmaxinvalid = true;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Tile::update_minmax(void)
{
        int16_t mi(std::numeric_limits<elev_t>::max()), ma(std::numeric_limits<elev_t>::min());
        unsigned int els(tilesize * tilesize);
        for (unsigned int i = 0; i < els; i++) {
                elev_t e(m_elev[i]);
                if (e == nodata)
                        continue;
                if (e == ocean)
                        e = 0;
                mi = std::min(mi, e);
                ma = std::max(ma, e);
        }
        m_minelev = mi;
        m_maxelev = ma;
        m_minmaxinvalid = false;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::ProfilePoint::ProfilePoint(double dist, elev_t elev, elev_t minelev, elev_t maxelev)
	: m_dist(dist), m_elev(elev), m_minelev(minelev), m_maxelev(maxelev)
{
	if (m_elev == ocean)
		m_elev = 0;
	if (m_minelev == ocean)
		m_minelev = 0;
	if (m_maxelev == ocean)
	        m_maxelev = 0;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::ProfilePoint::ProfilePoint(double dist, elev_t elev)
	: m_dist(dist), m_elev(elev), m_minelev(elev), m_maxelev(elev)
{
	if (m_elev == ocean)
		m_elev = m_minelev = m_maxelev = 0;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::ProfilePoint::ProfilePoint(double dist)
	: m_dist(dist), m_elev(nodata), m_minelev(nodata), m_maxelev(nodata)
{
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::ProfilePoint::scaledist(double ds)
{
	m_dist *= ds;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::ProfilePoint::adddist(double d)
{
	m_dist += d;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::ProfilePoint::add(elev_t elev)
{
	if (elev == nodata)
		return;
	if (elev == ocean)
		elev = 0;
	if (m_elev == nodata) {
		m_elev = elev;
		m_minelev = elev;
		m_maxelev = elev;
		return;
	}
	m_minelev = std::min(m_minelev, elev);
	m_maxelev = std::max(m_maxelev, elev);
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Profile::scaledist(double ds)
{
	for (typename base_t::iterator i(this->begin()), e(this->end()); i != e; ++i)
		i->scaledist(ds);
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::Profile::adddist(double d)
{
	for (typename base_t::iterator i(this->begin()), e(this->end()); i != e; ++i)
		i->adddist(d);
}

template<int resolution, int tilesize>
double TopoDbN<resolution,tilesize>::Profile::get_dist(void) const
{
	if (this->empty())
		return 0;
	return this->operator[](this->size() - 1).get_dist();
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::Profile::get_minelev(void) const
{
	elev_t el(std::numeric_limits<elev_t>::max());
	for (typename base_t::const_iterator i(this->begin()), e(this->end()); i != e; ++i) {
		elev_t ee(i->get_minelev());
		if (ee == nodata)
			continue;
		el = std::min(el, ee);
	}
	if (el == std::numeric_limits<elev_t>::max())
		return nodata;
	return el;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::Profile::get_maxelev(void) const
{
	elev_t el(std::numeric_limits<elev_t>::min());
	for (typename base_t::const_iterator i(this->begin()), e(this->end()); i != e; ++i) {
		elev_t ee(i->get_maxelev());
		if (ee == nodata)
			continue;
		el = std::max(el, ee);
	}
	if (el == std::numeric_limits<elev_t>::min())
		return nodata;
	return el;
}

template<int resolution, int tilesize>
TopoDbN<resolution,tilesize>::TopoDbN()
{
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::open(const Glib::ustring& path, const Glib::ustring& dbname)
{
        Glib::ustring dbname1(path);
        if (dbname1.empty())
                dbname1 = FPlan::get_userdbpath();
        dbname1 = Glib::build_filename(dbname1, dbname);
        m_db.open(dbname1);
        create_tables();
        create_indices();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::take(sqlite3 * dbh)
{
        m_db.take(dbh);
        if (!m_db.db())
                return;
        create_tables();
        create_indices();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::close(void)
{
        cache_commit();
        m_db.close();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::purgedb(void)
{
        m_tile_cache.clear();
        drop_indices();
        drop_tables();
        create_tables();
        create_indices();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::sync_off(void)
{
        sqlite3x::sqlite3_command cmd(m_db, "PRAGMA synchronous = OFF;");
        cmd.executenonquery();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::analyze(void)
{
        cache_commit();
        sqlite3x::sqlite3_command cmd(m_db, "ANALYZE;");
        cmd.executenonquery();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::vacuum(void)
{
        cache_commit();
        sqlite3x::sqlite3_command cmd(m_db, "VACUUM;");
        cmd.executenonquery();
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::interrupt(void)
{
        sqlite3_interrupt(m_db.db());
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::drop_tables(void)
{
        { sqlite3x::sqlite3_command cmd(m_db, "DROP TABLE IF EXISTS topo;"); cmd.executenonquery(); }
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::create_tables(void)
{
        {
                sqlite3x::sqlite3_command cmd(m_db, "CREATE TABLE IF NOT EXISTS topo (ID INTEGER PRIMARY KEY, "
                                              "MINELEV INTEGER,"
                                              "MAXELEV INTEGER,"
                                              "ELEV BLOB);");
                cmd.executenonquery();
        }
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::drop_indices(void)
{
        //{ sqlite3x::sqlite3_command cmd(m_db, "DROP INDEX IF EXISTS topo_bbox;"); cmd.executenonquery(); }
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::create_indices(void)
{
        //{ sqlite3x::sqlite3_command cmd(m_db, "CREATE INDEX IF NOT EXISTS topo_bbox ON topo(SWLAT,NELAT,SWLON,NELON);"); cmd.executenonquery(); }
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::Tile * TopoDbN<resolution,tilesize>::find(tile_index_t index)
{
        if (index >= nr_tiles)
                return 0;
        unsigned int nrctiles(0);
        for (typename tile_cache_t::iterator ti(m_tile_cache.begin()), te(m_tile_cache.end()); ti != te; ti++, nrctiles++) {
                if (ti->get_index() == index) {
                        // move *ti to the beginning of the list without copying it
                        m_tile_cache.splice(m_tile_cache.begin(), m_tile_cache, ti);
                        return &*ti;
                }
        }
        if (nrctiles >= tile_cache_size) {
                typename tile_cache_t::iterator ti(--m_tile_cache.end());
                ti->save(m_db);
                m_tile_cache.erase(ti);
        }
        typename tile_cache_t::iterator ti(m_tile_cache.insert(m_tile_cache.begin(), Tile(m_db, index)));
        return &*ti;
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::cache_commit(void)
{
        for (typename tile_cache_t::iterator ti(m_tile_cache.begin()), te(m_tile_cache.end()); ti != te; ti++) {
                ti->save(m_db);
        }
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::elev_t TopoDbN<resolution,tilesize>::get_elev(const TopoTileCoordinate & tc)
{
        Tile *t(find(tc.get_tile_index()));
        if (!t) {
                std::cerr << "TopoDbN: cannot find tile " << tc.get_tile_index() << std::endl;
                return nodata;
        }
        return t->get_elev(tc.get_pixel_index());
}

template<int resolution, int tilesize>
void TopoDbN<resolution,tilesize>::set_elev(const TopoTileCoordinate & tc, elev_t elev)
{
        Tile *t(find(tc.get_tile_index()));
        if (!t) {
                std::cerr << "TopoDbN: cannot find tile " << tc.get_tile_index() << std::endl;
                return;
        }
        t->set_elev(tc.get_pixel_index(), elev);
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_tile_minmax_elev(tile_index_t index)
{
        Tile *t(find(index));
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        if (!t) {
                std::cerr << "TopoDbN: cannot find tile " << index << std::endl;
                return ret;
        }
        ret.first = t->get_minelev();
        ret.second = t->get_maxelev();
        return ret;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_minmax_elev(const Rect& r)
{
        Point pdim(TopoCoordinate::get_pointsize());
        TopoTileCoordinate tmin(r.get_southwest()), tmax(r.get_northeast());
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        for (TopoTileCoordinate tc(tmin);;) {
                pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : tile_size - 1);
                pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : tile_size - 1);
                Rect rtile((Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmin, ymin),
                           (Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmax, ymax) + pdim);
                if (r.is_inside(rtile)) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        ret.first = std::min(ret.first, mm.first);
                        ret.second = std::max(ret.second, mm.second);
                } else {
                        for (pixel_index_t y(ymin); y <= ymax; y++) {
                                tc.set_lat_offs(y);
                                for (pixel_index_t x(xmin); x <= xmax; x++) {
                                        tc.set_lon_offs(x);
                                        elev_t e(get_elev(tc));
                                        if (e == ocean)
                                                e = 0;
                                        if (e != nodata) {
                                                ret.first = std::min(ret.first, e);
                                                ret.second = std::max(ret.second, e);
                                        }
                                }
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                tc.advance_tile_north();
        }
        if (ret.first == std::numeric_limits<elev_t>::max())
                ret.first = nodata;
        if (ret.second == std::numeric_limits<elev_t>::min())
                ret.second = nodata;
        return ret;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_minmax_elev(const PolygonSimple& p)
{
        if (p.empty())
                return minmax_elev_t(nodata, nodata);
        Point pdim(TopoCoordinate::get_pointsize()), pdim2(pdim.get_lon()/2, pdim.get_lat()/2);
        Rect r(p.get_bbox());
        TopoTileCoordinate tmin(r.get_southwest()), tmax(r.get_northeast());
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        for (TopoTileCoordinate tc(tmin);;) {
                pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : tile_size - 1);
                pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : tile_size - 1);
                Rect rtile((Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmin, ymin),
                           (Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmax, ymax) + pdim);
                bool intersect(p.is_intersection(rtile));
                bool poly_inside(!intersect && rtile.is_inside(p[0]));
                bool tile_inside(!intersect && !poly_inside && p.windingnumber(rtile.get_southwest()));
                if (false) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        std::cerr << "get_minmax_elev: bbox " << r << " tile " << rtile << " intersect " << (intersect ? "yes" : "no")
                                  << " poly_inside " << (poly_inside ? "yes" : "no") << " tile_inside " << (tile_inside ? "yes" : "no")
                                  << " tile elev " << mm.first << " / " << mm.second << std::endl;
                }
                if (tile_inside) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        ret.first = std::min(ret.first, mm.first);
                        ret.second = std::max(ret.second, mm.second);
                } else if (poly_inside || intersect) {
                        for (pixel_index_t y(ymin); y <= ymax; y++) {
                                tc.set_lat_offs(y);
                                for (pixel_index_t x(xmin); x <= xmax; x++) {
                                        tc.set_lon_offs(x);
                                        Point pt2(pdim2 + (Point)(TopoCoordinate)tc);
                                        if (r.is_inside(pt2) && p.windingnumber(pt2)) {
                                                elev_t e(get_elev(tc));
                                                if (false)
							std::cerr << "get_minmax_elev: " << e << " @ " << pt2 << std::endl;
                                                if (e == ocean)
                                                        e = 0;
                                                if (e != nodata) {
                                                        ret.first = std::min(ret.first, e);
                                                        ret.second = std::max(ret.second, e);
                                                }
                                        }
                                }
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                tc.advance_tile_north();
        }
        if (ret.first == std::numeric_limits<elev_t>::max())
                ret.first = nodata;
        if (ret.second == std::numeric_limits<elev_t>::min())
                ret.second = nodata;
        return ret;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_minmax_elev(const PolygonHole& p)
{
        if (p.get_exterior().empty())
                return minmax_elev_t(nodata, nodata);
        Point pdim(TopoCoordinate::get_pointsize()), pdim2(pdim.get_lon()/2, pdim.get_lat()/2);
        Rect r(p.get_bbox());
        TopoTileCoordinate tmin(r.get_southwest()), tmax(r.get_northeast());
        minmax_elev_t ret(std::numeric_limits<elev_t>::max(), std::numeric_limits<elev_t>::min());
        for (TopoTileCoordinate tc(tmin);;) {
                pixel_index_t ymin((tc.get_lat_tile() == tmin.get_lat_tile()) ? tmin.get_lat_offs() : 0);
                pixel_index_t ymax((tc.get_lat_tile() == tmax.get_lat_tile()) ? tmax.get_lat_offs() : tile_size - 1);
                pixel_index_t xmin((tc.get_lon_tile() == tmin.get_lon_tile()) ? tmin.get_lon_offs() : 0);
                pixel_index_t xmax((tc.get_lon_tile() == tmax.get_lon_tile()) ? tmax.get_lon_offs() : tile_size - 1);
                Rect rtile((Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmin, ymin),
                           (Point)(TopoCoordinate)TopoTileCoordinate(tc.get_lon_tile(), tc.get_lat_tile(), xmax, ymax) + pdim);
                bool intersect(p.is_intersection(rtile));
                bool poly_inside(!intersect && rtile.is_inside(p.get_exterior()[0]));
                bool tile_inside(!intersect && !poly_inside && p.windingnumber(rtile.get_southwest()));
                if (false) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        std::cerr << "get_minmax_elev: bbox " << r << " tile " << rtile << " intersect " << (intersect ? "yes" : "no")
                                  << " poly_inside " << (poly_inside ? "yes" : "no") << " tile_inside " << (tile_inside ? "yes" : "no")
                                  << " tile elev " << mm.first << " / " << mm.second << std::endl;
                }
                if (tile_inside) {
                        minmax_elev_t mm(get_tile_minmax_elev(tc));
                        ret.first = std::min(ret.first, mm.first);
                        ret.second = std::max(ret.second, mm.second);
                } else if (poly_inside || intersect) {
                        for (pixel_index_t y(ymin); y <= ymax; y++) {
                                tc.set_lat_offs(y);
                                for (pixel_index_t x(xmin); x <= xmax; x++) {
                                        tc.set_lon_offs(x);
                                        Point pt2(pdim2 + (Point)(TopoCoordinate)tc);
                                        if (r.is_inside(pt2) && p.windingnumber(pt2)) {
                                                elev_t e(get_elev(tc));
                                                if (false)
							std::cerr << "get_minmax_elev: " << e << " @ " << pt2 << std::endl;
                                                if (e == ocean)
                                                        e = 0;
                                                if (e != nodata) {
                                                        ret.first = std::min(ret.first, e);
                                                        ret.second = std::max(ret.second, e);
                                                }
                                        }
                                }
                        }
                }
                if (tc.get_lon_tile() != tmax.get_lon_tile()) {
                        tc.advance_tile_east();
                        continue;
                }
                if (tc.get_lat_tile() == tmax.get_lat_tile())
                        break;
                tc.set_lon_tile(tmin.get_lon_tile());
                tc.advance_tile_north();
        }
        if (ret.first == std::numeric_limits<elev_t>::max())
                ret.first = nodata;
        if (ret.second == std::numeric_limits<elev_t>::min())
                ret.second = nodata;
        return ret;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::minmax_elev_t TopoDbN<resolution,tilesize>::get_minmax_elev(const MultiPolygonHole& p)
{
	minmax_elev_t r(nodata, nodata);
	for (MultiPolygonHole::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
		minmax_elev_t m(get_minmax_elev(*pi));
		if (r.first == nodata)
			r.first = m.first;
		else if (m.first != nodata)
			r.first = std::min(r.first, m.first);
		if (r.second == nodata)
			r.second = m.second;
		else if (m.second != nodata)
			r.second = std::max(r.second, m.second);
	}
	return r;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution,tilesize>::Profile TopoDbN<resolution,tilesize>::get_profile(const Point& p0, const Point& p1, double corridor_nmi)
{
	if (p0.is_invalid() || p1.is_invalid() || corridor_nmi < 0 || corridor_nmi > 10)
		return Profile();
	double gcdist(p0.spheric_distance_nmi_dbl(p1));
	if (gcdist > 200)
		return Profile();
	Profile result;
	TopoCoordinate l0(p0), l1(p1), c0, c1;
	// corridor line coordinates
	{
		double crs(p0.spheric_true_course_dbl(p1));
		Point px0(p0.spheric_course_distance_nmi(crs + 90, corridor_nmi));
		Point px1(p0 + (p0 - px0));
		if (px0.is_invalid() || px1.is_invalid())
			return Profile();
		c0 = TopoCoordinate(px0);
		c1 = TopoCoordinate(px1);
	}
	// line step distance map
	double distmap[4];
	distmap[0] = 0;
	{
		Point ptsz(TopoCoordinate::get_pointsize());
		Point p(p0.halfway(p1));
		for (unsigned int i = 1; i < 4; ++i) {
			Point p1((i & 1) ? ptsz.get_lon() : 0, (i & 2) ? ptsz.get_lat() : 0);
			p1 += p;
			distmap[i] = p.spheric_distance_nmi_dbl(p1);
		}
	}
	typename TopoCoordinate::coords_t ldx(l0.lon_dist_signed(l1)), ldy(l0.lat_dist_signed(l1));
	bool lsx(ldx < 0), lsy(ldy < 0);
	ldx = abs(ldx);
	ldy = abs(ldy);
	typename TopoCoordinate::coords_t cdx(c0.lon_dist_signed(c1)), cdy(c0.lat_dist_signed(c1));
	bool csx(cdx < 0), csy(cdy < 0);
	cdx = abs(cdx);
	cdy = abs(cdy);
	double dist(0);
	typename TopoCoordinate::coords_t lerr(ldx - ldy);
	TopoCoordinate ll(l0);
	for (;;) {
		result.push_back(ProfilePoint(dist, get_elev(ll)));
		typename TopoCoordinate::coords_t cerr(cdx - cdy);
		TopoCoordinate cc(c0);
		for (;;) {
			result.back().add(get_elev(cc));
			if (cc == c1)
				break;
			typename TopoCoordinate::coords_t cerr2(2 * cerr);
			if (cerr2 > -cdy) {
				cerr -= cdy;
				if (csx)
					cc.advance_west();
				else
					cc.advance_east();
			}
			if (cerr2 < cdx) {
				cerr += cdx;
				if (csy)
					cc.advance_south();
				else
					cc.advance_north();
			}
		}
		if (ll == l1)
			break;
		typename TopoCoordinate::coords_t lerr2(2 * lerr);
		unsigned int dmi(0);
		if (lerr2 > -ldy) {
			lerr -= ldy;
			if (lsx) {
				ll.advance_west();
				c0.advance_west();
				c1.advance_west();
			} else {
				ll.advance_east();
				c0.advance_east();
				c1.advance_east();
			}
			dmi |= 1U;
		}
		if (lerr2 < ldx) {
			lerr += ldx;
			if (lsy) {
				ll.advance_south();
				c0.advance_south();
				c1.advance_south();
			} else {
				ll.advance_north();
				c0.advance_north();
				c1.advance_north();
			}
			dmi |= 2U;
		}
		dist += distmap[dmi];
	}
	if (dist > 0)
		result.scaledist(gcdist / dist);
	return result;
}

template<int resolution, int tilesize>
std::ostream & operator <<(std::ostream & os, const typename TopoDbN< resolution, tilesize >::TopoCoordinate & tc)
{
        return tc.print(os);
}

void TopoDb30::open(const Glib::ustring & path)
{
        TopoDbNN::open(path, "topo30.db");
}

template<int resolution, int tilesize>
TopoDbN<resolution, tilesize>::RGBColor::operator Gdk::Color(void) const
{
        Gdk::Color col;
        col.set_rgb((m_col >> 8) & 0xff00, m_col & 0xff00, (m_col << 8) & 0xff00);
        return col;
}

template<int resolution, int tilesize>
typename TopoDbN<resolution, tilesize>::RGBColor TopoDbN<resolution, tilesize>::color(elev_t x)
{
#include "gtopocolor.h"
        return RGBColor(0, 0, 255);
}

template<int resolution, int tilesize>
int TopoDbN<resolution, tilesize>::m_to_ft(elev_t elev)
{
        return (elev << 2) - elev + (elev >> 2) + (elev >> 5) - (elev >> 11);
}

template class TopoDbN<30,480>;
//const TopoDb30::elev_t TopoDb30::nodata = TopoDbNN::nodata;
