//
// C++ Interface: fplan
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2012
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef FPLAN_H
#define FPLAN_H

#include <sqlite3x.hpp>
#include <glibmm.h>
#include <time.h>
#include "geom.h"

#define USERDBPATH ".vfrnav"

class IcaoAtmosphere;
class FPlanLeg;
class FPlan;

class FPlanWaypoint {
        public:
                static const time_t unix_minus_palmos_epoch = 24472 * 24 * 60 * 60;
                static const uint16_t altflag_standard = 1 << 0;
                static const uint16_t altflag_ifr = 1 << 15;
                static const float ft_to_m = 0.3048;
                static const float m_to_ft = 1.0 / 0.3048;
		static const float to_rad = M_PI / (1 << 15);
		static const float to_deg = 180.0 / (1 << 15);
 		static const float from_rad = (1 << 15) / M_PI;
		static const float from_deg = (1 << 15) / 180.0;
                static const std::string reader_fields;

		typedef enum {
			pathcode_none,
			pathcode_sid,
			pathcode_star,
			pathcode_approach,
			pathcode_transition,
			pathcode_vfrdeparture,
			pathcode_vfrarrival,
			pathcode_vfrtransition,
			pathcode_airway,
			pathcode_directto
 		} pathcode_t;

                FPlanWaypoint(void);
                FPlanWaypoint(sqlite3x::sqlite3_cursor& cursor);

                const Glib::ustring& get_icao(void) const { return m_icao; }
                void set_icao(const Glib::ustring& t) { m_icao = t; m_dirty = true; }
                const Glib::ustring& get_name(void) const { return m_name; }
                void set_name(const Glib::ustring& t) { m_name = t; m_dirty = true; }
                Glib::ustring get_icao_name(void) const;
                const Glib::ustring& get_ident(void) const;
                const Glib::ustring& get_pathname(void) const { return m_pathname; }
                void set_pathname(const Glib::ustring& t) { m_pathname = t; m_dirty = true; }
		pathcode_t get_pathcode(void) const { return m_pathcode; }
                void set_pathcode(pathcode_t pc = pathcode_none) { m_pathcode = pc; m_dirty = true; }
		static const Glib::ustring& get_pathcode_name(pathcode_t pc);
		const Glib::ustring& get_pathcode_name(void) const { return get_pathcode_name(get_pathcode()); }
                const Glib::ustring& get_note(void) const { return m_note; }
                void set_note(const Glib::ustring& t) { m_note = t; m_dirty = true; }
                uint32_t get_time(void) const { return m_time; }
                void set_time(uint32_t t) { m_time = t; m_dirty = true; }
                time_t get_time_unix(void) const { return m_time - unix_minus_palmos_epoch; }
                void set_time_unix(time_t t) { m_time = t + unix_minus_palmos_epoch; m_dirty = true; }
                uint32_t get_save_time(void) const { return m_savetime; }
                void set_save_time(uint32_t t) { m_savetime = t; m_dirty = true; }
                time_t get_save_time_unix(void) const { return m_savetime - unix_minus_palmos_epoch; }
                void set_save_time_unix(time_t t) { m_savetime = t + unix_minus_palmos_epoch; m_dirty = true; }
                uint32_t get_frequency(void) const { return m_frequency; }
                void set_frequency(uint32_t f) { m_frequency = f; m_dirty = true; }
                Point get_coord(void) const { return m_coord; }
                Point::coord_t get_lon(void) const { return m_coord.get_lon(); }
                Point::coord_t get_lat(void) const { return m_coord.get_lat(); }
                void set_coord(const Point& c) { m_coord = c; m_dirty = true; }
                void set_lon(Point::coord_t lo) { m_coord.set_lon(lo); m_dirty = true; }
                void set_lat(Point::coord_t la) { m_coord.set_lat(la); m_dirty = true; }
                void set_lon_rad(float lo) { m_coord.set_lon_rad(lo); m_dirty = true; }
                void set_lat_rad(float la) { m_coord.set_lat_rad(la); m_dirty = true; }
                void set_lon_deg(float lo) { m_coord.set_lon_deg(lo); m_dirty = true; }
                void set_lat_deg(float la) { m_coord.set_lat_deg(la); m_dirty = true; }
                int16_t get_altitude(void) const { return m_alt; }
                void set_altitude(int16_t a) { m_alt = a; m_dirty = true; }
                uint16_t get_flags(void) const { return m_flags; }
                void set_flags(uint16_t f) { m_flags = f; m_dirty = true; }
                bool is_standard(void) const { return !!(get_flags() & altflag_standard); }
		bool is_ifr(void) const { return !!(get_flags() & altflag_ifr); }
		uint16_t get_winddir(void) const { return m_winddir; }
		float get_winddir_deg(void) const { return get_winddir() * to_deg; }
		float get_winddir_rad(void) const { return get_winddir() * to_rad; }
		void set_winddir(uint16_t dir) { m_winddir = dir; m_dirty = true; }
		void set_winddir_deg(float dir) { set_winddir(dir * from_deg); }
		void set_winddir_rad(float dir) { set_winddir(dir * from_rad); }
		uint16_t get_windspeed(void) const { return m_windspeed; }
		float get_windspeed_kts(void) const { return get_windspeed() * (1.f / 64.f); }
		void set_windspeed(uint16_t speed) { m_windspeed = speed; m_dirty = true; }
		void set_windspeed_kts(float speed) { set_windspeed(std::min(speed * 64.f, 65535.f)); }
		uint16_t get_qff(void) const { return m_qff; }
		float get_qff_hpa(void) const { return get_qff() * (1.f / 32.f); }
		void set_qff(uint16_t press) { m_qff = press; m_dirty = true; }
		void set_qff_hpa(float press) { set_qff(std::min(press * 32.f, 65535.f)); }
		int16_t get_isaoffset(void) const { return m_isaoffset; }
		float get_isaoffset_kelvin(void) const { return get_isaoffset() * (1.f / 256.f); }
		void set_isaoffset(int16_t temp) { m_isaoffset = temp; m_dirty = true; }
		void set_isaoffset_kelvin(float temp) { set_isaoffset(std::min(std::max(temp * 256.f, -32768.f), 32767.f)); }
		float get_sltemp_kelvin(void) const;
		void set_sltemp_kelvin(float temp);
		float get_oat_kelvin(void) const;
		void set_oat_kelvin(float temp);
		float get_pressure_altitude(void) const;
		float get_true_altitude(void) const;
		float get_density_altitude(void) const;

                void save_time(void) { m_savetime = m_time; m_dirty = true; }
                void restore_time(void) { m_time = m_savetime; m_dirty = true; }

		bool enforce_pathcode_vfrifr(void);

		static char get_ruleschar(uint16_t flags);
		char get_ruleschar(void) const { return get_ruleschar(get_flags()); }
		static char get_altchar(uint16_t flags);
		char get_altchar(void) const { return get_ruleschar(get_flags()); }

        private:
                Glib::ustring m_icao;
                Glib::ustring m_name;
                Glib::ustring m_pathname;
                Glib::ustring m_note;
                uint32_t m_time;
                uint32_t m_savetime;
                uint32_t m_frequency;
                Point m_coord;
                int16_t m_alt;
                uint16_t m_flags;
		uint16_t m_winddir;
		uint16_t m_windspeed;
		uint16_t m_qff;
		int16_t m_isaoffset;
		pathcode_t m_pathcode;
                bool m_dirty;

                friend class FPlan;
                friend class FPlanRoute;
};

class FPlanLeg {
        public:
                typedef enum {
                        legtype_invalid,
                        legtype_offblock_takeoff,
                        legtype_landing_onblock,
                        legtype_normal,
                } legtype_t;

                FPlanLeg(void);
                FPlanLeg(const Glib::ustring& fromicao, const Glib::ustring& fromname, const Point& fromcoord,
                         int16_t fromalt, uint16_t fromflags, uint32_t fromtime, uint32_t fromfreq,
                         const Glib::ustring& toicao, const Glib::ustring& toname, const Point& tocoord,
                         int16_t toalt, uint16_t toflags, uint32_t totime, uint32_t tofreq,
			 float winddir, float windspeed, float qnh, float isaoffset,
                         legtype_t legtype, bool future);

                void wind_correction(float vcruise_kts = 100.0);
		static int16_t baro_correction(float qnh, float isaoffset, int16_t alt, uint16_t altflags);
		void update(time_t curtime, const Point& coord = Point(), bool coordvalid = false, int16_t alt = 0, bool altvalid = false, float vgnd = 0.0 );

                const Glib::ustring& get_fromicao(void) const { return m_fromicao; }
                const Glib::ustring& get_fromname(void) const { return m_fromname; }
                Glib::ustring get_fromicaoname(void) const;
                const Glib::ustring& get_toicao(void) const { return m_toicao; }
                const Glib::ustring& get_toname(void) const { return m_toname; }
                Glib::ustring get_toicaoname(void) const;
                uint32_t get_fromtime(void) const { return m_fromtime; }
                time_t get_fromtime_unix(void) const { return m_fromtime - FPlanWaypoint::unix_minus_palmos_epoch; }
                uint32_t get_totime(void) const { return m_fromtime + m_legtime; }
                time_t get_totime_unix(void) const { return m_fromtime + m_legtime - FPlanWaypoint::unix_minus_palmos_epoch; }
                uint32_t get_legtime(void) const { return m_legtime; }
                Point get_fromcoord(void) const { return m_fromcoord; }
                Point get_tocoord(void) const { return m_tocoord; }
                int16_t get_fromaltitude(void) const { return m_fromalt; }
                int16_t get_toaltitude(void) const { return m_toalt; }
                uint16_t get_fromflags(void) const { return m_fromflags; }
                uint16_t get_toflags(void) const { return m_toflags; }
                int16_t get_fromtruealtitude(void) const { return m_fromtruealt; }
                int16_t get_totruealtitude(void) const { return m_totruealt; }
                uint32_t get_fromfrequency(void) const { return m_fromfreq; }
                uint32_t get_tofrequency(void) const { return m_tofreq; }

                float get_dist(void) const { return m_dist; }
                float get_vdist(void) const { return m_vdist; }
                float get_truetrack(void) const { return m_truetrack * to_deg; }
                float get_variance(void) const { return m_var * to_deg; }
                float get_wca(void) const { return m_wca * to_deg; }
                float get_heading(void) const { return (m_truetrack + m_var + m_wca) * to_deg; }

                bool is_future(void) const { return m_future; }
                legtype_t get_legtype(void) const { return m_legtype; }

                float get_crosstrack(void) const { return m_crosstrack; }
                float get_disttodest(void) const { return m_disttodest; }
                float get_legfraction(void) const { return m_legfraction * (1.0 / (1U << 15)); }
                int16_t get_altdiff(void) const { return m_altdiff; }
                int16_t get_referencealt(void) const { return m_referencealt; }
                int16_t get_mapalt(void) const { return m_mapalt; }
                float get_coursetodest(void) const { return m_coursetodest * to_deg; }
                Point get_mapcoord(void) const { return m_mapcoord; }
                bool is_gpsvalid(void) const { return m_gpsvalid; }
                bool is_gpsaltvalid(void) const { return m_gpsaltvalid; }
                float get_vgnd(void) const { return m_vgnd; }
                float get_verticalspeed(void) const;
                float get_verticalspeedtodest(void) const;

                static const float from_rad = (1 << 15) / M_PI;
                static const float to_rad = M_PI / (1 << 15);
                static const float from_deg = (1 << 15) / 180.0;
                static const float to_deg = 180.0 / (1 << 15);

        private:
                // Static Data
                Glib::ustring m_fromicao, m_toicao;
                Glib::ustring m_fromname, m_toname;
                Point m_fromcoord, m_tocoord;
                int16_t m_fromalt, m_toalt;
                int16_t m_fromtruealt, m_totruealt;
                uint16_t m_fromflags, m_toflags;
                uint32_t m_fromfreq, m_tofreq;
                uint32_t m_fromtime, m_legtime;
		float m_winddir;
		float m_windspeed;
                float m_dist;
                float m_vdist; /* speed (cruise+wind) projected onto the line connecting the two endpoints */
                uint16_t m_truetrack;  /* 2^16 = 360deg = 2pi rad */
                uint16_t m_var;  /* 2^16 = 360deg = 2pi rad */
                uint16_t m_wca;  /* wind correction angle, 2^16 = 360deg = 2pi rad */
                legtype_t m_legtype;
                bool m_future;
                // Dynamic Data
                float m_crosstrack;
                float m_disttodest;
                float m_vgnd;
                uint16_t m_legfraction;
                int16_t m_altdiff;
                int16_t m_referencealt;
                int16_t m_mapalt;
                Point m_mapcoord;
                uint16_t m_coursetodest;
                bool m_gpsvalid;
                bool m_gpsaltvalid;

		void wind_correction(float vcruise_kts, float winddir_deg, float vwind_kts);
                void baro_correction(float qnh, float isaoffset);
};

class FPlanRoute {
        public:
                typedef int32_t id_t;
                FPlanRoute(FPlan& fpp);

                id_t get_id(void) const { return m_id; }
                const Glib::ustring& get_note(void) const { return m_note; }
                void set_note(const Glib::ustring& t) { m_note = t; m_dirty = true; }
                uint16_t get_nrwpt(void) const { return m_wpts.size(); }
                uint16_t get_curwpt(void) const { return m_curwpt; }
                void set_curwpt(uint16_t cw) { m_curwpt = cw; m_dirty = true; }
                FPlanWaypoint& operator[](uint32_t nr);
                const FPlanWaypoint& operator[](uint32_t nr) const;
                FPlanWaypoint& get_wpt(uint32_t nr) { return operator[](nr); }
                const FPlanWaypoint& get_wpt(uint32_t nr) const { return operator[](nr); }
                FPlanWaypoint& get_wpt(void) { return get_wpt(m_curwpt); }
                const FPlanWaypoint& get_wpt(void) const { return get_wpt(m_curwpt); }
                void insert_wpt(uint32_t nr = ~0, const FPlanWaypoint& wpt = FPlanWaypoint());
                void delete_wpt(uint32_t nr);
                uint32_t get_time_offblock(void) const { return m_time_offblock; }
                void set_time_offblock(uint32_t t) { m_time_offblock = t; m_dirty = true; }
                time_t get_time_offblock_unix(void) const { return m_time_offblock - FPlanWaypoint::unix_minus_palmos_epoch; }
                void set_time_offblock_unix(time_t t) { m_time_offblock = t + FPlanWaypoint::unix_minus_palmos_epoch; m_dirty = true; }
                uint32_t get_save_time_offblock(void) const { return m_savetime_offblock; }
                void set_save_time_offblock(uint32_t t) { m_savetime_offblock = t; m_dirty = true; }
                time_t get_save_time_offblock_unix(void) const { return m_savetime_offblock - FPlanWaypoint::unix_minus_palmos_epoch; }
                void set_save_time_offblock_unix(time_t t) { m_savetime_offblock = t + FPlanWaypoint::unix_minus_palmos_epoch; m_dirty = true; }
                uint32_t get_time_onblock(void) const { return m_time_onblock; }
                void set_time_onblock(uint32_t t) { m_time_onblock = t; m_dirty = true; }
                time_t get_time_onblock_unix(void) const { return m_time_onblock - FPlanWaypoint::unix_minus_palmos_epoch; }
                void set_time_onblock_unix(time_t t) { m_time_onblock = t + FPlanWaypoint::unix_minus_palmos_epoch; m_dirty = true; }
                uint32_t get_save_time_onblock(void) const { return m_savetime_onblock; }
                void set_save_time_onblock(uint32_t t) { m_savetime_onblock = t; m_dirty = true; }
                time_t get_save_time_onblock_unix(void) const { return m_savetime_onblock - FPlanWaypoint::unix_minus_palmos_epoch; }
                void set_save_time_onblock_unix(time_t t) { m_savetime_onblock = t + FPlanWaypoint::unix_minus_palmos_epoch; m_dirty = true; }

                void save_time_offblock(void) { m_savetime_offblock = m_time_offblock; m_dirty = true; }
                void restore_time_offblock(void) { m_time_offblock = m_savetime_offblock; m_dirty = true; }
                void save_time_onblock(void) { m_savetime_onblock = m_time_onblock; m_dirty = true; }
                void restore_time_onblock(void) { m_time_onblock = m_savetime_onblock; m_dirty = true; }

                void save_time(uint32_t nr);
                void restore_time(uint32_t nr);
                void set_time(uint32_t nr, uint32_t t);
                void set_time_unix(uint32_t nr, time_t t) { set_time(nr, t + FPlanWaypoint::unix_minus_palmos_epoch); }
                uint32_t get_time(uint32_t nr) const;
                time_t get_time_unix(uint32_t nr) const { return get_time(nr) - FPlanWaypoint::unix_minus_palmos_epoch; }

                FPlanLeg get_leg(uint32_t nr);
                FPlanLeg get_directleg(void);

                void load_fp(void);
                void load_fp(id_t nid);
                void save_fp(void);
                void new_fp(void);
                void duplicate_fp(void);
                void delete_fp(void);
                void reverse_fp(void);
                bool load_first_fp(void);
                bool load_next_fp(void);

                float total_distance_km(void) const;
                float total_distance_nmi(void) const { return total_distance_km() * Point::km_to_nmi; }
                double total_distance_km_dbl(void) const;
                double total_distance_nmi_dbl(void) const { return total_distance_km_dbl() * Point::km_to_nmi_dbl; }
                int16_t max_altitude(void) const;
                int16_t max_altitude(uint16_t& flags) const;
		Rect get_bbox(void) const;
		char get_flightrules(void) const;

		bool has_pathcodes(void) const;
		bool enforce_pathcode_vfrdeparr(void);
		bool enforce_pathcode_vfrifr(void);

                bool is_dirty(void) const;

        private:
                void set_dirty(bool d = true);

        private:
                FPlan *m_fp;
                id_t m_id;
                Glib::ustring m_note;
                uint32_t m_time_offblock;
                uint32_t m_savetime_offblock;
                uint32_t m_time_onblock;
                uint32_t m_savetime_onblock;
                uint32_t m_curwpt;
                typedef std::vector<FPlanWaypoint> waypoints_t;
                waypoints_t m_wpts;
                bool m_dirty;

                friend class FPlan;
};

class FPlan {
        public:
                typedef enum {
                        comp_startswith,
                        comp_contains,
                        comp_exact
                } comp_t;

                FPlan(void);
                FPlan(sqlite3 *dbh);
                ~FPlan() throw ();
                static Glib::ustring get_userdbpath(void);

                void delete_database(void);

                std::vector<FPlanWaypoint> find_by_name(const Glib::ustring& nm = "", unsigned int limit = ~0, unsigned int skip = 0, comp_t comp = comp_contains);
                std::vector<FPlanWaypoint> find_by_icao(const Glib::ustring& ic = "", unsigned int limit = ~0, unsigned int skip = 0, comp_t comp = comp_contains);
                std::vector<FPlanWaypoint> find_nearest(const Point& pt, unsigned int limit = ~0, unsigned int skip = 0, const Rect& rect = Rect());

        private:
                sqlite3x::sqlite3_connection db;

                void init_database(void);
                sqlite3x::sqlite3_connection& get_db(void) { return db; }

                static std::string escape_string(const std::string& s);

                static void dbfunc_simpledist(sqlite3_context *ctxt, int, sqlite3_value **values);

                friend class FPlanRoute;
                friend class FPlanWaypoint;
};

#endif /* FPLAN_H */
