/*****************************************************************************/

/*
 *      navlog.cc  --  Navigation Log Creation.
 *
 *      Copyright (C) 2012, 2013
 *          Thomas Sailer (t.sailer@alumni.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*****************************************************************************/

#include "aircraft.h"
#include "fplan.h"
#include "airdata.h"
#include "engine.h"
#include "baro.h"
#include "wmm.h"
#include "wind.h"
#include "icaofpl.h"

#include <cmath>
#include <stdexcept>
#include <limits>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <ext/stdio_filebuf.h>

#ifdef __WIN32__
#include <windows.h>
#endif

class Aircraft::NavFPlan {
public:
	NavFPlan(const Aircraft& acft, Engine& engine, double cruise_rpm, double cruise_mp);
	~NavFPlan();
	void begin(const WeightBalance::elementvalues_t& wbev = WeightBalance::elementvalues_t());
	void fpl(const FPlanRoute& fpl);
	void profile(const FPlanRoute& fpl);
	void end(const Glib::ustring& templatefile);

protected:
        class FPlanWptExtra {
          public:
                FPlanWptExtra(time_t ft = 0, double truealt = 0, double tt = 0, double th = 0,
                              double decl = 0, double dist = 0, double fuel = 0, double tas = 0, double gs = 0,
                              double rpm = std::numeric_limits<double>::quiet_NaN(),
                              double mp = std::numeric_limits<double>::quiet_NaN());

                time_t get_flighttime(void) const { return m_flighttime; }
                double get_truealt(void) const { return m_truealt; }
                double get_truetrack(void) const { return m_tt; }
                double get_trueheading(void) const { return m_th; }
                double get_declination(void) const { return m_decl; }
                double get_magnetictrack(void) const;
                double get_magneticheading(void) const;
                double get_dist(void) const { return m_dist; }
                double get_fuel(void) const { return m_fuel; }
                double get_tas(void) const { return m_tas; }
                double get_rpm(void) const { return m_rpm; }
                double get_mp(void) const { return m_mp; }
                double get_gs(void) const { return m_gs; }

		void set_flighttime(time_t ft) { m_flighttime = ft; }
                void set_truealt(double x) { m_truealt = x; }
                void set_truetrack(double tt) { m_tt = tt; }
                void set_trueheading(double th) { m_th = th; }
                void set_declination(double decl) { m_decl = decl; }
                void set_dist(double d) { m_dist = d; }
                void set_fuel(double x) { m_fuel = x; }
                void set_tas(double x) { m_tas = x; }
                void set_rpm(double x) { m_rpm = x; }
                void set_mp(double x) { m_mp = x; }
                void set_gs(double x) { m_gs = x; }

          protected:
                time_t m_flighttime;
                double m_truealt;
                double m_tt;
                double m_th;
                double m_decl;
                double m_dist;
                double m_fuel;
                double m_tas;
                double m_rpm;
                double m_mp;
                double m_gs;
        };

	class Profile {
	  public:
		Profile(Engine& engine);
		std::ostream& profile(std::ostream& os, const FPlanRoute& fpl, double corridor_width = 5);

	  protected:
		TopoDb30::Profile m_profile;
		Glib::RefPtr<Engine::ElevationProfileResult> m_topoquery;
		Glib::Cond m_cond;
		Glib::Mutex m_mutex;
		Engine& m_engine;

		void async_cancel(void);
		void async_clear(void);
		void async_connectdone(void);
		void async_done(void);
		void async_waitdone(void);
	};


	const Aircraft& m_aircraft;
	Engine& m_engine;
	__gnu_cxx::stdio_filebuf<char> *m_tempfile;
	std::string m_tempfilename;
        typedef std::vector<FPlanWptExtra> routeextra_t;
        routeextra_t m_routeextra;
	double m_cruise_rpm;
	double m_cruise_mp;

	void recompute_wptextra(const FPlanRoute& m_route);
};

Aircraft::NavFPlan::FPlanWptExtra::FPlanWptExtra(time_t ft, double truealt, double tt, double th, double decl,
                                      double dist, double fuel, double tas, double gs, double rpm, double mp)
        : m_flighttime(ft), m_truealt(truealt), m_tt(tt), m_th(th), m_decl(decl), m_dist(dist), m_fuel(fuel), m_tas(tas), m_rpm(rpm), m_mp(mp), m_gs(gs)
{
}

double Aircraft::NavFPlan::FPlanWptExtra::get_magnetictrack(void) const
{
        double x(get_truetrack() - get_declination());
        if (x < 0)
                x += 360;
        if (x >= 360)
                x -= 360;
        return x;
}

double Aircraft::NavFPlan::FPlanWptExtra::get_magneticheading(void) const
{
        double x(get_trueheading() - get_declination());
        if (x < 0)
                x += 360;
        if (x >= 360)
                x -= 360;
        return x;
}

Aircraft::NavFPlan::Profile::Profile(Engine& engine)
	: m_engine(engine)
{
}

std::ostream& Aircraft::NavFPlan::Profile::profile(std::ostream& os, const FPlanRoute& fpl, double corridor_width)
{
	double dist(0);
	unsigned int nrwpt(fpl.get_nrwpt());
	for (unsigned int i = 1; i < nrwpt;) {
		const FPlanWaypoint& wpt0(fpl[i - 1]);
		const FPlanWaypoint& wpt1(fpl[i]);
		++i;
		bool lastleg(i >= nrwpt);
		async_cancel();
		async_clear();
		m_topoquery = m_engine.async_elevation_profile(wpt0.get_coord(), wpt1.get_coord(), corridor_width);
		async_connectdone();
		async_waitdone();
		unsigned int profsize(m_profile.size());
		for (unsigned int j = 0; j < profsize;) {
			const TopoDb30::ProfilePoint& pp(m_profile[j]);
			TopoDb30::elev_t el(pp.get_elev()), maxel(pp.get_maxelev());
			if (el == TopoDb30::nodata) {
				++j;
				continue;
			}
			if (el == TopoDb30::ocean)
				el = 0;
			if (maxel == TopoDb30::nodata)
				maxel = el;
			else if (maxel == TopoDb30::ocean)
				maxel = 0;
			os << "VPROF|";
			++j;
			if (j == 1)
				os << wpt0.get_icao_name() << '|' << wpt0.get_true_altitude();
			else if (lastleg && j >= profsize)
				os << wpt1.get_icao_name() << '|' << wpt1.get_true_altitude();
			else
				os << '|';
			os << '|' << (dist + pp.get_dist()) << '|' << el << '|' << maxel << std::endl;
		}
		dist += wpt0.get_coord().spheric_distance_nmi_dbl(wpt1.get_coord());
	}
	return os;
}

void Aircraft::NavFPlan::Profile::async_cancel(void)
{
        Engine::ElevationProfileResult::cancel(m_topoquery);
}

void Aircraft::NavFPlan::Profile::async_clear(void)
{
	m_topoquery = Glib::RefPtr<Engine::ElevationProfileResult>();
	m_profile.clear();
}

void Aircraft::NavFPlan::Profile::async_connectdone(void)
{
	if (m_topoquery)
		m_topoquery->connect(sigc::mem_fun(*this, &Profile::async_done));
}

void Aircraft::NavFPlan::Profile::async_done(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        m_cond.broadcast();
}

void Aircraft::NavFPlan::Profile::async_waitdone(void)
{
        Glib::Mutex::Lock lock(m_mutex);
        for (;;) {
		if (m_topoquery && m_topoquery->is_done()) {
			m_profile.clear();
			if (!m_topoquery->is_error())
				m_profile = m_topoquery->get_result();
			m_topoquery = Glib::RefPtr<Engine::ElevationProfileResult>();
		}
		if (!m_topoquery)
			break;
		m_cond.wait(m_mutex);
	}
}

Aircraft::NavFPlan::NavFPlan(const Aircraft& acft, Engine& engine, double cruise_rpm, double cruise_mp)
	: m_aircraft(acft), m_engine(engine), m_tempfile(0), m_cruise_rpm(cruise_rpm), m_cruise_mp(cruise_mp)
{
}

Aircraft::NavFPlan::~NavFPlan()
{
	if (m_tempfile) {
		m_tempfile->close();
		delete m_tempfile;
	}
}

void Aircraft::NavFPlan::begin(const WeightBalance::elementvalues_t& wbev)
{	
	if (m_tempfile) {
		m_tempfile->close();
		delete m_tempfile;
		m_tempfile = 0;
	}
	int fd(Glib::file_open_tmp(m_tempfilename, "vfrnav_navlog_"));
	m_tempfile = new __gnu_cxx::stdio_filebuf<char>(fd, std::ios::out);
	std::ostream os(m_tempfile);
	os << "ACFT|" << m_aircraft.get_callsign()
	   << '|' << m_aircraft.get_aircrafttype()
	   << '|' << m_aircraft.get_icaotype()
	   << '|' << m_aircraft.get_equipment()
	   << '|' << m_aircraft.get_transponder()
	   << '|' << m_aircraft.get_pbn()
	   << '|' << m_aircraft.get_colormarking()
	   << '|' << m_aircraft.get_wakecategory() << std::endl
	   << "ACFTD|" << m_aircraft.get_mtom()
	   << '|' << m_aircraft.get_vr0()
	   << '|' << m_aircraft.get_vr1()
	   << '|' << m_aircraft.get_vx0()
	   << '|' << m_aircraft.get_vx1()
	   << '|' << m_aircraft.get_vy()
	   << '|' << m_aircraft.get_vs0()
	   << '|' << m_aircraft.get_vs1()
	   << '|' << m_aircraft.get_vno()
	   << '|' << m_aircraft.get_vne()
	   << '|' << m_aircraft.get_vref0()
	   << '|' << m_aircraft.get_vref1()
	   << '|' << m_aircraft.get_va()
	   << '|' << m_aircraft.get_vfe()
	   << '|' << m_aircraft.get_vgearext()
	   << '|' << m_aircraft.get_vgearret()
	   << '|' << m_aircraft.get_vbestglide()
	   << '|' << m_aircraft.get_glideslope()
	   << '|' << m_aircraft.get_fuelmass()
	   << '|' << m_aircraft.get_maxbhp() << std::endl;
	{
		const Aircraft::WeightBalance& wb(m_aircraft.get_wb());
		for (unsigned int i = 0; i < wb.get_nrelements(); ++i) {
			const Aircraft::WeightBalance::Element& el(wb.get_element(i));
			os << "WBEL|" << el.get_name();
			if (i >= wbev.size())
				os << "|0|0";
			else
				os << '|' << wbev[i].get_value() << '|' << wbev[i].get_unit();
			os << '|' << el.get_massmin()
			   << '|' << el.get_massmax()
			   << '|' << el.get_arm() << '|';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_fixed)
				os << 'F';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_binary)
				os << 'B';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_avgas)
				os << 'A';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_oil)
				os << 'O';
			if (el.get_flags() & Aircraft::WeightBalance::Element::flag_gear)
				os << 'G';
			for (unsigned int j = 0; j < el.get_nrunits(); ++j) {
				const Aircraft::WeightBalance::Element::Unit& u(el.get_unit(j));
				os << '|' << u.get_name() 
				   << '|' << u.get_factor()
				   << '|' << u.get_offset();
			}
			os << std::endl;
		}
		for (unsigned int i = 0; i < wb.get_nrenvelopes(); ++i) {
			const Aircraft::WeightBalance::Envelope& env(wb.get_envelope(i));
			os << "WBENV|" << env.get_name();
			for (unsigned int j = 0; j < env.size(); ++j) {
				const Aircraft::WeightBalance::Envelope::Point& pt(env[j]);
				os << '|' << pt.get_mass() << '|' << pt.get_arm();
			}
			os << std::endl;
		}
	}
	{
		const Aircraft::Distances& dists(m_aircraft.get_dist());
		for (unsigned int i = 0; i < dists.get_nrtakeoffdist(); ++i) {
			const Aircraft::Distances::Distance& dist(dists.get_takeoffdist(i));
			os << "TKOFFDIST|" << dist.get_name()
			   << '|' << dist.get_vrotate()
			   << '|' << dist.get_mass() << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_gndpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_obstpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << std::endl;
		}
		for (unsigned int i = 0; i < dists.get_nrldgdist(); ++i) {
			const Aircraft::Distances::Distance& dist(dists.get_ldgdist(i));
			os << "LDGDIST|" << dist.get_name()
			   << '|' << dist.get_vrotate()
			   << '|' << dist.get_mass() << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_gndpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << '|';
			{
				const Aircraft::Poly1D<double>& p(dist.get_obstpoly());
				bool subseq(false);
				for (Aircraft::Poly1D<double>::const_iterator pi(p.begin()), pe(p.end()); pi != pe; ++pi) {
					if (subseq)
						os << ',';
					subseq = true;
					os << *pi;
				}
			}
			os << std::endl;
		}
	}
}

void Aircraft::NavFPlan::fpl(const FPlanRoute& fpl)
{
	if (!m_tempfile || fpl.get_nrwpt() < 1)
		return;
	recompute_wptextra(fpl);
	std::ostream os(m_tempfile);
	{
		IcaoFlightPlan icaofpl(m_engine);
		icaofpl.set_aircraft(m_aircraft, m_cruise_rpm, m_cruise_mp);
		icaofpl.populate(fpl, false);
		os << "FPLH|" << fpl.get_time_offblock_unix()
		   << '|' << fpl.get_time_onblock_unix()
		   << '|' << icaofpl.get_fpl() << std::endl;
	}
	for (unsigned int i = 0; i < 2; ++i) {
		const FPlanWaypoint& wpt(fpl[(fpl.get_nrwpt() - 1) & -i]);
		{
			double tas(0), fuelflow(0), bhp(0), rpm(m_cruise_rpm), mp(m_cruise_mp), da(wpt.get_density_altitude() + 1500.0);
			m_aircraft.get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
			os << (i ? "FPLDESTHLD|" : "FPLDEPHLD|") << wpt.get_density_altitude()
			   << '|' << fuelflow << '|' << tas << '|' << bhp << std::endl;
		}
		IcaoFlightPlan::FindCoord fc(m_engine);
		if (!fc.find(wpt.get_icao(), wpt.get_name(), IcaoFlightPlan::FindCoord::flag_subtables | IcaoFlightPlan::FindCoord::flag_airport, wpt.get_coord()))
			continue;
		AirportsDb::Airport arpt;
		if (!fc.get_airport(arpt))
			continue;
		if (!arpt.is_valid())
			continue;
		for (unsigned int j = 0; j < arpt.get_nrrwy(); ++j) {
			const AirportsDb::Airport::Runway& rwy(arpt.get_rwy(j));
			os << (i ? "FPLDESTRWY" : "FPLDEPRWY")
			   << '|' << rwy.get_ident_he()
			   << '|' << rwy.get_ident_le()
			   << '|' << rwy.get_length()
			   << '|' << rwy.get_width()
			   << '|' << rwy.get_surface()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_he_coord().get_lat_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_he_coord().get_lon_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_le_coord().get_lat_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << rwy.get_le_coord().get_lon_deg_dbl()
			   << '|' << rwy.get_he_tda()
			   << '|' << rwy.get_he_lda()
			   << '|' << rwy.get_he_disp()
			   << '|' << rwy.get_he_hdg()
			   << '|' << rwy.get_he_elev()
			   << '|' << rwy.get_le_tda()
			   << '|' << rwy.get_le_lda()
			   << '|' << rwy.get_le_disp()
			   << '|' << rwy.get_le_hdg()
			   << '|' << rwy.get_le_elev()
			   << '|' << (int)rwy.is_he_usable()
			   << '|' << (int)rwy.is_le_usable();
			for (unsigned int k = 0; k < 8; ++k)
				os << '|' << AirportsDb::Airport::Runway::get_light_string(rwy.get_he_light(k));
			for (unsigned int k = 0; k < 8; ++k)
				os << '|' << AirportsDb::Airport::Runway::get_light_string(rwy.get_le_light(k));
			os << std::endl;
		}
		for (unsigned int j = 0; j < arpt.get_nrhelipads(); ++j) {
			const AirportsDb::Airport::Helipad& hp(arpt.get_helipad(j));
			os << (i ? "FPLDESTHP" : "FPLDEPHP")
			   << '|' << hp.get_ident()
			   << '|' << hp.get_length()
			   << '|' << hp.get_width()
			   << '|' << hp.get_surface()
			   << '|' << std::fixed << std::setprecision(9) << hp.get_coord().get_lat_deg_dbl()
			   << '|' << std::fixed << std::setprecision(9) << hp.get_coord().get_lon_deg_dbl()
			   << '|' << hp.get_hdg()
			   << '|' << hp.get_elev()
			   << '|' << (int)hp.is_usable() << std::endl;
		}
		for (unsigned int j = 0; j < arpt.get_nrcomm(); ++j) {
			const AirportsDb::Airport::Comm& comm(arpt.get_comm(j));
			std::string sector(comm.get_sector());
			std::string oprhours(comm.get_oprhours());
			for (std::string::iterator si(sector.begin()), se(sector.end()); si != se; ++si)
				if (*si == '\n' || *si == '\r')
					*si = ' ';
			for (std::string::iterator si(oprhours.begin()), se(oprhours.end()); si != se; ++si)
				if (*si == '\n' || *si == '\r')
					*si = ' ';
			os << (i ? "FPLDESTCOMM" : "FPLDEPCOMM")
			   << '|' << comm.get_name()
			   << '|' << sector
			   << '|' << oprhours;
			for (unsigned int k = 0; k < 5; ++k)
				os << '|' << comm[k];
			os << std::endl;
		}
	}
        for (unsigned int i = 0; i < m_routeextra.size(); ++i) {
                const FPlanWaypoint& wpt(fpl[i]);
                const FPlanWptExtra& wpte(m_routeextra[i]);
		os << "FPLL|" << wpt.get_icao() << '|' << wpt.get_name() << '|' << wpt.get_pathname()
		   << '|' << wpt.get_pathcode_name() << '|';
		if (!i)
			os << (wpt.get_time_unix() - fpl.get_time_offblock_unix());
		else
			os << (wpte.get_flighttime() - m_routeextra[i - 1].get_flighttime());
		os << '|' << wpt.get_frequency()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_coord().get_lat_deg_dbl()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_coord().get_lon_deg_dbl()
		   << '|' << wpt.get_altitude()
		   << '|' << ((wpt.get_flags() & FPlanWaypoint::altflag_standard) ? "STD" : "QNH")
		   << '|' << ((wpt.get_flags() & FPlanWaypoint::altflag_ifr) ? "IFR" : "VFR")
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_truealt()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_truetrack()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_trueheading()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_declination()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_magnetictrack()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_magneticheading()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_dist()
		   << '|';
		if (i)
			os << std::fixed << std::setprecision(9) << (wpte.get_fuel() - m_routeextra[i - 1].get_fuel());
		else
			os << '0';
		os << '|' << std::fixed << std::setprecision(9) << wpte.get_tas()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_gs()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_rpm()
		   << '|' << std::fixed << std::setprecision(9) << wpte.get_mp()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_winddir_deg()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_windspeed_kts()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_qff_hpa()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_isaoffset_kelvin()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_oat_kelvin()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_pressure_altitude()
		   << '|' << std::fixed << std::setprecision(9) << wpt.get_density_altitude()
		   << std::endl;
	}
}

void Aircraft::NavFPlan::profile(const FPlanRoute& fpl)
{
	if (!m_tempfile || fpl.get_nrwpt() < 2)
		return;
	std::ostream os(m_tempfile);
	Profile p(m_engine);
	p.profile(os, fpl);
}

void Aircraft::NavFPlan::end(const Glib::ustring& templatefile)
{
	if (m_tempfile) {
		m_tempfile->close();
		delete m_tempfile;
		m_tempfile = 0;
	}
	std::vector<std::string> argv;
	std::vector<std::string> env(Glib::listenv());
	{
		bool needfile(true);
		for (std::vector<std::string>::iterator i(env.begin()), e(env.end()); i != e; ++i) {
			if (*i == "VFRNAV_FPLFILE") {
				*i += "=" + m_tempfilename;
				needfile = false;
				continue;
			}
			*i += "=" + Glib::getenv(*i);
		}
		if (needfile)
			env.push_back("VFRNAV_FPLFILE=" + m_tempfilename);
	}
	if (false) {
		std::cerr << "Environment:" << std::endl;
		for (std::vector<std::string>::const_iterator i(env.begin()), e(env.end()); i != e; ++i)
			std::cerr << *i << std::endl;
	}
#ifdef __WIN32__
	{
		std::string version;
		std::string path;
		static const char * const regkeys[] = {
			"Software\\LibreOffice\\LibreOffice",
			"Software\\OpenOffice.org\\OpenOffice.org",
			0
		};
		for (const char * const *regkey = regkeys; *regkey; ++regkey) {
			HKEY hKey;
			if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, *regkey, 0, KEY_READ, &hKey) == ERROR_SUCCESS) {
				if (false)
					std::cerr << "Enumerating keys in HKLM\\" << *regkey << std::endl;
				for (DWORD dwIndex = 0;; ++dwIndex) {
					TCHAR lpName[1024];
					DWORD lpcbName(sizeof(lpName));
					FILETIME fTime;
					if (RegEnumKeyEx(hKey, dwIndex, lpName, &lpcbName, NULL, NULL, NULL, &fTime) != ERROR_SUCCESS)
						break;
					std::string ver1(lpName);
					if (false)
						std::cerr << "Subkey " << dwIndex << ": " << ver1 << std::endl;
					if (!path.empty() && ver1 < version)
						continue;
					//TCHAR lpData[1024];
					BYTE lpData[1024];
					DWORD cbData(sizeof(lpData));
					//LONG err(RegGetValue(hKey, lpName, "Path", RRF_RT_REG_SZ, 0, &lpData, &cbData));
					HKEY hKey2;
					LONG err(RegOpenKeyEx(hKey, lpName, 0, KEY_READ, &hKey2));
					if (err != ERROR_SUCCESS) {
						if (true)
							std::cerr << "Cannot open key HKLM\\" << *regkey << "\\" << lpName
								  << " Error " << g_win32_error_message(err) << " (" << err << ')' << std::endl;
						continue;
					}
					DWORD lpType;
					err = RegQueryValueEx(hKey2, "Path", 0, &lpType, lpData, &cbData);
					if (cbData < sizeof(lpData))
						cbData[lpData] = 0;
					RegCloseKey(hKey2);
					if (err != ERROR_SUCCESS) {
						if (true)
							std::cerr << "Cannot get key HKLM\\" << *regkey << "\\" << lpName
								  << " value: Error " << g_win32_error_message(err) << " (" << err << ')' << std::endl;
						continue;
					}
					version = ver1;
					path = (char *)lpData;
					
				}
				RegCloseKey(hKey);
			}
		}
		if (!path.empty()) {
			std::cerr << "Found libreoffice V" << version << ": " << path << std::endl;
			argv.push_back(path);
		} else {
			std::cerr << "Libreoffice not found; trying soffice.exe" << std::endl;
			argv.push_back("soffice.exe");
		}
	}
#else
	argv.push_back("soffice");
#endif
	argv.push_back("--calc");
	argv.push_back("--norestore");
	argv.push_back("--nologo");
	argv.push_back("-n");
	argv.push_back(templatefile);
	argv.push_back("macro://./Standard.FPLImport.Main()");
	Glib::spawn_async_with_pipes(Glib::get_home_dir(), argv, env, Glib::SPAWN_SEARCH_PATH);
}

void Aircraft::NavFPlan::recompute_wptextra(const FPlanRoute& m_route)
{
        m_routeextra.resize(m_route.get_nrwpt());
        if (m_routeextra.empty())
                return;
        AirData ad;
        WMM wmm;
        time_t curtime(time(0));
        {
                const FPlanWaypoint& wpt(m_route[0]);
                FPlanWptExtra& wpte(m_routeextra[0]);
                wpte.set_flighttime(0);
                wpte.set_fuel(0);
                wpte.set_rpm(0);
                wpte.set_mp(0);
		ad.set_qnh_tempoffs(wpt.get_qff_hpa(), wpt.get_isaoffset_kelvin());
                if (wpt.get_flags() & FPlanWaypoint::altflag_standard)
                        ad.set_pressure_altitude(wpt.get_altitude());
                else
                        ad.set_true_altitude(wpt.get_altitude());
                wpte.set_truealt(ad.get_true_altitude());
                {
                        double tm(m_aircraft.get_climb().altitude_to_time(ad.get_density_altitude()));
                        wpte.set_tas(m_aircraft.get_climb().time_to_tas(tm));
                }
                wmm.compute(ad.get_true_altitude() * (FPlanWaypoint::ft_to_m / 1000.0), wpt.get_coord(), curtime);
                wpte.set_truetrack(0);
                wpte.set_trueheading(0);
                wpte.set_declination(wmm.get_dec());
                wpte.set_dist(0);
        }
        for (unsigned int i = 1; i < m_routeextra.size(); ++i) {
                const FPlanWaypoint& wpt(m_route[i]);
                FPlanWptExtra& wpte(m_routeextra[i]);
                const FPlanWaypoint& wpto(m_route[i - 1]);
                const FPlanWptExtra& wpteo(m_routeextra[i - 1]);
                wpte.set_tas(0);
                wpte.set_rpm(0);
                wpte.set_mp(0);
		ad.set_qnh_tempoffs(wpt.get_qff_hpa(), wpt.get_isaoffset_kelvin());
                double dist(wpto.get_coord().spheric_distance_nmi_dbl(wpt.get_coord()));
                double tt(wpto.get_coord().spheric_true_course(wpt.get_coord()));
                wpte.set_dist(dist);
                wpte.set_truetrack(tt);
                double prevda(ad.get_density_altitude());
                double prevta(ad.get_true_altitude());
                if (wpt.get_flags() & FPlanWaypoint::altflag_standard)
                        ad.set_pressure_altitude(wpt.get_altitude());
                else
                        ad.set_true_altitude(wpt.get_altitude());
                double nextda(ad.get_density_altitude());
                double ltime(wpteo.get_flighttime()), fuel(wpteo.get_fuel());
                double th(tt);
                Wind<double> wind;
                wind.set_wind(wpto.get_winddir_deg(), wpto.get_windspeed_kts());
                // check for climb segment
                bool climbseg(nextda - prevda > 100);
                if (climbseg) {
                        double prevt(m_aircraft.get_climb().altitude_to_time(prevda));
                        double nextt(m_aircraft.get_climb().altitude_to_time(nextda));
                        double prevd(m_aircraft.get_climb().time_to_distance(prevt));
                        double nextd(m_aircraft.get_climb().time_to_distance(nextt));
                        double climbd(nextd - prevd);
                        // check for wind
                        double speedfactor(1);
			wpte.set_gs(m_aircraft.get_climb().time_to_tas(prevt));
                        if (wind.get_windspeed() > 0) {
                                double tas(m_aircraft.get_climb().time_to_tas(prevt));
                                wind.set_crs_tas(tt, tas);
                                if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0) {
                                        climbd = 0;
                                } else {
                                        th = wind.get_hdg();
                                        speedfactor = wind.get_gs() / tas;
                                        climbd *= speedfactor;
					wpte.set_gs(wind.get_gs());
                                        if (false)
                                                std::cerr << "recompute_wptextra: climb: windcorr: TT " << tt << " TH " << th
                                                          << " TAS " << tas << " GS " << wind.get_gs() << " Wind " << wind.get_winddir() << '/' << wind.get_windspeed() << std::endl;
                                }
                        }
                        if (climbd <= dist) {
                                dist -= climbd;
                        } else {
                                // climb not finished at waypoint
                                nextt = m_aircraft.get_climb().distance_to_time(prevd + dist / speedfactor);
                                dist = 0;
                                ad.set_density_altitude(m_aircraft.get_climb().time_to_altitude(nextt));
                        }
                        ltime += nextt - prevt;
                        fuel += m_aircraft.get_climb().time_to_fuel(nextt) -
                                m_aircraft.get_climb().time_to_fuel(prevt);
                        wpte.set_tas(m_aircraft.get_climb().time_to_tas(nextt));
                }
                // check for cruise segment
                if (dist > 0) {
                        double tas(0), fuelflow(0), bhp(0), rpm(m_cruise_rpm), mp(m_cruise_mp), da(ad.get_density_altitude());
                        m_aircraft.get_cruise().calculate(tas, fuelflow, bhp, rpm, mp, da);
                        if (!climbseg)
                                wpte.set_tas(tas);
                        double gs(tas);
                        if (wind.get_windspeed() > 0) {
                                wind.set_crs_tas(tt, tas);
                                if (std::isnan(wind.get_gs()) || std::isnan(wind.get_hdg()) || wind.get_gs() <= 0) {
                                        gs = 1;
                                } else if (!climbseg) {
                                        th = wind.get_hdg();
                                        gs = wind.get_gs();
                                        if (false)
                                                std::cerr << "recompute_wptextra: cruise: windcorr: TT " << tt << " TH " << th
                                                          << " TAS " << tas << " GS " << gs << " Wind " << wind.get_winddir() << '/' << wind.get_windspeed() << std::endl;
                                }
                        }
                        if (!climbseg)
				wpte.set_gs(gs);
                        double ct(dist / gs);
                        ltime += ct * 3600.0;
                        fuel += ct * fuelflow;
                        if (!std::isnan(rpm) && rpm > 0)
                                wpte.set_rpm(rpm);
                        if (!std::isnan(mp) && mp > 0)
                                wpte.set_mp(mp);
                }
                wpte.set_flighttime(ltime);
                wpte.set_fuel(fuel);
                double nextta(ad.get_true_altitude());
                wpte.set_truealt(nextta);
                {
                        double nextta(ad.get_true_altitude());
                        wmm.compute((prevta + nextta) * (FPlanWaypoint::ft_to_m / 2000.0), wpto.get_coord().halfway(wpt.get_coord()), curtime);
                        wpte.set_declination(wmm.get_dec());
                }
                wpte.set_trueheading(th);
        }
}

void Aircraft::navfplan(const Glib::ustring& templatefile, Engine& engine, const FPlanRoute& fplan,
			double cruise_rpm, double cruise_mp, const WeightBalance::elementvalues_t& wbev) const
{
	NavFPlan n(*this, engine, cruise_rpm, cruise_mp);
	n.begin(wbev);
	n.fpl(fplan);
	n.profile(fplan);
	n.end(templatefile);
}
